/*! @file glmain.cpp
	
	@brief GLUT{\
 
	@author Makoto Fujisawa
	@date   2010-02
*/

#pragma comment(lib, "glew32.lib")


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <GL/glew.h>

#include "rx_common.h"

// VFCv}b`O@
#include "rx_shape_matching.h"

// 3Df
#include "rx_obj.h"
#include "rx_vrml.h"

// VhE}bsO@
#include "rx_shadow.h"

// }EXɂ鎋_ړƃ}EXsbN
#include "rx_trackball.h"
#include "rx_pick.h"


//-----------------------------------------------------------------------------
// O
//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
// `/萔
//-----------------------------------------------------------------------------
const string RX_PROGRAM_NAME = "rx_shape_matching";

const GLfloat RX_FOV = 45.0f;

const GLfloat RX_LIGHT0_POS[4] = {  0.5f, 3.0f, 0.5f, 0.0f };

const double RX_GOUND_HEIGHT = -1.0;

// _F
const Vec3 VERTEX_COLOR[] = { Vec3(1.0, 0.0, 0.0), 
							  Vec3(0.0, 1.0, 0.0), 
							  Vec3(0.0, 0.0, 1.0), 
							  Vec3(0.0, 1.0, 1.0), 
							  Vec3(1.0, 0.0, 1.0), 
							  Vec3(1.0, 1.0, 0.0) };



enum{
	RXD_VERTEX		= (1 << 0),  
	RXD_EDGE		= (1 << 1), 
	RXD_FACE		= (1 << 2),
};



//-----------------------------------------------------------------------------
// O[oϐ
//-----------------------------------------------------------------------------
bool g_bIdle = false;

// EBhE
int g_iWinW = 800;		//!< `EBhE̕
int g_iWinH = 800;		//!< `EBhE̍
int g_iWinX = 100;		//!< `EBhEʒux
int g_iWinY = 50;		//!< `EBhEʒuy

bool g_bFullScreen = false;		//!< tXN[\

// }EX{^̏
int g_iMouseButton = -1;

//! `tO
int g_iDraw = RXD_FACE;

// gbN{[
rxTrackball g_tbView;

//! }EXsbN
rxGLPick g_Pick;					//!< }EXsbN
vector<int> g_vSelectedVertices;	//!< I𒸓_(_Î߂ɔzɂĂ())
int g_iPickedObj = -1;				//!< sbNꂽIuWFNg
double g_fPickDist = 1.0;			//!< sbNꂽ_܂ł̋

// |SIuWFNg
struct rxObject
{
	string filename;			//!< t@C
	rxPolygons poly;			//!< |S
	rxShapeMatching *deform;	//!< VFCv}b`O@ɂό`
	int vstart, vend;			//!< Ŝ̒_ɂʒu
};
vector<rxObject> g_vObj;

// VhE}bsO
rxShadowMap g_ShadowMap;
int g_iShadowMapSize = 4096;

int g_iStep = 0;

// V~[Vݒ
Vec3 g_v3EnvMin(-3.0, RX_GOUND_HEIGHT, -3.0);
Vec3 g_v3EnvMax(3.0, RX_GOUND_HEIGHT+3.0, 3.0);
double g_fDt = 0.01;


//-----------------------------------------------------------------------------
// ֐vg^Cv錾
//-----------------------------------------------------------------------------
void OnExitApp(int id = -1);

void Idle(void);
void CleanGL(void);


//-----------------------------------------------------------------------------
// bVό`
//-----------------------------------------------------------------------------
/*!
 * bVǂݍ
 * @param[in] fn t@C
 * @param[out] poly |S
 * @param[in] cen,ext f̒SʒuƑ傫
 */
void ReadFile(const string &fn, rxPolygons &poly, Vec3 cen, Vec3 ext)
{
	if(GetExtension(fn) == "obj"){
		ReadOBJ(fn, poly);
	}
	else if(GetExtension(fn) == "wrl"){
		ReadVRML(fn, poly);
	}

	if(poly.open){
		std::cout << fn << " have been read." << endl;
		std::cout << " the number of vertex   : " << poly.vertices.size() << endl;
		std::cout << " the number of normal   : " << poly.normals.size() << endl;
		std::cout << " the number of polygon  : " << poly.faces.size() << endl;
		std::cout << " the number of material : " << poly.materials.size() << endl;

		FitVertices(cen, ext, poly.vertices);
	}
}



/*!
 * bVIuWFNgǉ
 * @param[in] fn t@C
 * @param[in] cen,ext fǉʒuƑ傫
 * @param[in] beta ގ(,[0,1])
 */
void AddMeshObject(const string &fn, Vec3 cen, Vec3 ext, double beta)
{
	rxObject obj;

	obj.deform = new rxShapeMatching((int)g_vObj.size());

	// t@Cǂݍ
	ReadFile(fn, obj.poly, cen, ext);
	obj.filename = fn;

	int vn = (int)obj.poly.vertices.size();
	int pn = (int)obj.poly.faces.size();

	// }EXsbNpɑŜ̒_ɂʒuLĂ
	obj.vstart = (g_vObj.empty() ? 0 : g_vObj.back().vend+1);
	obj.vend = obj.vstart+(vn-1);

	// Shape Matching̐ݒ
	obj.deform->SetSimulationSpace(g_v3EnvMin, g_v3EnvMax);
	obj.deform->SetTimeStep(g_fDt);
	obj.deform->SetCollisionFunc(0);
	obj.deform->SetStiffness(0.9, beta);

	// _WZbg
	for(int i = 0; i < vn; ++i){
		obj.deform->AddVertex(obj.poly.vertices[i], 0.01);
	}

	g_vObj.push_back(obj);
}

/*!
 * bV
 */
void InitMesh(void)
{
	g_iStep = 0;

	AddMeshObject("bunny.wrl", Vec3(-0.7, 0.3, 0.0), Vec3(0.2), 0.9);
	AddMeshObject("bunny.wrl", Vec3( 0.0, 0.3, 0.0), Vec3(0.2), 0.8);
	AddMeshObject("bunny.wrl", Vec3( 0.7, 0.3, 0.0), Vec3(0.2), 0.7);

	g_vSelectedVertices.clear();
	g_iPickedObj = -1;
}

/*!
 * bVXV
 */
void UpdateMesh(void)
{
	int c = 0;
	vector<rxObject>::iterator itr = g_vObj.begin();
	while(itr != g_vObj.end()){
		itr->deform->Update();

		int n = itr->poly.vertices.size();
		for(int i = 0; i < n; ++i){
			itr->poly.vertices[i] = itr->deform->GetVertexPos(i);
		}
		CalVertexNormals(itr->poly);
		itr++;
	}
}


//-----------------------------------------------------------------------------
// AvP[V
//-----------------------------------------------------------------------------

/*!
 * ACh֐ON/OFF
 * @param[in] on trueON, falseOFF
 */
void SwitchIdle(int on)
{
	g_bIdle = (on == -1) ? !g_bIdle : (on ? true : false);
	glutIdleFunc((g_bIdle ? Idle : 0));
	cout << "idle " << (g_bIdle ? "on" : "off") << endl;
}

/*!
 * tXN[/EBhE\̐؂ւ
 */
void SwitchFullScreen(void)
{
	static int pos0[2] = { 0, 0 };
	static int win0[2] = { 500, 500 };
	if(g_bFullScreen){
		glutPositionWindow(pos0[0], pos0[1]);
		glutReshapeWindow(win0[0], win0[1]);
		g_bFullScreen = false;
	}
	else{
		pos0[0] = glutGet(GLUT_WINDOW_X);
		pos0[1] = glutGet(GLUT_WINDOW_Y);
		win0[0] = glutGet(GLUT_WINDOW_WIDTH);
		win0[1] = glutGet(GLUT_WINDOW_HEIGHT);

		glutFullScreen();
		g_bFullScreen = true;
	}

}

/*!
 * AvP[VI  Quit{^̃Cxgnh
 * @param[in] id {^̃CxgID
 */
void OnExitApp(int id)
{
	CleanGL();

	exit(1);
}





//-----------------------------------------------------------------------------
// `֐Ȃ
//-----------------------------------------------------------------------------
/*!
 * eϊ
 */
void Projection(void)
{
	gluPerspective(RX_FOV, (double)g_iWinW/(double)g_iWinH, 0.01f, 20.0f);
}
static void Projection_s(void* x){ Projection(); }

/*!
 * V[`
 */
void RenderScene(int draw)
{
	// ގ
	static const GLfloat difw[] = { 0.8f, 0.8f, 0.8f, 1.0f };	// gUF : 
	static const GLfloat difr[] = { 1.0f, 0.4f, 0.4f, 1.0f };	// gUF : 
	static const GLfloat difg[] = { 0.4f, 0.6f, 0.4f, 1.0f };	// gUF : 
	static const GLfloat difb[] = { 0.4f, 0.4f, 1.0f, 1.0f };	// gUF : 
	static const GLfloat spec[] = { 0.3f, 0.3f, 0.3f, 1.0f };	// ʔːF
	static const GLfloat ambi[] = { 0.1f, 0.1f, 0.1f, 1.0f };	// 


	glPushMatrix();

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	glEnable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	// ގݒ
	glMaterialfv(GL_FRONT, GL_SPECULAR, spec);
	glMaterialfv(GL_FRONT, GL_AMBIENT,  ambi);
	glMaterialf(GL_FRONT, GL_SHININESS, 50.0);

	glColor3f(0.0, 0.0, 1.0);
	
	// 
	glPushMatrix();
	glMaterialfv(GL_FRONT, GL_DIFFUSE, difg);
	glTranslatef(0, RX_GOUND_HEIGHT-0.105, 0);
	glScaled(10.0, 0.2, 10.0);
	glutSolidCube(1.0);
	glPopMatrix();

	// IuWFNg
	glPushMatrix();
	glMaterialfv(GL_FRONT, GL_DIFFUSE, difb);
	if(draw & (RXD_FACE | RXD_EDGE | RXD_VERTEX)){
		GLenum mode = (draw & RXD_VERTEX) ? GL_POINTS : ((draw & RXD_EDGE) ? GL_LINE : GL_FILL);
		glPolygonMode(GL_FRONT_AND_BACK, mode);

		vector<rxObject>::iterator itr = g_vObj.begin();
		while(itr != g_vObj.end()){
			int vn = (int)itr->poly.vertices.size();

			int pn = (int)itr->poly.faces.size();
			// ׂẴ|S`
			for(int i = 0; i < pn; ++i){
				rxFace *face = &(itr->poly.faces[i]);
				int n = (int)face->vert_idx.size();

				// |S`
				glBegin(GL_TRIANGLES);
				for(int j = 0; j < n; ++j){
					int idx = face->vert_idx[j];
					if(idx >= 0 && idx < vn){
						glNormal3dv(itr->poly.normals[idx].data);
						if(!face->texcoords.empty()) glTexCoord2dv(face->texcoords[j].data);
						glVertex3dv((itr->poly.vertices[idx]).data);
					}
				}
				glEnd();
			}

			itr++;
		}

	}
	glPopMatrix();

	glPopMatrix();
}
static void RenderScene_s(void* x){ RenderScene((g_iDraw & 15)); }


//------------------------------------------------------------------
// OpenGLLoXp̃Cxgnh
//------------------------------------------------------------------
/*!
 * ĕ`Cxg֐
 */
void Display(void)
{
	glClearColor(0.8, 0.8, 0.9, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	GLfloat sky_color[4] = {0.9f, 0.9f, 1.0f, 1.0f};

	// ʒu̎Zo
	Vec4 light_dir = Vec4(0.2, 1.0, 0.5, 0.0);
	normalize(light_dir);
	GLfloat light_dirf[4];
	for(int i = 0; i < 4; ++i) light_dirf[i] = (GLfloat)light_dir[i];

	// n_ʒuC̎Zo
	Vec3 eye_pos(0.0), eye_dir(0.0);
	g_tbView.CalLocalPos(eye_pos.data, Vec3(0.0).data);
	g_tbView.CalLocalRot(eye_dir.data, Vec3(0.0, 0.0, -1.0).data);

	// VhE}bv̍쐬
	vector<Vec3> bcns, bsls;
	bcns.push_back(Vec3(0.0));
	bsls.push_back(Vec3(2.0));
	g_ShadowMap.MakeShadowMap(light_dir, eye_pos, eye_dir, RenderScene_s, NULL, bcns, bsls);

	// t[obt@ƃfvXobt@NA
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	g_tbView.Apply();	// }EXɂ]Esړ̓Kp

	// ݒ
	glLightfv(GL_LIGHT0, GL_POSITION, light_dirf);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, sky_color);

	// etŕ`
	g_ShadowMap.RenderSceneWithShadow(RenderScene_s, NULL, g_iWinW, g_iWinH);

	// VhE}bṽfobOp
	//g_ShadowMap.DrawDepthTex();
	//g_ShadowMap.OverviewCam(RenderScene_s, NULL, g_iWinW, g_iWinH);

	glutSwapBuffers();
}

/*!
 * }EXsbNp`֐
 */
void DisplayForPick(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glPushMatrix();

	// }EXɂ]Esړ̓Kp
	g_tbView.Apply();

	// _`
	int c = 1;
	vector<rxObject>::iterator itr = g_vObj.begin();
	while(itr != g_vObj.end()){
		int vn = (int)itr->poly.vertices.size();
		glDisable(GL_LIGHTING);
		glPointSize(16.0);
		for(int i = 0; i < vn; ++i){
			glLoadName(c++);
			glBegin(GL_POINTS);
			glVertex3dv((itr->poly.vertices[i]).data);
			glEnd();
		}

		itr++;
	}
	glPopMatrix();

	glutSwapBuffers();
}
static void DisplayForPick_s(void* x){ DisplayForPick(); }


/*! 
 * TCYCxg֐
 * @param[in] w LoX(sNZ)
 * @param[in] h LoX(sNZ)
 */
void Resize(int w, int h)
{
	g_iWinW = w;
	g_iWinH = h;

	glViewport(0, 0, w, h);
	g_tbView.SetRegion(w, h);

	// ϊs̐ݒ
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	Projection();

	// fr[ϊs̐ݒ
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	g_ShadowMap.CalAllFrustums(w, h, RX_FOV);
}

/*!
 * }EXsbNꂽ_̃NA
 */
void ClearPick(void)
{
	g_vSelectedVertices.clear();
	if(g_iPickedObj != -1){
		for(int i = 0; i < g_vObj[g_iPickedObj].deform->GetNumVertices(); ++i) g_vObj[g_iPickedObj].deform->UnFixVertex(i);
		g_iPickedObj = -1;
	}
}

/*!
 * }EXCxg֐
 * @param[in] button }EX{^(GLUT_LEFT_BUTTON,GLUT_MIDDLE_BUTTON,GLUT_RIGHT_BUTTON)
 * @param[in] state }EX{^̏(GLUT_UP, GLUT_DOWN)
 * @param[in] x,y }EXW(XN[Wn)
 */
void Mouse(int button, int state, int x, int y)
{
	int mod = glutGetModifiers();	// SHIFT,CTRL,ALT̏Ԏ擾

	if(button == GLUT_LEFT_BUTTON){
		if(state == GLUT_DOWN){
			int hit = g_Pick.Pick(x, y);
			if(hit >= 1){
				int v = hit-1;

				// ǂ̃IuWFNg̒_Ȃ̂𔻕
				g_iPickedObj = -1;
				int size = (int)g_vObj.size();
				for(int i = 0; i < size; ++i){
					if(g_vObj[i].vstart <= v && v <= g_vObj[i].vend){
						g_iPickedObj = i;
						break;
					}
				}

				if(g_iPickedObj != -1){
					g_vSelectedVertices.resize(1);
					g_vSelectedVertices[0] = v-g_vObj[g_iPickedObj].vstart;
					cout << "vertex " << g_vSelectedVertices[0] << " is selected - " << g_iPickedObj << endl;
				
					// _sbN_܂ł̋vZ
					Vec3 ray_from;
					Vec3 init_pos = Vec3(0.0);
					g_tbView.CalLocalPos(ray_from, init_pos);
					Vec3 pos = g_vObj[g_iPickedObj].deform->GetVertexPos(g_vSelectedVertices[0]);

					g_fPickDist = length(pos-ray_from);
				}
			}
			else{
				g_tbView.Start(x, y, mod+1);
				ClearPick();
			}
		}
		else if(state == GLUT_UP){
			g_tbView.Stop(x, y);
			ClearPick();
		}
	}
	else if(button == GLUT_MIDDLE_BUTTON){
	}
	else if(button == GLUT_RIGHT_BUTTON){
	}

	glutPostRedisplay();
}

/*!
 * [VCxg֐(}EX{^܂܃hbO)
 * @param[in] x,y }EXW(XN[Wn)
 */
void Motion(int x, int y)
{
	if(g_vSelectedVertices.empty()){
		g_tbView.Motion(x, y);
	}
	else{
		Vec3 ray_from, ray_to;
		Vec3 init_pos = Vec3(0.0);
		g_tbView.CalLocalPos(ray_from, init_pos);
		g_tbView.GetRayTo(x, y, RX_FOV, ray_to);

		Vec3 dir = Unit(ray_to-ray_from);	// _}EXʒuւ̃xNg

		int v = g_vSelectedVertices[0];
		Vec3 cur_pos = g_vObj[g_iPickedObj].deform->GetVertexPos(v);
		Vec3 new_pos = ray_from+dir*g_fPickDist;
		g_vObj[g_iPickedObj].deform->FixVertex(v, new_pos);
	}

	glutPostRedisplay();
}


/*! 
 * AChCxg֐
 */
void Idle(void)
{
	UpdateMesh();

	g_iStep++;
	glutPostRedisplay();
}

/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void Keyboard(unsigned char key, int x, int y)
{
	switch(key){
	case '\033':  // '\033'  ESC  ASCII R[h
		OnExitApp();
		break;

	case ' ':	// V~[V1Xebvi߂
		Idle();
		break;

	case 's':	// Aj[VON/OFF
		SwitchIdle(-1);
		break;

	case 'F':	// tXN[ON/OFF
		SwitchFullScreen();
		break;

	case 'v':
		g_iDraw ^= RXD_VERTEX;
		break;

	case 'e':
		g_iDraw ^= RXD_EDGE;
		break;

	case 'f':
		g_iDraw ^= RXD_FACE;
		break;

	case 'c':	// Ỉ
		ClearPick();
		break;

	default:
		break;
	}

	glutPostRedisplay();
}

/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void SpecialKey(int key, int x, int y)
{
	switch(key){
	case GLUT_KEY_LEFT:
		break;
	case GLUT_KEY_RIGHT:
		break;
	default:
		break;
	}

	glutPostRedisplay();
}



/*! 
 * GL̏֐
 */
void InitGL(void)
{
	RXCOUT << "OpenGL Ver. " << glGetString(GL_VERSION) << endl;
	glewInit();

	// wiF
	glClearColor(0.8, 0.8, 0.9, 1.0);
	glClearDepth(1.0f);


	glEnable(GL_CULL_FACE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);

	// |Sݒ
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	// ̏ݒ
	static const GLfloat difw[] = { 1.0f, 1.0f, 1.0f, 1.0f };	// gUF : 
	static const GLfloat spec[] = { 0.6f, 0.6f, 0.6f, 1.0f };	// ʔːF
	static const GLfloat ambi[] = { 0.3f, 0.3f, 0.3f, 1.0f };	// 
	//glLightfv(GL_LIGHT0, GL_POSITION, RX_LIGHT0_POS);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  difw);
	glLightfv(GL_LIGHT0, GL_SPECULAR, spec);
	glLightfv(GL_LIGHT0, GL_AMBIENT,  ambi);
	glEnable(GL_LIGHTING);

	glShadeModel(GL_SMOOTH);

	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	// _̏
	g_tbView.SetScaling(-3.0);
	g_tbView.SetTranslation(0.0, 0.5);

	// VhE}bv
	g_ShadowMap.InitShadowMap(g_iWinW, g_iWinH, RX_FOV, 4, 4096);
	g_ShadowMap.SetShadowType(3);

	// Shape Matching
	InitMesh();
}

/*! 
 * GL̏I֐
 */
void CleanGL(void)
{
	vector<rxObject>::iterator itr = g_vObj.begin();
	while(itr != g_vObj.end()){
		if(itr->deform) delete (itr->deform);
		itr->deform = 0;
		itr++;
	}
}

/*!
 * Cj[
 * @param[in] id j[ID
 */
void OnMainMenu(int id)
{
	Keyboard((unsigned char)id, 0, 0);
}

/*!
 * GLUT̉ENbNj[쐬
 */
void InitMenu(void)
{
	// Cj[
	glutCreateMenu(OnMainMenu);
	glutAddMenuEntry("Toggle vertex drawing [v]",	'v');
	glutAddMenuEntry("Toggle edge drawing [e]",		'e');
	glutAddMenuEntry("Toggle face drawing [f]",		'f');
	glutAddMenuEntry("------------", -1);
	glutAddMenuEntry("Toggle Fullscreen [F]",		'F');
	glutAddMenuEntry("Toggle Animation [s]",		's');
	glutAddMenuEntry("Step Animation [ ]",			' ');
	glutAddMenuEntry(" ------------------------ ",	-1);
	glutAddMenuEntry("Exit [q]",					'q');
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}



/*!
 * C[`
 * @param[in] argc R}hC̐
 * @param[in] argv R}hC
 */
int main(int argc, char *argv[])
{
	glutInitWindowPosition(g_iWinX, g_iWinY);
	glutInitWindowSize(g_iWinW, g_iWinH);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow(argv[0]);

	glutDisplayFunc(Display);
	glutReshapeFunc(Resize);
	glutMouseFunc(Mouse);
	glutMotionFunc(Motion);
	glutKeyboardFunc(Keyboard);
	glutSpecialFunc(SpecialKey);
	glutCloseFunc(CleanGL);

	SwitchIdle(1);

	InitGL();
	InitMenu();

	g_Pick.Set(DisplayForPick_s, 0, Projection_s, 0);

	glutMainLoop();

	return 0;
}


