/*! 
  @file rx_levelset3d.cpp
	
  @brief 3xZbg@
		  - t̍\zƑx̊O}
		  - Ql : http://code.google.com/p/levelset2d/
 
  @author Makoto Fujisawa
  @date 2013-10
*/


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_levelset3d.h"


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
// אڃObhCfbNXe[u
#define NNGRID(i, j, k) { {i-1,j-1,k-1},{i,j-1,k-1},{i+1,j-1,k-1},{i-1,j,k-1},{i+1,j,k-1},{i-1,j+1,k-1},{i,j+1,k-1},{i+1,j+1,k-1},{i,j,k-1}, \
						  {i-1,j-1,k},  {i,j-1,k},  {i+1,j-1,k},  {i-1,j,k},  {i+1,j,k},  {i-1,j+1,k},  {i,j+1,k},  {i+1,j+1,k},             \
						  {i-1,j-1,k+1},{i,j-1,k+1},{i+1,j-1,k+1},{i-1,j,k+1},{i+1,j,k+1},{i-1,j+1,k+1},{i,j+1,k+1},{i+1,j+1,k+1},{i,j,k+1} }


//-----------------------------------------------------------------------------
// rxLevelSet3DNX̎
//-----------------------------------------------------------------------------
//! ftHgRXgN^
rxLevelSet3D::rxLevelSet3D()
{
	m_pGrids = 0;
	m_pGrad[0] = m_pGrad[1] = m_pGrad[2] = 0;

	m_cShowGrid = 1;
	m_cShowDist = 1;
	m_cShowRegion = 1;
}

//! fXgN^
rxLevelSet3D::~rxLevelSet3D()
{
	RX_FREE<rxGrid>(m_pGrids);
	RX_FREE<RXREAL>(m_pGrad[0]);
	RX_FREE<RXREAL>(m_pGrad[1]);
	RX_FREE<RXREAL>(m_pGrad[2]);
}

//! 
/*!
 * Obh
 * @param[in] nx,ny,nz Obh(̃\oƈقȂ苫EObh܂ޒlw)
 * @param[in] h Obh
 */
void rxLevelSet3D::Init(int nx, int ny, int nz, RXREAL h)
{
	m_iNx = nx;
	m_iNy = ny;
	m_iNz = nz;
	m_fH = h;

	int size = m_iNx*m_iNy*m_iNz;

	// Obh̊蓖
	m_pGrids   = RX_ALLOC<rxGrid>(size);
	m_pGrad[0] = RX_ALLOC<RXREAL>(size);
	m_pGrad[1] = RX_ALLOC<RXREAL>(size);
	m_pGrad[2] = RX_ALLOC<RXREAL>(size);

	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		rxGrid g;
		g.i = i;
		g.j = j;
		g.k = k;

		m_pGrids[ix(i,j,k)] = g;
		m_pGrad[0][ix(i,j,k)] = m_pGrad[1][ix(i,j,k)] = m_pGrad[2][ix(i,j,k)] = 0.0;
	}
}

/*!
 * xZbg֐l̐ݒ
 * @param[in] func A֐l^֐(t̓ŕCOŐ̒lƂȂ悤ɂ)
 * @param[in] tol  t𐶐ohC̒l^ƋꐶsȂ
 */
void rxLevelSet3D::Set(RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, RXREAL tol, bool app)
{
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		RXREAL x = i/(RXREAL)(m_iNx-1);
		RXREAL y = j/(RXREAL)(m_iNy-1);
		RXREAL z = k/(RXREAL)(m_iNz-1);
		RXREAL f = func(x, y, z, ptr);

		rxGrid &g = m_pGrids[ix(i,j,k)];

		g.state = FMM_ACCEPTED;
		g.dist = (app ? ((g.dist >= 0.0 && f < 0.0) ? f : g.dist) : f);
		g.pos = Vec3(0.0);
	}

	if(tol >= 0.0) Reinitialize(tol);

	//dump();
}
void rxLevelSet3D::Set(RXREAL *f, RXREAL tol, bool app)
{
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		rxGrid &g = m_pGrids[ix(i,j,k)];

		int idx = ix(i,j,k);
		g.state = FMM_ACCEPTED;
		g.dist = (app ? ((g.dist >= 0.0 && f[idx] < 0.0) ? f[idx] : g.dist) : f[idx]);
		g.pos = Vec3(0.0);
	}

	if(tol >= 0.0) Reinitialize(tol);

	//dump();
}

/*!
 * xZbg֐l̎擾
 * @param[out] f ֐li[(1z)
 */
void rxLevelSet3D::Get(RXREAL *f)
{
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		rxGrid &g = m_pGrids[ix(i,j,k)];

		int idx = ix(i,j,k);
		f[idx] = (g.state == FMM_ACCEPTED) ? g.dist : 1.0;
	}
}


/*!
 * ď
 *  - t̐
 * @param[in] tol 𐶐oh
 */
void rxLevelSet3D::Reinitialize(RXREAL tol)
{
	RXREAL w = m_fH;

	// lݒ肳ĂȂƂdist=1ɂ
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		if(m_pGrids[ix(i,j,k)].state != FMM_ACCEPTED){
			m_pGrids[ix(i,j,k)].dist = 1.0;
		}
		m_pGrids[ix(i,j,k)].state = FMM_ACCEPTED;
	}


	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		rxGrid &g = m_pGrids[ix(i,j,k)];

		bool far_cell = true;
		int query[][3] = NNGRID(i, j, k);
		const int qnum = 26;

		Vec3 pos[qnum];
		bool fnd[qnum];

		// ͂̃ObhTāCقȂ̂炻̊Ԃɕ\ʂƂČ_(pos)vZ
		for(int q = 0; q < qnum; ++q){
			int qi = query[q][0];
			int qj = query[q][1];
			int qk = query[q][2];

			fnd[q] = false;
			if(qi >= 0 && qi < m_iNx && qj >= 0 && qj < m_iNy && qk >= 0 && qk < m_iNz){
				if((m_pGrids[ix(qi,qj,qk)].state == FMM_ACCEPTED) && (m_pGrids[ix(qi,qj,qk)].dist*g.dist < 0.0)){
					far_cell = false;
					
					// V_
					RXREAL f0 = g.dist;
					RXREAL f1 = m_pGrids[ix(qi,qj,qk)].dist;
					RXREAL a = f0/(f0-f1);
					Vec3 p0 = Vec3(w*i, w*j, w*k);
					Vec3 p1 = Vec3(w*qi, w*qj, w*qk);
					pos[q] = (1.0-a)*p0+a*p1;
					fnd[q] = true;
				}
			}
		}

		// \ʂ痣Ăꍇ̓tO0ɂCłȂȂ1ɂċlČvZ
		if(far_cell){
			g.state = FMM_NONE;
		}
		else{
			g.state = FMM_ACCEPTED;

			// \ʂ܂ł̕tvZ
			RXREAL min_d = RX_FEQ_INF;

			int pquery[][8] = {	{13, 16, 15, 14, 12, 9, 10, 11}, 
								{13, 22, 21, 20, 12, 3, 4, 5}, 
								{15, 24, 21, 18, 10, 1, 4, 7}, 
								{22, 25, 15, 6, 3, 0, 10, 19}, 
								{5, 8, 15, 23, 20, 17, 10, 2}, 
								{13, 25, 24, 23, 12, 0, 1, 2}, 
								{13, 19, 18, 17, 12, 6, 7, 8}, 
								{16, 25, 21, 17, 9, 0, 4, 8 }, 
								{14, 23, 21, 19, 11, 2, 4, 6} };
			const int pnum = 9;

			for(int p = 0; p < pnum; ++p){
				// א2_
				for(int qi = 0; qi < 8; ++qi){
					int q = pquery[p][qi];
					for(int rpi = 1; rpi < 3; ++rpi){
						int qri = (qi+rpi)%8;
						int qr = pquery[p][qri];

						if(fnd[qr] && fnd[q]){
							Vec3 x = Vec3(i*w, j*w, k*w);
							intersect(pos[qr], pos[q], x);
							RXREAL d = (RXREAL)(norm(x-Vec3(i*w, j*w, k*w)));
							if(d < min_d){	// ŏōXV
								min_d = d;
								g.pos = x;
								g.dist = (g.dist > 0.0 ? 1.0 : -1.0)*RX_MAX(d, (RXREAL)1.0e-8);
							}
						}
					}
				}

				// 1_̂
				for(int qi = 0; qi < 8; ++qi){
					int q = pquery[p][qi];
					int q1 = pquery[p][(qi+1)%8];
					int q2 = pquery[p][(qi+2)%8];
					if(!fnd[q] && fnd[q1] && !fnd[q2]){
						Vec3 x = pos[q1];
						RXREAL d = (RXREAL)(norm(x-Vec3(i*w, j*w, k*w)));
						if(d < min_d){	// ŏōXV
							min_d = d;
							g.pos = x;
							g.dist = (g.dist > 0.0 ? 1.0 : -1.0)*RX_MAX(d, (RXREAL)1.0e-8);
						}
					}
				}
			}

		}

	} // RXFOR2(0, m_iN, 0, m_iN)

	// lݒ肳ĂȂƂdist=1ɏ
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		if(m_pGrids[ix(i,j,k)].state != FMM_ACCEPTED){
			m_pGrids[ix(i,j,k)].dist = 1.0;
		}
	}

	// Fast Marching Methodŋl\ʎ͂ɍL
	fastMarching(tol);
}

/*!
 * Fast Marching Methodŋl\ʎ͂ɍL
 * @param[in] tol 𐶐oh
 */
void rxLevelSet3D::fastMarching(RXREAL tol)
{
	GridQueue tentatives;	// Obhi[L[

	// mObh(vZς)T
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		rxGrid &g = m_pGrids[ix(i,j,k)];

		Vec3 pos;
		RXREAL dist;
		if(g.state != FMM_ACCEPTED){	// m肳ĂȂObh̏ꍇ
			g.state = FMM_NONE;
			if(updateDistance(i, j, k, pos, dist)){	// ߖTɊmObh炻ŋXVCL[ɓ
				g.dist = dist;
				if(fabs(dist) < tol){
					g.pos = pos;
					g.state = FMM_TENTATIVE;	// ɏԂύX
					tentatives.push(&m_pGrids[ix(i,j,k)]);
				}
			}
			else{
				g.dist = 1.0;
			}
		}
	}

	// ObhȂȂ܂ŋߖTmObhɂ鋗vZJԂ
	while(!tentatives.empty()){
		rxGrid *mingrid = tentatives.top();	// L[(vZς݂)mObhЂƂo
		int i = mingrid->i;
		int j = mingrid->j;
		int k = mingrid->k;
		tentatives.pop();
		mingrid->state = FMM_ACCEPTED;	// ObhVmObhɂ
		
		// VmObh̎͂ɋ`dČObhɂ
		int query[][3] = NNGRID(i, j, k);
		const int qnum = 26;
		for(int q = 0; q < qnum; ++q){
			int qi = query[q][0];
			int qj = query[q][1];
			int qk = query[q][2];

			if(qi >= 0 && qi < m_iNx && qj >= 0 && qj < m_iNy && qk >= 0 && qk < m_iNz){
				rxGrid &gq = m_pGrids[ix(qi,qj,qk)];
				if(gq.state == FMM_NONE){
					Vec3 pos;
					RXREAL dist;
					// ̖͂mObh̋XVāCL[ɓ
					if(updateDistance(qi, qj, qk, pos, dist)){
						if(fabs(dist) < tol){
							gq.dist = dist;
							gq.pos = pos;
							gq.state = FMM_TENTATIVE;
							tentatives.push(&m_pGrids[ix(qi,qj,qk)]);
						}
					}
				}
			}
		}
	}

	// \ʂtolÕObh̋l[t̓:-1,t̊O:1]ɐݒ
	scanline(0, m_iNx, m_iNy, m_iNz);
	scanline(1, m_iNz, m_iNx, m_iNy);
	scanline(2, m_iNy, m_iNz, m_iNx);
}

/*!
 * XLCTŕ\ʂtolÕObh̋l[t̓:-1,t̊O:1]ɐݒ
 * @param[in] 
 * @param[out] 
 * @return 
 */
void rxLevelSet3D::scanline(int d, int nx, int ny, int nz)
{
	for(int k = 0; k < nz; ++k){
		for(int j = 0; j < ny; ++j){
			RXREAL sgn = 1.0;
			
			// Forward Search
			for(int i = 0; i < nx; ++i){
				int idx[][3] = { {i,j,k}, {k,i,j}, {j,k,i} };
				rxGrid &g = m_pGrids[ix(idx[d][0],idx[d][1],idx[d][2])];

				if(g.state == FMM_ACCEPTED){
					sgn = g.dist < 0.0 ? -1.0 : 1.0;
				}
				else if(sgn < 0.0){
					g.dist = -1.0;
					g.state = FMM_ACCEPTED;
				}
			}
			
			// Backward Search
			for(int i = nx-1; i >= 0; --i){
				int idx[][3] = { {i,j,k}, {k,i,j}, {j,k,i} };
				rxGrid &g = m_pGrids[ix(idx[d][0],idx[d][1],idx[d][2])];
				
				if(g.state == FMM_ACCEPTED){
					sgn = g.dist < 0.0 ? -1.0 : 1.0;
				}
				else if(sgn < 0.0){
					g.dist = -1.0;
					g.state = FMM_ACCEPTED;
				}
			}
		}
	}
}


/*!
 * m̋ߖTVvZ
 * @param[in] i,j mObhʒu
 * @param[out] pos \ʍŋߓ_
 * @param[out] dist \ʂ܂ł̋
 * @return Vtrue
 */
bool rxLevelSet3D::updateDistance(int i, int j, int k, Vec3 pos, RXREAL &dist)
{
	Vec3 x = Vec3(i/(double)(m_iNx-1), j/(double)(m_iNy-1), k/(double)(m_iNz-1));
	dist = 1.0;
	bool updated = false;
	
	int query[][3] = NNGRID(i, j, k);
	const int qnum = 26;
	for(int q = 0; q < qnum; ++q){
		int qi = query[q][0];
		int qj = query[q][1];
		int qk = query[q][2];
		if(qi >= 0 && qi < m_iNx && qj >= 0 && qj < m_iNy && qk >= 0 && qk < m_iNz){
			RXREAL sgn = m_pGrids[ix(qi,qj,qk)].dist > 0 ? 1.0 : -1.0;

			// ߖTmObhłȂ
			if(m_pGrids[ix(qi,qj,qk)].state == FMM_ACCEPTED){
				// mߖTObhɊi[ꂽ\ʍŋߓ_܂ŋvZ
				RXREAL d = (RXREAL)(norm(m_pGrids[ix(qi,qj,qk)].pos-x));

				// ݂̋ȂΒu
				if(d < fabs(dist)){
					dist = sgn*d;
					pos = m_pGrids[ix(qi,qj,qk)].pos;
					updated = true;
				}
			}
		}
	}
	return updated;
}

/*!
 * t̊Oւ̕ʂ̊O}
 * @param[inout] f O}镨
 * @param[in] region
 */
void rxLevelSet3D::Extrapolate(RXREAL *f)
{
	GridQueue tantatives;	// Obhi[L[

	char *computed = RX_ALLOC<char>(m_iNx*m_iNy*m_iNz);

	// ObhT
	RXFOR3(0, m_iNx, 0, m_iNy, 0, m_iNz){
		rxGrid &g = m_pGrids[ix(i,j,k)];

		// t̊OŋlvZς݂Ȃɓ
		// ohȂ炷ׂČObhɂȂ
		if(g.dist >= 0.0 && g.state == FMM_ACCEPTED){
			tantatives.push(&g);
		}
		computed[ix(i,j,k)] = (g.dist < 0.0) ? 1 : 0;
	}

	// ObhȂȂ܂ŋߖTmObhɂvZJԂ
	while(!tantatives.empty()){
		// lłObho
		rxGrid *mingrid = tantatives.top();
		int i = mingrid->i;
		int j = mingrid->j;
		int k = mingrid->k;
		
		// 26Obh̕ςŒlXV
		int sum = 0;
		RXREAL sumq = 0.0;
		int query[][3] = NNGRID(i, j, k);
		const int qnum = 26;
		for(int nq = 0; nq < qnum; ++nq){
			int qi = query[nq][0];
			int qj = query[nq][1];
			int qk = query[nq][2];
			if(qi >= 0 && qi < m_iNx && qj >= 0 && qj < m_iNy && qk >= 0 && qk < m_iNz){
				if(computed[ix(qi,qj,qk)]){
					sumq += f[ix(qi,qj,qk)];
					sum++;
				}
			}
		}
		if(sum){
			f[ix(i,j,k)] = sumq/sum;
		}

		tantatives.pop();
		computed[ix(i,j,k)] = 1;
	}

	RX_FREE(computed);
}


//! f[^̃t@Co
void rxLevelSet3D::dump(int k)
{
	ofstream fo;
	fo.open("_sdf.txt");
	fo.precision(4);
	RXFOR2(0, m_iNy, 0, m_iNx){
		rxGrid &g = m_pGrids[ix(j,m_iNy-i-1,k)];
		fo << setw(5) << right << g.dist << (j == m_iNx-1 ? "\n" : " ");
	}
	fo << endl;
	fo.close();
}
