/*!
  @file rx_mc.h
	
  @brief A֐\ʂ̃|S(MC@)
			- Ql : http://paulbourke.net/geometry/polygonise/

  @author Makoto Fujisawa
  @date   2010-03
*/


#ifndef _RX_MC_MESH_H_
#define _RX_MC_MESH_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cstdlib>

// STL
#include <map>
#include <vector>
#include <string>

#include <iostream>

#include "rx_utility.h"
#include "rx_mesh.h"


//-----------------------------------------------------------------------------
// O
//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
typedef unsigned int uint;

#ifndef RXREAL
	#define RXREAL float
#endif


struct RxVertexID
{
	uint newID;
	double x, y, z;
};

typedef std::map<uint, RxVertexID> ID2VertexID;

struct RxTriangle
{
	uint pointID[3];
};

typedef std::vector<RxTriangle> RxTriangleVector;

struct RxScalarField
{
	uint iNum[3];
	Vec3 fWidth;
	Vec3 fMin;
};



//-----------------------------------------------------------------------------
// rxMCMeshCPUNX
//-----------------------------------------------------------------------------
class rxMCMeshCPU
{
protected:
	uint m_nVertices;						//!< lʃbV̒_
	uint m_nNormals;						//!< lʃbV̒_@(쐬Ă @=_)
	uint m_nTriangles;						//!< lʃbV̎Op`|S

	ID2VertexID m_i2pt3idVertices;			//!< lʂ`钸_̃Xg
	RxTriangleVector m_trivecTriangles;		//!< Op`|S`钸_̃Xg

	RxScalarField m_Grid;					//!< Obh

	// A֐l(XJ[l)擾pϐ(ǂ炩̂ݗp)
	const RXREAL* m_ptScalarField;			//!< XJ[lێTv{[
	RXREAL (*m_fpScalarFunc)(RXREAL, RXREAL, RXREAL, void*);	//!< XJ[lԂ֐|C^
	void *m_pFuncPtr;

	RXREAL m_tIsoLevel;						//!< 臒l
	bool m_bValidSurface;					//!< bV̉

	// bV\zp̃e[u
	static const uint m_edgeTable[256];
	static const int m_triTable[256][16];


public:
	// RXgN^
	rxMCMeshCPU();

	// fXgN^
	~rxMCMeshCPU();
	
	//! A֐Op`bV𐶐
	bool CreateMesh(RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, Vec3 min_p, double h, int n[3], RXREAL threshold, 
					vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[Op`bV𐶐
	bool CreateMeshV(RXREAL *field, Vec3 min_p, double h, int n[3], RXREAL threshold, 
					 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face);

	//! Tv{[瓙lʃbV
	void GenerateSurface(const RxScalarField sf, RXREAL *field, RXREAL threshold, 
						 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! ֐Tv{[쐬ēlʃbV
	void GenerateSurfaceV(const RxScalarField sf, RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, RXREAL threshold, 
						  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! ֐瓙lʃbV
	void GenerateSurfaceF(const RxScalarField sf, RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, RXREAL threshold, 
						  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris);

	//! lʍ쐬trueԂ
	bool IsSurfaceValid() const { return m_bValidSurface; }

	//! 쐬lʃbV̔j
	void DeleteSurface();

	//! bVɗpObh̑傫(bV쐬ĂȂꍇ͕Ԓl-1)
	int GetVolumeLengths(double& fVolLengthX, double& fVolLengthY, double& fVolLengthZ);

	// 쐬bV̏
	uint GetNumVertices(void) const { return m_nVertices; }
	uint GetNumTriangles(void) const { return m_nTriangles; }
	uint GetNumNormals(void) const { return m_nNormals; }

protected:
	//! GbWID
	uint getEdgeID(uint i, uint j, uint k, uint edge);

	//! _ID
	uint getVertexID(uint i, uint j, uint k);

	// GbW̓l_vZ
	RxVertexID calIntersection(uint nX, uint nY, uint nZ, uint nEdgeNo);
	RxVertexID calIntersectionF(uint nX, uint nY, uint nZ, uint nEdgeNo);

	//! ObhGbW[̉A֐l^Ԃœl_vZ
	RxVertexID interp(double fX1, double fY1, double fZ1, double fX2, double fY2, double fZ2, RXREAL tVal1, RXREAL tVal2);

	//! _CbV􉽏o͌`Ŋi[
	void packVertex(vector<Vec3> &vrts, uint &nvrts, vector<int> &tris, uint &ntris);

	//! _@vZ
	void calNormals(const vector<Vec3> &vrts, uint nvrts, const vector<int> &tris, uint ntris, 
					vector<Vec3> &nrms, uint &nnrms);

};



#endif // _RX_MC_MESH_H_

