/*!
  @file rx_mc_cpu.cpp
	
  @brief A֐\ʂ̃|S(MC@)
			- Ql : http://paulbourke.net/geometry/polygonise/
	
  @author Makoto Fujisawa
  @date   2010-03
*/


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <math.h>
#include "rx_mc.h"
#include "rx_mc_tables.h"



//-----------------------------------------------------------------------------
// rxMCMeshCPU̎
//-----------------------------------------------------------------------------
/*!
 * RXgN^
 */
rxMCMeshCPU::rxMCMeshCPU()
{
	// MARK:RXgN^
	m_Grid.fMin = Vec3(0.0);
	m_Grid.fWidth = Vec3(0.0);
	m_Grid.iNum[0] = 0;
	m_Grid.iNum[1] = 0;
	m_Grid.iNum[2] = 0;

	m_nTriangles = 0;
	m_nNormals = 0;
	m_nVertices = 0;

	m_ptScalarField = NULL;
	m_fpScalarFunc = 0;
	m_pFuncPtr = 0;
	m_tIsoLevel = 0;
	m_bValidSurface = false;
}

/*!
 * fXgN^
 */
rxMCMeshCPU::~rxMCMeshCPU()
{
	DeleteSurface();
}


/*!
 * A֐Op`bV𐶐
 * @param[in] func A֐l擾p֐|C^
 * @param[in] min_p Obh̍ŏW
 * @param[in] h Obh̕
 * @param[in] n[3] Obh(x,y,z)
 * @param[in] threshold l(A֐l̒l̂ƂbV)
 * @param[in] method bV@("mc", "rmt", "bloomenthal")
 * @param[out] vrts _W
 * @param[out] nrms _@
 * @param[out] tris bV
 * @retval true  bV
 * @retval false bVs
 */
bool rxMCMeshCPU::CreateMesh(RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, Vec3 min_p, double h, int n[3], RXREAL threshold, 
							 vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face)
{
	if(func == NULL) return false;

	RxScalarField sf;
	for(int i = 0; i < 3; ++i){
		sf.iNum[i] = n[i];
		sf.fWidth[i] = h;
		sf.fMin[i] = min_p[i];
	}

	vector<int> tris;
	GenerateSurfaceV(sf, func, ptr, threshold, vrts, nrms, tris);

	if(IsSurfaceValid()){
		int nv = (int)GetNumVertices();
		int nm = (int)GetNumTriangles();
		int nn = (int)GetNumNormals();

		face.resize(nm);
		for(int i = 0; i < nm; ++i){
			face[i].vert_idx.resize(3);
			for(int j = 0; j < 3; ++j){
				face[i][j] = tris[3*i+(2-j)];
			}
		}

		return true;
	}

	return false;
}




/*!
 * A֐Op`bV𐶐
 * @param[in] field Tv{[
 * @param[in] min_p Obh̍ŏW
 * @param[in] h Obh̕
 * @param[in] n[3] Obh(x,y,z)
 * @param[in] threshold l(A֐l̒l̂ƂbV)
 * @param[in] method bV@("mc", "rmt", "bloomenthal")
 * @param[out] vrts _W
 * @param[out] nrms _@
 * @param[out] tris bV
 * @retval true  bV
 * @retval false bVs
 */
bool rxMCMeshCPU::CreateMeshV(RXREAL *field, Vec3 min_p, double h, int n[3], RXREAL threshold, 
								  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<rxFace> &face)
{
	if(field == NULL) return false;

	RxScalarField sf;
	for(int i = 0; i < 3; ++i){
		sf.iNum[i] = n[i];
		sf.fWidth[i] = h;
		sf.fMin[i] = min_p[i];
	}

	vector<int> tris;
	GenerateSurface(sf, field, threshold, vrts, nrms, tris);

	if(IsSurfaceValid()){
		int nv = (int)GetNumVertices();
		int nm = (int)GetNumTriangles();
		int nn = (int)GetNumNormals();

		face.resize(nm);
		for(int i = 0; i < nm; ++i){
			face[i].vert_idx.resize(3);
			for(int j = 0; j < 3; ++j){
				face[i][j] = tris[3*i+(2-j)];
			}
		}

		return true;
	}

	return false;
}


/*!
 * bV
 * @param[in] sf Obh
 * @param[in] field Tv{[
 * @param[in] threshold 臒l
 * @param[out] vrts bV_
 * @param[out] nrms bV_@
 * @param[out] tris bV􉽏(_ڑ)
 */
void rxMCMeshCPU::GenerateSurface(const RxScalarField sf, RXREAL *field, RXREAL threshold, 
								  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris)
{
	if(m_bValidSurface){
		DeleteSurface();
	}

	m_tIsoLevel = threshold;
	m_Grid.iNum[0] = sf.iNum[0];
	m_Grid.iNum[1] = sf.iNum[1];
	m_Grid.iNum[2] = sf.iNum[2];
	m_Grid.fWidth = sf.fWidth;
	m_Grid.fMin = sf.fMin;
	m_ptScalarField = field;

	uint slice0 = (m_Grid.iNum[0]+1);
	uint slice1 = slice0*(m_Grid.iNum[1]+1);

	// lʂ̐
	for(uint z = 0; z < m_Grid.iNum[2]; ++z){
		for(uint y = 0; y < m_Grid.iNum[1]; ++y){
			for(uint x = 0; x < m_Grid.iNum[0]; ++x){
				// Obh̒_zue[uQƗpCfbNX̌vZ
				uint tableIndex = 0;
				if(m_ptScalarField[z*slice1+y*slice0+x] < m_tIsoLevel)
					tableIndex |= 1;
				if(m_ptScalarField[z*slice1+(y+1)*slice0+x] < m_tIsoLevel)
					tableIndex |= 2;
				if(m_ptScalarField[z*slice1+(y+1)*slice0+(x+1)] < m_tIsoLevel)
					tableIndex |= 4;
				if(m_ptScalarField[z*slice1+y*slice0+(x+1)] < m_tIsoLevel)
					tableIndex |= 8;
				if(m_ptScalarField[(z+1)*slice1+y*slice0+x] < m_tIsoLevel)
					tableIndex |= 16;
				if(m_ptScalarField[(z+1)*slice1+(y+1)*slice0+x] < m_tIsoLevel)
					tableIndex |= 32;
				if(m_ptScalarField[(z+1)*slice1+(y+1)*slice0+(x+1)] < m_tIsoLevel)
					tableIndex |= 64;
				if(m_ptScalarField[(z+1)*slice1+y*slice0+(x+1)] < m_tIsoLevel)
					tableIndex |= 128;

				if(edgeTable[tableIndex] != 0){
					// GbW̒_Zo
					if(edgeTable[tableIndex] & 8){
						RxVertexID pt = calIntersection(x, y, z, 3);
						uint id = getEdgeID(x, y, z, 3);
						m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
					}
					if(edgeTable[tableIndex] & 1){
						RxVertexID pt = calIntersection(x, y, z, 0);
						uint id = getEdgeID(x, y, z, 0);
						m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
					}
					if(edgeTable[tableIndex] & 256){
						RxVertexID pt = calIntersection(x, y, z, 8);
						uint id = getEdgeID(x, y, z, 8);
						m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
					}
					
					if(x == m_Grid.iNum[0] - 1){
						if(edgeTable[tableIndex] & 4){
							RxVertexID pt = calIntersection(x, y, z, 2);
							uint id = getEdgeID(x, y, z, 2);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
						if(edgeTable[tableIndex] & 2048){
							RxVertexID pt = calIntersection(x, y, z, 11);
							uint id = getEdgeID(x, y, z, 11);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					}
					if(y == m_Grid.iNum[1] - 1){
						if(edgeTable[tableIndex] & 2){
							RxVertexID pt = calIntersection(x, y, z, 1);
							uint id = getEdgeID(x, y, z, 1);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
						if(edgeTable[tableIndex] & 512){
							RxVertexID pt = calIntersection(x, y, z, 9);
							uint id = getEdgeID(x, y, z, 9);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					}
					if(z == m_Grid.iNum[2] - 1){
						if(edgeTable[tableIndex] & 16){
							RxVertexID pt = calIntersection(x, y, z, 4);
							uint id = getEdgeID(x, y, z, 4);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
						if(edgeTable[tableIndex] & 128){
							RxVertexID pt = calIntersection(x, y, z, 7);
							uint id = getEdgeID(x, y, z, 7);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					}
					if((x==m_Grid.iNum[0] - 1) && (y==m_Grid.iNum[1] - 1))
						if(edgeTable[tableIndex] & 1024){
							RxVertexID pt = calIntersection(x, y, z, 10);
							uint id = getEdgeID(x, y, z, 10);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					if((x==m_Grid.iNum[0] - 1) && (z==m_Grid.iNum[2] - 1))
						if(edgeTable[tableIndex] & 64){
							RxVertexID pt = calIntersection(x, y, z, 6);
							uint id = getEdgeID(x, y, z, 6);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					if((y==m_Grid.iNum[1] - 1) && (z==m_Grid.iNum[2] - 1))
						if(edgeTable[tableIndex] & 32){
							RxVertexID pt = calIntersection(x, y, z, 5);
							uint id = getEdgeID(x, y, z, 5);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					
					// |S
					for(uint i = 0; triTable[tableIndex][i] != 255; i += 3){
						RxTriangle triangle;
						uint pointID0, pointID1, pointID2;
						pointID0 = getEdgeID(x, y, z, triTable[tableIndex][i]);
						pointID1 = getEdgeID(x, y, z, triTable[tableIndex][i+1]);
						pointID2 = getEdgeID(x, y, z, triTable[tableIndex][i+2]);
						triangle.pointID[0] = pointID0;
						triangle.pointID[1] = pointID1;
						triangle.pointID[2] = pointID2;
						m_trivecTriangles.push_back(triangle);
					}
				}
			}
		}
	}

	// _l߂
	packVertex(vrts, m_nVertices, tris, m_nTriangles);

	// @vZ
	calNormals(vrts, m_nVertices, tris, m_nTriangles, nrms, m_nNormals);

	m_bValidSurface = true;
}

/*!
 * bV(Tv{[쐬)
 * @param[in] sf Obh
 * @param[in] func A֐l擾֐|C^
 * @param[in] threshold 臒l
 * @param[out] vrts bV_
 * @param[out] nrms bV_@
 * @param[out] tris bV􉽏(_ڑ)
 */
void rxMCMeshCPU::GenerateSurfaceV(const RxScalarField sf, RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, RXREAL threshold, 
										  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris)
{
	int nx, ny, nz;
	nx = sf.iNum[0]+1;
	ny = sf.iNum[1]+1;
	nz = sf.iNum[2]+1;

	Vec3 minp = sf.fMin;
	Vec3 d = sf.fWidth;

	RXREAL *field = new RXREAL[nx*ny*nz];
	for(int k = 0; k < nz; ++k){
		for(int j = 0; j < ny; ++j){
			for(int i = 0; i < nx; ++i){
				int idx = k*nx*ny+j*nx+i;
				Vec3 pos = minp+Vec3(i, j, k)*d;

				RXREAL val = func(pos[0], pos[1], pos[2], ptr);
				field[idx] = val;
			}
		}
	}

	GenerateSurface(sf, field, threshold, vrts, nrms, tris);

	delete [] field;
}

/*!
 * bV(֐)
 * @param[in] sf Obh
 * @param[in] func A֐l擾֐|C^
 * @param[in] threshold 臒l
 * @param[out] vrts bV_
 * @param[out] nrms bV_@
 * @param[out] tris bV􉽏(_ڑ)
 */
void rxMCMeshCPU::GenerateSurfaceF(const RxScalarField sf, RXREAL (*func)(RXREAL, RXREAL, RXREAL, void*), void* ptr, RXREAL threshold, 
										  vector<Vec3> &vrts, vector<Vec3> &nrms, vector<int> &tris)
{
	if(m_bValidSurface){
		DeleteSurface();
	}

	m_tIsoLevel = threshold;
	m_Grid.iNum[0] = sf.iNum[0];
	m_Grid.iNum[1] = sf.iNum[1];
	m_Grid.iNum[2] = sf.iNum[2];
	m_Grid.fWidth = sf.fWidth;
	m_Grid.fMin = sf.fMin;
	m_fpScalarFunc = func;
	m_pFuncPtr = ptr;

	uint slice0 = (m_Grid.iNum[0]+1);
	uint slice1 = slice0*(m_Grid.iNum[1]+1);

	double dx = m_Grid.fWidth[0];
	double dy = m_Grid.fWidth[1];
	double dz = m_Grid.fWidth[2];
	
	for(uint k = 0; k < m_Grid.iNum[2]; ++k){
		for(uint j = 0; j < m_Grid.iNum[1]; ++j){
			for(uint i = 0; i < m_Grid.iNum[0]; ++i){
				double x, y, z;
				x = m_Grid.fMin[0]+i*m_Grid.fWidth[0];
				y = m_Grid.fMin[1]+j*m_Grid.fWidth[1];
				z = m_Grid.fMin[2]+k*m_Grid.fWidth[2];

				// Obh̒_zue[uQƗpCfbNX̌vZ
				uint tableIndex = 0;
				if(m_fpScalarFunc(x,    y,    z,    m_pFuncPtr) < m_tIsoLevel) tableIndex |= 1;
				if(m_fpScalarFunc(x,    y+dy, z,    m_pFuncPtr) < m_tIsoLevel) tableIndex |= 2;
				if(m_fpScalarFunc(x+dx, y+dy, z,    m_pFuncPtr) < m_tIsoLevel) tableIndex |= 4;
				if(m_fpScalarFunc(x+dx, y,    z,    m_pFuncPtr) < m_tIsoLevel) tableIndex |= 8;
				if(m_fpScalarFunc(x,    y,    z+dz, m_pFuncPtr) < m_tIsoLevel) tableIndex |= 16;
				if(m_fpScalarFunc(x,    y+dy, z+dz, m_pFuncPtr) < m_tIsoLevel) tableIndex |= 32;
				if(m_fpScalarFunc(x+dx, y+dy, z+dz, m_pFuncPtr) < m_tIsoLevel) tableIndex |= 64;
				if(m_fpScalarFunc(x+dx, y,    z+dz, m_pFuncPtr) < m_tIsoLevel) tableIndex |= 128;

				// _C􉽏񐶐
				if(edgeTable[tableIndex] != 0){
					if(edgeTable[tableIndex] & 8){
						RxVertexID pt = calIntersectionF(i, j, k, 3);
						uint id = getEdgeID(i, j, k, 3);
						m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
					}
					if(edgeTable[tableIndex] & 1){
						RxVertexID pt = calIntersectionF(i, j, k, 0);
						uint id = getEdgeID(i, j, k, 0);
						m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
					}
					if(edgeTable[tableIndex] & 256){
						RxVertexID pt = calIntersectionF(i, j, k, 8);
						uint id = getEdgeID(i, j, k, 8);
						m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
					}
					
					if(i == m_Grid.iNum[0] - 1){
						if(edgeTable[tableIndex] & 4){
							RxVertexID pt = calIntersectionF(i, j, k, 2);
							uint id = getEdgeID(i, j, k, 2);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
						if(edgeTable[tableIndex] & 2048){
							RxVertexID pt = calIntersectionF(i, j, k, 11);
							uint id = getEdgeID(i, j, k, 11);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					}
					if(j == m_Grid.iNum[1] - 1){
						if(edgeTable[tableIndex] & 2){
							RxVertexID pt = calIntersectionF(i, j, k, 1);
							uint id = getEdgeID(i, j, k, 1);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
						if(edgeTable[tableIndex] & 512){
							RxVertexID pt = calIntersectionF(i, j, k, 9);
							uint id = getEdgeID(i, j, k, 9);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					}
					if(k == m_Grid.iNum[2] - 1){
						if(edgeTable[tableIndex] & 16){
							RxVertexID pt = calIntersectionF(i, j, k, 4);
							uint id = getEdgeID(i, j, k, 4);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
						if(edgeTable[tableIndex] & 128){
							RxVertexID pt = calIntersectionF(i, j, k, 7);
							uint id = getEdgeID(i, j, k, 7);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					}
					if((i == m_Grid.iNum[0]-1) && (j == m_Grid.iNum[1]-1))
						if(edgeTable[tableIndex] & 1024){
							RxVertexID pt = calIntersectionF(i, j, k, 10);
							uint id = getEdgeID(i, j, k, 10);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					if((i == m_Grid.iNum[0]-1) && (k == m_Grid.iNum[2]-1))
						if(edgeTable[tableIndex] & 64){
							RxVertexID pt = calIntersectionF(i, j, k, 6);
							uint id = getEdgeID(i, j, k, 6);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					if((j == m_Grid.iNum[1]-1) && (k == m_Grid.iNum[2]-1))
						if(edgeTable[tableIndex] & 32){
							RxVertexID pt = calIntersectionF(i, j, k, 5);
							uint id = getEdgeID(i, j, k, 5);
							m_i2pt3idVertices.insert(ID2VertexID::value_type(id, pt));
						}
					
					for(uint t = 0; triTable[tableIndex][t] != 255; t += 3){
						RxTriangle triangle;
						uint pointID0, pointID1, pointID2;
						pointID0 = getEdgeID(i, j, k, triTable[tableIndex][t]);
						pointID1 = getEdgeID(i, j, k, triTable[tableIndex][t+1]);
						pointID2 = getEdgeID(i, j, k, triTable[tableIndex][t+2]);
						triangle.pointID[0] = pointID0;
						triangle.pointID[1] = pointID1;
						triangle.pointID[2] = pointID2;
						m_trivecTriangles.push_back(triangle);
					}
				}
			}
		}
	}
	
	packVertex(vrts, m_nVertices, tris, m_nTriangles);
	calNormals(vrts, m_nVertices, tris, m_nTriangles, nrms, m_nNormals);
	m_bValidSurface = true;
}

/*!
 * lʃbV̔j
 */
void rxMCMeshCPU::DeleteSurface()
{
	m_Grid.fWidth[0] = 0;
	m_Grid.fWidth[1] = 0;
	m_Grid.fWidth[2] = 0;
	m_Grid.iNum[0] = 0;
	m_Grid.iNum[1] = 0;
	m_Grid.iNum[2] = 0;

	m_nTriangles = 0;
	m_nNormals = 0;
	m_nVertices = 0;

	m_ptScalarField = NULL;
	m_tIsoLevel = 0;
	m_bValidSurface = false;
}

/*!
 * bVɗpObh̑傫
 * @param[out] fVolLength* Obh̑傫
 * @return bVĂ1, łȂ-1
 */
int rxMCMeshCPU::GetVolumeLengths(double& fVolLengthX, double& fVolLengthY, double& fVolLengthZ)
{
	if(IsSurfaceValid()){
		fVolLengthX = m_Grid.fWidth[0]*m_Grid.iNum[0];
		fVolLengthY = m_Grid.fWidth[1]*m_Grid.iNum[1];
		fVolLengthZ = m_Grid.fWidth[2]*m_Grid.iNum[2];
		return 1;
	}
	else
		return -1;
}

/*!
 * GbWID̎擾
 * @param[in] nx,ny,nz Obhʒu
 * @param[in] edge GbWԍ
 * @return GbWID
 */
uint rxMCMeshCPU::getEdgeID(uint i, uint j, uint k, uint edge)
{
	switch(edge){
	case 0:
		return getVertexID(i, j, k)+1;
	case 1:
		return getVertexID(i, j+1, k);
	case 2:
		return getVertexID(i+1, j, k)+1;
	case 3:
		return getVertexID(i, j, k);
	case 4:
		return getVertexID(i, j, k+1)+1;
	case 5:
		return getVertexID(i, j+1, k+1);
	case 6:
		return getVertexID(i+1, j, k+1)+1;
	case 7:
		return getVertexID(i, j, k+1);
	case 8:
		return getVertexID(i, j, k)+2;
	case 9:
		return getVertexID(i, j+1, k)+2;
	case 10:
		return getVertexID(i+1, j+1, k)+2;
	case 11:
		return getVertexID(i+1, j, k)+2;
	default:
		return -1;
	}
}

/*!
 * _ID̎擾
 * @param[in] nx,ny,nz Obhʒu
 * @return _ID
 */
uint rxMCMeshCPU::getVertexID(uint i, uint j, uint k)
{
	return 3*(k*(m_Grid.iNum[1]+1)*(m_Grid.iNum[0]+1)+j*(m_Grid.iNum[0]+1)+i);
}


/*!
 * ԂɂGbW̓l_vZ(Tv{[)
 * @param[in] nx,ny,nz Obhʒu
 * @param[in] edge GbWԍ
 * @return bV_
 */
RxVertexID rxMCMeshCPU::calIntersection(uint nx, uint ny, uint nz, uint edge)
{
	double x1, y1, z1, x2, y2, z2;
	uint v1x = nx, v1y = ny, v1z = nz;
	uint v2x = nx, v2y = ny, v2z = nz;
	
	switch(edge){
	case 0:
		v2y += 1;
		break;
	case 1:
		v1y += 1;
		v2x += 1;
		v2y += 1;
		break;
	case 2:
		v1x += 1;
		v1y += 1;
		v2x += 1;
		break;
	case 3:
		v1x += 1;
		break;
	case 4:
		v1z += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 5:
		v1y += 1;
		v1z += 1;
		v2x += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 6:
		v1x += 1;
		v1y += 1;
		v1z += 1;
		v2x += 1;
		v2z += 1;
		break;
	case 7:
		v1x += 1;
		v1z += 1;
		v2z += 1;
		break;
	case 8:
		v2z += 1;
		break;
	case 9:
		v1y += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 10:
		v1x += 1;
		v1y += 1;
		v2x += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 11:
		v1x += 1;
		v2x += 1;
		v2z += 1;
		break;
	}

	// 2[_W
	x1 = m_Grid.fMin[0]+v1x*m_Grid.fWidth[0];
	y1 = m_Grid.fMin[1]+v1y*m_Grid.fWidth[1];
	z1 = m_Grid.fMin[2]+v1z*m_Grid.fWidth[2];
	x2 = m_Grid.fMin[0]+v2x*m_Grid.fWidth[0];
	y2 = m_Grid.fMin[1]+v2y*m_Grid.fWidth[1];
	z2 = m_Grid.fMin[2]+v2z*m_Grid.fWidth[2];

	// 2[_ł̊֐l
	RXREAL val1 = m_ptScalarField[v1z*(m_Grid.iNum[1]+1)*(m_Grid.iNum[0]+1)+v1y*(m_Grid.iNum[0]+1)+v1x];
	RXREAL val2 = m_ptScalarField[v2z*(m_Grid.iNum[1]+1)*(m_Grid.iNum[0]+1)+v2y*(m_Grid.iNum[0]+1)+v2x];

	// Ԃɂl߂
	RxVertexID intersection = interp(x1, y1, z1, x2, y2, z2, val1, val2);
	
	return intersection;
}

/*!
 * ԂɂGbW̓l_vZ(֐gp)
 * @param[in] nx,ny,nz Obhʒu
 * @param[in] edge GbWԍ
 * @return bV_
 */
RxVertexID rxMCMeshCPU::calIntersectionF(uint nx, uint ny, uint nz, uint edge)
{
	double x1, y1, z1, x2, y2, z2;
	uint v1x = nx, v1y = ny, v1z = nz;
	uint v2x = nx, v2y = ny, v2z = nz;
	
	switch(edge){
	case 0:
		v2y += 1;
		break;
	case 1:
		v1y += 1;
		v2x += 1;
		v2y += 1;
		break;
	case 2:
		v1x += 1;
		v1y += 1;
		v2x += 1;
		break;
	case 3:
		v1x += 1;
		break;
	case 4:
		v1z += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 5:
		v1y += 1;
		v1z += 1;
		v2x += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 6:
		v1x += 1;
		v1y += 1;
		v1z += 1;
		v2x += 1;
		v2z += 1;
		break;
	case 7:
		v1x += 1;
		v1z += 1;
		v2z += 1;
		break;
	case 8:
		v2z += 1;
		break;
	case 9:
		v1y += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 10:
		v1x += 1;
		v1y += 1;
		v2x += 1;
		v2y += 1;
		v2z += 1;
		break;
	case 11:
		v1x += 1;
		v2x += 1;
		v2z += 1;
		break;
	}

	// 2[_W
	x1 = m_Grid.fMin[0]+v1x*m_Grid.fWidth[0];
	y1 = m_Grid.fMin[1]+v1y*m_Grid.fWidth[1];
	z1 = m_Grid.fMin[2]+v1z*m_Grid.fWidth[2];
	x2 = m_Grid.fMin[0]+v2x*m_Grid.fWidth[0];
	y2 = m_Grid.fMin[1]+v2y*m_Grid.fWidth[1];
	z2 = m_Grid.fMin[2]+v2z*m_Grid.fWidth[2];

	// 2[_ł̊֐l
	RXREAL val1 = m_fpScalarFunc(x1, y1, z1, m_pFuncPtr);
	RXREAL val2 = m_fpScalarFunc(x2, y2, z2, m_pFuncPtr);

	// Ԃɂl߂
	RxVertexID intersection = interp(x1, y1, z1, x2, y2, z2, val1, val2);
	
	return intersection;
}

/*!
 * ObhGbW[̉A֐l^Ԃœl_vZ
 * @param[in] fX1,fY1,fZ1 [_W1
 * @param[in] fX2,fY2,fZ2 [_W2
 * @param[in] tVal1 [_W1ł̃XJ[l
 * @param[in] tVal2 [_W2ł̃XJ[l
 * @return _
 */
RxVertexID rxMCMeshCPU::interp(double fX1, double fY1, double fZ1, double fX2, double fY2, double fZ2, RXREAL tVal1, RXREAL tVal2)
{
	RxVertexID interp;
	RXREAL mu;

	mu = RXREAL((m_tIsoLevel - tVal1))/(tVal2 - tVal1);
	interp.x = fX1+mu*(fX2 - fX1);
	interp.y = fY1+mu*(fY2 - fY1);
	interp.z = fZ1+mu*(fZ2 - fZ1);

	return interp;
}


/*!
 * bV_C􉽏o͌`Ŋi[
 * @param[out] vrts _W
 * @param[out] nvrts _
 * @param[out] tris Op`|S􉽏
 * @param[out] ntris Op`|S
 */
void rxMCMeshCPU::packVertex(vector<Vec3> &vrts, uint &nvrts, vector<int> &tris, uint &ntris)
{
	uint nextID = 0;

	// _l߂Ƃ̐VCfbNXi[
	ID2VertexID::iterator vitr = m_i2pt3idVertices.begin();
	while(vitr != m_i2pt3idVertices.end()){
		(*vitr).second.newID = nextID++;
		vitr++;
	}

	// _ʒu̕ωɉăbV\钸_CfbNX쐬
	RxTriangleVector::iterator titr = m_trivecTriangles.begin();
	while(titr != m_trivecTriangles.end()){
		for(uint i = 0; i < 3; i++){
			uint newID = m_i2pt3idVertices[(*titr).pointID[i]].newID;
			(*titr).pointID[i] = newID;
		}
		titr++;
	}

	// 쐬񂩂pbNꂽ_ƃbV쐬
	vitr = m_i2pt3idVertices.begin();
	nvrts = (int)m_i2pt3idVertices.size();
	vrts.resize(nvrts);
	for(uint i = 0; i < nvrts; i++, vitr++){
		vrts[i][0] = (*vitr).second.x;
		vrts[i][1] = (*vitr).second.y;
		vrts[i][2] = (*vitr).second.z;
	}
	titr = m_trivecTriangles.begin();
	ntris = (int)m_trivecTriangles.size();
	tris.resize(ntris*3);
	for(uint i = 0; i < ntris; i++, titr++){
		tris[3*i+0] = (*titr).pointID[0];
		tris[3*i+1] = (*titr).pointID[1];
		tris[3*i+2] = (*titr).pointID[2];
	}

	m_i2pt3idVertices.clear();
	m_trivecTriangles.clear();
}

/*!
 * _@vZ
 * @param[in] vrts _W
 * @param[in] nvrts _
 * @param[in] tris Op`|S􉽏
 * @param[in] ntris Op`|S
 * @param[out] nrms @
 * @param[out] nnrms @(=_)
 */
void rxMCMeshCPU::calNormals(const vector<Vec3> &vrts, uint nvrts, const vector<int> &tris, uint ntris, 
									   vector<Vec3> &nrms, uint &nnrms)
{
	nnrms = nvrts;
	nrms.resize(nnrms);
	
	for(uint i = 0; i < nnrms; i++){
		nrms[i][0] = 0;
		nrms[i][1] = 0;
		nrms[i][2] = 0;
	}

	// OςŃbV@߂Ăe_ɐώZ
	for(uint i = 0; i < ntris; i++){
		Vec3 vec1, vec2, normal;
		uint id0, id1, id2;
		id0 = tris[3*i+0];
		id1 = tris[3*i+1];
		id2 = tris[3*i+2];

		vec1 = vrts[id1]-vrts[id0];
		vec2 = vrts[id2]-vrts[id0];
		normal = cross(vec1, vec2);

		nrms[id0] += normal;
		nrms[id1] += normal;
		nrms[id2] += normal;
	}

	// ڑ郁bV̖@𕽋ς邱ƂŒ_@vZ
	for(uint i = 0; i < nnrms; i++){
		normalize(nrms[i]);
	}
}

