/*!
  @file rx_ps.h
	
  @brief p[eBNV~[V̊NX
 
  @author Makoto Fujisawa
  @date 2011-06
*/

#ifndef _RX_PS_H_
#define _RX_PS_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <GL/glew.h>
#include "rx_common.h"

#include "rx_sph_solid.h"



//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
#ifndef DIM
	#define DIM 4
#endif

const int RX_MAX_STEPS = 100000;


//-----------------------------------------------------------------------------
// p[eBNV~[V̊NX
//-----------------------------------------------------------------------------
class rxParticleSystemBase
{
public:

	enum rxParticleConfig
	{
		RX_CONFIG_RANDOM,
		RX_CONFIG_GRID,
		RX_CONFIG_BOX,
		RX_CONFIG_NONE, 
		_NUM_CONFIGS
	};

	enum rxParticleArray
	{
		RX_POSITION = 0,
		RX_VELOCITY,
		RX_NORMAL, 
		RX_FORCE, 
		RX_DENSITY, 
		RX_PRESSURE, 

		RX_CONSTANT, 
		RX_RAMP, 
		RX_NONE, 

		RX_PSDATA_END, 
	};

protected:
	bool m_bInitialized;		//!< ς݃tO
	bool m_bUseOpenGL;			//!< VBOL/

	uint m_uNumParticles;		//!< ݂̃p[eBN
	uint m_uMaxParticles;		//!< őp[eBN

	RXREAL m_fParticleRadius;	//!< p[eBNa
	Vec3   m_v3Gravity;			//!< d͉x
	RXREAL m_fRestitution;		//!< W

	RXREAL *m_hPos;				//!< p[eBNʒu
	RXREAL *m_hVel;				//!< p[eBNx

	uint m_posVBO;				//!< p[eBNWVBO
	uint m_colorVBO;			//!< J[VBO

	Vec3 m_v3EnvMin;			//!< AABBŏW
	Vec3 m_v3EnvMax;			//!< AABBőW
	
	int m_iColorType;			//!< `掞̐F^Cv

protected:
	rxParticleSystemBase(){}

public:
	//! RXgN^
	rxParticleSystemBase(bool bUseOpenGL) : 
		m_bInitialized(false),
		m_bUseOpenGL(bUseOpenGL), 
		m_hPos(0),
		m_hVel(0)
	{
		m_v3Gravity = Vec3(0.0, -9.82, 0.0);
		m_fRestitution = 0.0;
		m_fParticleRadius = 0.1;
	}

	//! fXgN^
	virtual ~rxParticleSystemBase(){}


	// V~[Vԏ̎擾
	Vec3 GetMax(void) const { return m_v3EnvMax; }
	Vec3 GetMin(void) const { return m_v3EnvMin; }
	Vec3 GetDim(void) const { return m_v3EnvMax-m_v3EnvMin; }
	Vec3 GetCen(void) const { return 0.5*(m_v3EnvMax+m_v3EnvMin); }

	// p[eBN̎擾
	int	GetNumParticles() const { return m_uNumParticles; }
	int	GetMaxParticles() const { return m_uMaxParticles; }

	// p[eBNa̎擾
	float GetParticleRadius(){ return m_fParticleRadius; }

	// V~[Vݒ
	void SetGravity(RXREAL x){ m_v3Gravity = Vec3(0.0, x, 0.0); }	//!< d

	// p[eBNVBO̎擾
	unsigned int GetCurrentReadBuffer() const { return m_posVBO; }
	unsigned int GetColorBuffer()	   const { return m_colorVBO; }

	// `pJ[ݒ
	void SetColorType(int type){ m_iColorType = type; }
	int  GetColorType(void) const { return m_iColorType; }

public:
	//
	// z֐
	//
	virtual bool Update(RXREAL dt, int step = 0) = 0;

	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true, int num = -1) = 0;
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count) = 0;
	virtual void SetColorVBO(int type) = 0;

	virtual RXREAL* GetParticle(void) = 0;

public:
	//
	// z֐
	//
	// ő̃IuWFNgݒ
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, Vec3 vel, int flg){}
	virtual void SetSphereObstacle(Vec3 cen, double rad, Vec3 vel, int flg){}

	// p[eBNߖTTZɊi[
	virtual void SetParticlesToCell(void) = 0;
	virtual void SetParticlesToCell(RXREAL *prts, int n, RXREAL h) = 0;

	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z){ return 0.0; }
	virtual double GetImplicitAniso(double x, double y, double z){ return 0.0; }

	// `֐
	virtual void DrawCell(int i, int j, int k){}
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0){}
	virtual void DrawObstacles(void){}

public:
	// V[ݒ
	void Reset(rxParticleConfig config);
	bool Set(const vector<Vec3> &ppos, const vector<Vec3> &pvel);

	// t̂̒ǉ
	void AddSphere(int start, RXREAL *pos, RXREAL *vel, int r, RXREAL spacing);
	void AddBox(int start, Vec3 cen, Vec3 dim, Vec3 vel, RXREAL spacing);

	// `Fݒ
	RXREAL SetColorVBOFromArray(RXREAL *hVal, int d, bool use_max = true, RXREAL vmax = 1.0);
	void SetColorVBO(void){ SetColorVBO(m_iColorType); }

	// bV쐬pA֐擾
	static RXREAL GetImplicit_s(RXREAL x, RXREAL y, RXREAL z, void *ptr)
	{
		return ((rxParticleSystemBase*)ptr)->GetImplicit(x, y, z);
	}
	static RXREAL GetImplicitAniso_s(RXREAL x, RXREAL y, RXREAL z, void *ptr)
	{
		return ((rxParticleSystemBase*)ptr)->GetImplicitAniso(x, y, z);
	}


protected:
	//! VBO쐬
	uint createVBO(uint size)
	{
		GLuint vbo;
		glGenBuffers(1, &vbo);
		glBindBuffer(GL_ARRAY_BUFFER, vbo);
		glBufferData(GL_ARRAY_BUFFER, size, 0, GL_DYNAMIC_DRAW);
		glBindBuffer(GL_ARRAY_BUFFER, 0);
		return vbo;
	}
};



#endif	// _PS_H_

