/*!
  @file rx_sph_cpu.cpp
	
  @brief SPH@̎
 
  @author Makoto Fujisawa
  @date   2008-10,2011-06
*/
// FILE --rx_sph.cpp--

//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <GL/glew.h>
#include "rx_sph.h"


//-----------------------------------------------------------------------------
// rxSPHNX̎
//-----------------------------------------------------------------------------
/*!
 * RXgN^
 * @param[in] use_opengl VBOgptO
 */
rxSPH::rxSPH(bool use_opengl) : 
	rxParticleSystemBase(use_opengl), 
	m_hNrm(0), 
	m_hFrc(0),
	m_hDens(0), 
	m_hPres(0), 
	m_pBoundary(0)
{
	// p[^l̏
	m_v3Gravity = Vec3(0.0, -9.82, 0.0);
	m_fViscosity = 0.01;
	m_fGasStiffness = 3.0;
	m_fRestitution = 0.0;

	// ߖTTZ
	m_pNNGrid = new rxNNGrid(DIM);

	m_uNumParticles = 0;
	m_iColorType = RX_RAMP;
}

/*!
 * fXgN^
 */
rxSPH::~rxSPH()
{
	Finalize();
}


/*!
 * V~[V̏
 * @param[in] max_particles őp[eBN
 * @param[in] boundary_ext Ȇ傫(eӂ̒1/2)
 * @param[in] dens x
 * @param[in] mass p[eBN̎
 * @param[in] kernel_particle Lahȓ̃p[eBN
 */
void rxSPH::Initialize(const rxEnviroment &env)
{
	m_fRestDens        = env.dens;
	m_fMass            = env.mass;
	m_iKernelParticles = env.kernel_particles;

	RXREAL volume = m_iKernelParticles*m_fMass/m_fRestDens;

	// Lam_iKernelParticles̐p[eBNƂm_fRestDensƂȂ悤ɔavZ
	m_fEffectiveRadius = pow(((3.0*volume)/(4.0*RX_PI)), 1.0/3.0);
	m_fParticleRadius = pow((RX_PI/(6.0*m_iKernelParticles)), 1.0/3.0)*m_fEffectiveRadius;

	m_fKernelRadius = m_fEffectiveRadius;
	m_fViscosity = env.viscosity;
	m_fGasStiffness = env.gas_k;

	RXREAL h = m_fEffectiveRadius;
	RXREAL r = m_fParticleRadius;

	// J[l֐̒萔
	m_fAw = KernelCoefPoly6(h, 3, 1);
	m_fAg = KernelCoefSpiky(h, 3, 2);
	m_fAl = KernelCoefVisc(h, 3, 3);

	// J[l֐
	m_fpW  = KernelPoly6;
	m_fpGW = KernelSpikyG<Vec3>;
	m_fpLW = KernelViscL;

	// x̌vZ(ݒt@CŋLq̂ƑقȂ\)
	m_fRestDens = calRestDensity(h);


	//
	// Eݒ
	//
	m_pBoundary = new rxSolidBox(env.boundary_cen-env.boundary_ext, env.boundary_cen+env.boundary_ext, -1);

	// V~[V̑傫
	m_v3EnvMin = m_pBoundary->GetMin();
	m_v3EnvMax = m_pBoundary->GetMax();

	Vec3 world_size = m_v3EnvMax-m_v3EnvMin;
	Vec3 world_origin = m_v3EnvMin;
	
	double expansion = 0.01;
	world_origin -= 0.5*expansion*world_size;
	world_size *= (1.0+expansion); // V~[VŜ𕢂悤ɏ傫߂ɐݒ

	m_v3EnvMin = world_origin;
	m_v3EnvMax = world_origin+world_size;

	m_uNumParticles = 0;

	Allocate(env.max_particles);
}

/*!
 * ̊m
 *  - őp[eBNŊm
 * @param[in] max_particles őp[eBN
 */
void rxSPH::Allocate(int max_particles)
{
	assert(!m_bInitialized);

	m_uMaxParticles = max_particles;
	
	unsigned int size  = m_uMaxParticles*DIM;
	unsigned int size1 = m_uMaxParticles;
	unsigned int mem_size  = sizeof(RXREAL)*size;
	unsigned int mem_size1 = sizeof(RXREAL)*size1;

	//
	// m
	//
	m_hPos = new RXREAL[size];
	m_hVel = new RXREAL[size];
	m_hNrm = new RXREAL[size];
	m_hFrc = new RXREAL[size];
	memset(m_hPos, 0, mem_size);
	memset(m_hVel, 0, mem_size);
	memset(m_hNrm, 0, mem_size);
	memset(m_hFrc, 0, mem_size);

	m_hDens = new RXREAL[size1];
	m_hPres = new RXREAL[size1];
	memset(m_hDens, 0, mem_size1);
	memset(m_hPres, 0, mem_size1);

	// ߖTp[eBN
	m_vNeighs.resize(m_uMaxParticles);

	// `pVBO
	if(m_bUseOpenGL){
		m_posVBO = createVBO(mem_size);	
		m_colorVBO = createVBO(m_uMaxParticles*4*sizeof(RXREAL));

		SetColorVBO(RX_RAMP);
	}

	// Zݒ
	m_pNNGrid->Setup(m_v3EnvMin, m_v3EnvMax, m_fEffectiveRadius, m_uMaxParticles);
	m_vNeighs.resize(m_uMaxParticles);

	m_bInitialized = true;
}

/*!
 * mۂ̉
 */
void rxSPH::Finalize(void)
{
	assert(m_bInitialized);

	// 
	if(m_hPos) delete [] m_hPos;
	if(m_hVel) delete [] m_hVel;
	if(m_hNrm) delete [] m_hNrm;
	if(m_hFrc) delete [] m_hFrc;

	if(m_hDens) delete [] m_hDens;
	if(m_hPres) delete [] m_hPres;

	m_vNeighs.clear();

	if(m_bUseOpenGL){
		glDeleteBuffers(1, (const GLuint*)&m_posVBO);
		glDeleteBuffers(1, (const GLuint*)&m_colorVBO);
	}

	if(m_pNNGrid) delete m_pNNGrid;
	m_vNeighs.clear();

	if(m_pBoundary) delete m_pBoundary;

	int num_solid = (int)m_vSolids.size();
	for(int i = 0; i < num_solid; ++i){
		delete m_vSolids[i];
	}
	m_vSolids.clear();	

	m_uNumParticles = 0;
	m_uMaxParticles = 0;
}


/*!
 * SPH1Xebvi߂
 * @param[in] dt ԃXebv
 * @param[in] step ݂̃Xebv
 */
bool rxSPH::Update(RXREAL dt, int step)
{
	RXREAL h = m_fEffectiveRadius;

	// ߖTqT
	SetParticlesToCell();

	// xvZ
	calDensity(m_hPos, m_hDens, h);

	// p[eBNɂ(ۂɂ͉x)vZ
	calForce(m_hPos, m_hVel, m_hDens, m_hPres, m_hFrc, h);

	// ʒuEx̍XV
	integrate(m_hPos, m_hVel, m_hDens, m_hFrc, m_hPos, m_hVel, dt);

	// `pVBOֈʒuƐF̏]
	SetArrayVBO(RX_POSITION, m_hPos, 0, m_uNumParticles);
	SetColorVBO(m_iColorType);

	return true;
}

/*!
 * p[eBNf[^̎擾
 * @return p[eBNf[^̃foCX|C^
 */
RXREAL* rxSPH::GetParticle(void)
{
	return m_hPos;
}


/*!
 * J[lpVBO̕ҏW
 * @param[in] type F̃x[XƂ镨l
 */
void rxSPH::SetColorVBO(int type)
{
	switch(type){
	case RX_DENSITY:
		SetColorVBOFromArray(m_hDens, 1, false, m_fRestDens*2.0);
		break;

	case RX_PRESSURE:
		SetColorVBOFromArray(m_hPres, 1);
		break;

	case RX_CONSTANT:
		if(m_bUseOpenGL){
			// J[obt@ɒlݒ
			glBindBufferARB(GL_ARRAY_BUFFER, m_colorVBO);
			RXREAL *data = (RXREAL*)glMapBufferARB(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
			RXREAL *ptr = data;
			for(uint i = 0; i < m_uNumParticles; ++i){
				RXREAL t = i/(RXREAL)m_uNumParticles;
				*ptr++ = 0.15f;
				*ptr++ = 0.15f;
				*ptr++ = 0.95f;
				*ptr++ = 1.0f;
			}
			glUnmapBufferARB(GL_ARRAY_BUFFER);
		}
		break;

	case RX_RAMP:
		if(m_bUseOpenGL){
			// J[obt@ɒlݒ
			glBindBufferARB(GL_ARRAY_BUFFER, m_colorVBO);
			RXREAL *data = (RXREAL*)glMapBufferARB(GL_ARRAY_BUFFER, GL_WRITE_ONLY);
			RXREAL *ptr = data;
			for(uint i = 0; i < m_uNumParticles; ++i){
				RXREAL t = i/(RXREAL)m_uNumParticles;
#if 0
				*ptr++ = rand()/(RXREAL)RAND_MAX;
				*ptr++ = rand()/(RXREAL)RAND_MAX;
				*ptr++ = rand()/(RXREAL)RAND_MAX;
#else
				RX_COLOR_RAMP(t, ptr);
				ptr += 3;
#endif
				*ptr++ = 1.0f;
			}
			glUnmapBufferARB(GL_ARRAY_BUFFER);
		}
		break;

	default:
		break;
	}

}



/*!
 * xvZ
 * @param[in] ppos p[eBNSʒuz
 * @param[out] pdens p[eBNxi[z
 * @param[in] h La
 */
void rxSPH::calDensity(const RXREAL *ppos, RXREAL *pdens, RXREAL h)
{
	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0;
		pos0[0] = ppos[DIM*i+0];
		pos0[1] = ppos[DIM*i+1];
		pos0[2] = ppos[DIM*i+2];

		pdens[i] = 0.0;

		// ߖTq
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0) continue;

			Vec3 pos1;
			pos1[0] = ppos[DIM*j+0];
			pos1[1] = ppos[DIM*j+1];
			pos1[2] = ppos[DIM*j+2];

			RXREAL r = sqrt(itr->Dist2);

			pdens[i] += m_fMass*m_fpW(r, h, m_fAw);
		}
	}
}


/*!
 * ͍CŠvZ
 * @param[in] ppos p[eBNSʒuz
 * @param[in] pvel p[eBNxz
 * @param[in] pdens p[eBNxz
 * @param[out] ppres p[eBN͂i[z(`,fobOp)
 * @param[out] pfrc p[eBNɂ(x)i[z
 * @param[in] h La
 */
void rxSPH::calForce(const RXREAL *ppos, const RXREAL *pvel, const RXREAL *pdens, RXREAL *ppres, RXREAL *pfrc, RXREAL h)
{
	RXREAL r0 = m_fRestDens;
	
	for(uint i = 0; i < m_uNumParticles; ++i){
		pfrc[4*i+0] = 0.0;
		pfrc[4*i+1] = 0.0;
		pfrc[4*i+2] = 0.0;
		pfrc[4*i+3] = 0.0;

		ppres[i] = m_fGasStiffness*(pdens[i]-r0);
	}


	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0, vel0;
		pos0 = Vec3(ppos[4*i+0], ppos[4*i+1], ppos[4*i+2]);
		vel0 = Vec3(pvel[4*i+0], pvel[4*i+1], pvel[4*i+2]);

		RXREAL prsi = ppres[i]/(pdens[i]*pdens[i]);

		Vec3 Fp(0.0), Fev(0.0);
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0 || i == j) continue;

			Vec3 pos1, vel1;
			pos1 = Vec3(ppos[4*j+0], ppos[4*j+1], ppos[4*j+2]);
			vel1 = Vec3(pvel[4*j+0], pvel[4*j+1], pvel[4*j+2]);

			Vec3 rij = pos0-pos1;
			Vec3 vji = vel1-vel0;

			RXREAL r = norm(rij);

			RXREAL prsj = ppres[j]/(pdens[j]*pdens[j]);

			// 
			Fp += m_fMass*(prsi+prsj)*m_fpGW(r, h, m_fAg, rij);
			//Fp += m_fMass*(ppres[j]/(pdens[i]*pdens[j]))*m_fpGW(r, h, m_fAg, rij);

			// S
			Fev += m_fMass*(vji/pdens[j])*m_fpLW(r, h, m_fAl, 3);
		}

		Vec3 force(0.0);
		force += -Fp;
		force += m_fViscosity*Fev;

		// O͍
		force += m_v3Gravity;

		pfrc[4*i+0] += force[0];
		pfrc[4*i+1] += force[1];
		pfrc[4*i+2] += force[2];
	}
}


/*!
 * rest density̌vZ
 *  - ߖTɃp[eBN~l߂ĂƂĖxvZ
 * @param[in] h La
 * @return rest density
 */
RXREAL rxSPH::calRestDensity(RXREAL h)
{
	RXREAL r0 = 0.0;
	RXREAL l = 2*GetParticleRadius();
	int n = (int)ceil(m_fKernelRadius/l)+1;
	for(int x = -n; x <= n; ++x){
		for(int y = -n; y <= n; ++y){
			for(int z = -n; z <= n; ++z){
				Vec3 rij = Vec3(x*l, y*l, z*l);
				r0 += m_fMass*m_fpW(norm(rij), h, m_fAw);
			}
		}
	}
	return r0;
}


/*!
 * ő̃IuWFNgƂ̏Փ˔CՓˉ
 *  - Ep[eBNgĂꍇ͓˂h~p̈ʒû
 * @param[in] pos0 OXebv̈ʒu
 * @param[inout] pos1 Vʒu
 * @param[inout] vel x
 * @param[in] dt ^CXebv
 * @return Փ˃IuWFNg̐
 */
int rxSPH::calCollisionSolid(Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt)
{
	int c = 0;
	rxCollisionInfo coli;

	// ő̃IuWFNgƂ̏Փˏ
	for(vector<rxSolid*>::iterator i = m_vSolids.begin(); i != m_vSolids.end(); ++i){
		if((*i)->GetDistanceR(pos1, m_fParticleRadius, coli)){
			RXREAL res = m_fRestitution;
			res = (res > 0) ? (res*fabs(coli.Penetration())/(dt*norm(vel))) : 0.0f;
			vel -= (1+res)*dot(vel, coli.Normal())*coli.Normal();
			pos1 = coli.Contact();
		}
	}

	// V~[VԋEƂ̏Փˏ
	if(m_pBoundary->GetDistanceR(pos1, m_fParticleRadius, coli)){
		RXREAL res = m_fRestitution;
		res = (res > 0) ? (res*fabs(coli.Penetration())/(dt*norm(vel))) : 0.0f;
		vel -= (1+res)*dot(vel, coli.Normal())*coli.Normal();
		pos1 = coli.Contact();
	}

	return c;
}

/*!
 * ʒuEx̍XV
 * @param[in] pos p[eBNSʒuz
 * @param[in] vel p[eBNxz
 * @param[in] acc p[eBNxz
 * @param[out] pos_new XVp[eBNʒui[z
 * @param[out] vel_new XVp[eBNxi[z
 * @param[in] dt ^CXebv
 */
void rxSPH::integrate(const RXREAL *pos, const RXREAL *vel, const RXREAL *dens, const RXREAL *acc, 
						  RXREAL *pos_new, RXREAL *vel_new, RXREAL dt)
{
	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 x, x_old, v, a, v_old;
		for(int k = 0; k < 3; ++k){
			x[k] = pos[DIM*i+k];
			v[k] = vel[DIM*i+k];
			a[k] = acc[DIM*i+k];
		}
		x_old = x;

		// Vxƈʒu
		v += dt*a;
		x += dt*v;

		// E/̂Ƃ̏Փ˔
		calCollisionSolid(x_old, x, v, dt);

		// VxƈʒuōXV
		for(int k = 0; k < 3; ++k){
			pos_new[DIM*i+k] = x[k];
			vel_new[DIM*i+k] = v[k];
		}

	}
}


//-----------------------------------------------------------------------------
// @vZ
//-----------------------------------------------------------------------------
/*!
 * @vZ
 *  - z쏃C"q@ɂ闬̃V~[V̍i_O"CÉwƘ_C2007.
 *  - p.24, 3.4.3 @̏C  (3.8) 
 *  - (3.8) ł w(v-r)(v-r) ƂȂĂ邪 w(v-r) ̊ԈႢ
 */
void rxSPH::CalNormal(void)
{
	RXREAL h = m_fEffectiveRadius;

	// ߖTqT
	SetParticlesToCell();

	for(uint i = 0; i < m_uNumParticles; ++i){
		Vec3 pos0 = Vec3(m_hPos[DIM*i+0], m_hPos[DIM*i+1], m_hPos[DIM*i+2]);
		Vec3 nrm(0.0);

		// ߖTq
		for(vector<rxNeigh>::iterator itr = m_vNeighs[i].begin() ; itr != m_vNeighs[i].end(); ++itr){
			int j = itr->Idx;
			if(j < 0 || i == j) continue;

			Vec3 pos1 = Vec3(m_hPos[DIM*j+0], m_hPos[DIM*j+1], m_hPos[DIM*j+2]);
			Vec3 rij = pos0-pos1;

			RXREAL rr = sqrt(itr->Dist2)/h;

			RXREAL w = (rr < 1.0 ? 1.0/rr-1.0 : 0.0);

			nrm += w*rij;
		}

		m_hNrm[DIM*i+0] = nrm[0];
		m_hNrm[DIM*i+1] = nrm[1];
	}
}



//-----------------------------------------------------------------------------
// ߖTT
//-----------------------------------------------------------------------------
/*!
 * ߖTqT
 * @param[in] idx TSp[eBNCfbNX
 * @param[in] prts p[eBNʒu
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighbors(int idx, RXREAL *prts, vector<rxNeigh> &neighs, RXREAL h)
{
	if(idx < 0 || idx >= (int)m_uNumParticles) return;

	Vec3 pos;
	pos[0] = prts[DIM*idx+0];
	pos[1] = prts[DIM*idx+1];
	pos[2] = prts[DIM*idx+2];

	if(h < 0.0) h = m_fEffectiveRadius;

	m_pNNGrid->GetNN(pos, prts, m_uNumParticles, neighs, h);
}

/*!
 * ߖTqT
 * @param[in] idx TSp[eBNCfbNX
 * @param[out] neighs Tʊi[ߖTRei
 * @param[in] h La
 */
void rxSPH::GetNearestNeighbors(Vec3 pos, vector<rxNeigh> &neighs, RXREAL h)
{
	if(h < 0.0) h = m_fEffectiveRadius;

	m_pNNGrid->GetNN(pos, m_hPos, m_uNumParticles, neighs, h);
}

/*!
 * Sp[eBN𕪊ZɊi[
 */
void rxSPH::SetParticlesToCell(RXREAL *prts, int n, RXREAL h)
{
	// Zɗqo^
	m_pNNGrid->SetObjectToCell(prts, n);

	// ߖTqT
	if(h < 0.0) h = m_fEffectiveRadius;
	for(int i = 0; i < (int)m_uNumParticles; i++){
		m_vNeighs[i].clear();
		GetNearestNeighbors(i, prts, m_vNeighs[i], h);
	}
}
void rxSPH::SetParticlesToCell(void)
{
	SetParticlesToCell(m_hPos, m_uNumParticles, m_fEffectiveRadius);
}




/*!
 * TpZ̕`
 * @param[in] i,j,k Obh̃CfbNX
 */
void rxSPH::DrawCell(int i, int j, int k)
{
	if(m_pNNGrid) m_pNNGrid->DrawCell(i, j, k);
}

/*!
 * TpObh̕`
 * @param[in] col p[eBN܂܂Z̐F
 * @param[in] col2 |S܂܂Z̐F
 * @param[in] sel _ɑIꂽẐݕ`(1ŐVZIC2łłɑIĂZ`C0łׂẴZ`)
 */
void rxSPH::DrawCells(Vec3 col, Vec3 col2, int sel)
{
	if(m_pNNGrid) m_pNNGrid->DrawCells(col, col2, sel, m_hPos);

}

/*!
 * ő̏Q̕`
 */
void rxSPH::DrawObstacles(void)
{
	for(vector<rxSolid*>::iterator i = m_vSolids.begin(); i != m_vSolids.end(); ++i){
		(*i)->Draw();
	}
}




/*!
 * {bNX^Q
 * @param[in] cen {bNXSW
 * @param[in] ext {bNX̑傫(ӂ̒1/2)
 * @param[in] ang {bNX̊px(IC[p)
 * @param[in] flg L/tO
 */
void rxSPH::SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, Vec3 vel, int flg)
{
	rxSolidBox *box = new rxSolidBox(cen-ext, cen+ext, 1);
	double m[16];
	EulerToMatrix(m, ang[0], ang[1], ang[2]);
	box->SetMatrix(m);

	m_vSolids.push_back(box);

}

/*!
 * ^Q
 * @param[in] cen ̒SW
 * @param[in] rad ̂̔a
 * @param[in] flg L/tO
 */
void rxSPH::SetSphereObstacle(Vec3 cen, double rad, Vec3 vel, int flg)
{
	rxSolidSphere *sphere = new rxSolidSphere(cen, rad, 1);
	m_vSolids.push_back(sphere);
}


/*!
 * VBOzXgփf[^]C擾
 * @param[in] type f[^̎
 * @return zXg̃f[^
 */
RXREAL* rxSPH::GetArrayVBO(rxParticleArray type, bool d2h, int num)
{
	assert(m_bInitialized);
 
	if(num == -1) num = m_uNumParticles;
	RXREAL* hdata = 0;

	unsigned int vbo = 0;

	switch(type){
	case RX_POSITION:
		hdata = m_hPos;
		vbo = m_posVBO;
		break;

	case RX_VELOCITY:
		hdata = m_hVel;
		break;

	case RX_DENSITY:
		hdata = m_hDens;
		break;

	case RX_PRESSURE:
		hdata = m_hPres;
		break;

	case RX_NORMAL:
		hdata = m_hNrm;
		break;

	case RX_FORCE:
		hdata = m_hFrc;
		break;

	default:
		break;
	}

	return hdata;
}

/*!
 * zXgVBOփf[^]
 * @param[in] type f[^̎
 * @param[in] data zXg̃f[^
 * @param[in] start f[^̊JnCfbNX
 * @param[in] count ǉ
 */
void rxSPH::SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count)
{
	assert(m_bInitialized);
 
	switch(type){
	case RX_POSITION:
		{
			if(m_bUseOpenGL){
				glBindBuffer(GL_ARRAY_BUFFER, m_posVBO);
				glBufferSubData(GL_ARRAY_BUFFER, start*4*sizeof(RXREAL), count*4*sizeof(RXREAL), data);
				glBindBuffer(GL_ARRAY_BUFFER, 0);
			}
		}
		break;

	default:
		break;
	}	   
}




//-----------------------------------------------------------------------------
// A֐l
//-----------------------------------------------------------------------------
/*!
 * ^{[ɂA֐̌vZ
 *  - xvẐ܂܎gp
 * @param[in] x,y,z Wl
 * @return A֐l
 */
double rxSPH::GetImplicit(double x, double y, double z)
{
	RXREAL c = 0.0;
	Vec3 pos(x, y, z);

	if(pos[0] < m_v3EnvMin[0]) return c;
	if(pos[0] > m_v3EnvMax[0]) return c;
	if(pos[1] < m_v3EnvMin[1]) return c;
	if(pos[1] > m_v3EnvMax[1]) return c;
	if(pos[2] < m_v3EnvMin[2]) return c;
	if(pos[2] > m_v3EnvMax[2]) return c;

	RXREAL h = m_fEffectiveRadius;

	vector<rxNeigh> ne;
	m_pNNGrid->GetNN(pos, m_hPos, m_uNumParticles, ne, h);

	// ߖTq
	for(vector<rxNeigh>::iterator itr = ne.begin(); itr != ne.end(); ++itr){
		int j = itr->Idx;
		if(j < 0) continue;

		Vec3 pos1;
		pos1[0] = m_hPos[DIM*j+0];
		pos1[1] = m_hPos[DIM*j+1];
		pos1[2] = m_hPos[DIM*j+2];

		RXREAL r = sqrt(itr->Dist2);

		c += m_fMass*m_fpW(r, h, m_fAw);
	}

	return c;
}


