/*!
  @file rx_sph.h
	
  @brief SPH@
 
  @author Makoto Fujisawa
  @date 2008-10,2011-06
*/

#ifndef _RX_SPH_H_
#define _RX_SPH_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_ps.h"			// p[eBNVXeNX
#include "rx_nnsearch.h"	// ObhɂߖTT

#include "rx_sph_solid.h"	// ő̃IuWFNgNX
#include "rx_kernel.h"		// J[l(d݊֐)


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------

//! SPHV[̃p[^
struct rxEnviroment
{
	int max_particles;			//!< őp[eBN
	Vec3 boundary_cen;			//!< E̒S
	Vec3 boundary_ext;			//!< Ȇ傫(eӂ̒1/2)
	RXREAL dens;				//!< x
	RXREAL mass;				//!< p[eBN̎
	RXREAL kernel_particles;	//!< Lahȓ̃p[eBN
	RXREAL dt;					//!< ԃXebv
	RXREAL viscosity;			//!< SW
	RXREAL gas_k;				//!< KX萔

	// \ʃbV
	int mesh_max_n;				//!< MC@pObh̍ő啪
	RXREAL mesh_thr;			//!< MC@臒l

	rxEnviroment()
	{
		max_particles = 50000;
		boundary_cen = Vec3(0.0);
		boundary_ext = Vec3(1.0, 1.0, 1.0);
		dens = (RXREAL)998.29;
		mass = (RXREAL)0.04;
		kernel_particles = (RXREAL)20.0;
		dt = 0.01;
		viscosity = 1.0e-3;
		gas_k = 3.0;

		mesh_max_n = 128;
		mesh_thr = 700.0;
	}
};



//-----------------------------------------------------------------------------
// MARK:rxSPHNX̐錾
//-----------------------------------------------------------------------------
class rxSPH : public rxParticleSystemBase
{
private:
	// p[eBN
	RXREAL *m_hNrm;					//!< p[eBN@
	RXREAL *m_hFrc;					//!< p[eBNɂ
	RXREAL *m_hDens;				//!< p[eBNx
	RXREAL *m_hPres;				//!< p[eBN

	// EEő
	rxSolid *m_pBoundary;			//!< V~[VԂ̋E
	vector<rxSolid*> m_vSolids;		//!< ő̕

	// ԕiq֘A
	rxNNGrid *m_pNNGrid;			//!< ObhɂߖTT
	vector< vector<rxNeigh> > m_vNeighs;	//!< ߖTp[eBN

	// qp[^
	uint m_iKernelParticles;		//!< J[l̃p[eBN
	RXREAL m_fRestDens, m_fMass;	//!< xC
	RXREAL m_fEffectiveRadius;		//!< La
	RXREAL m_fKernelRadius;			//!< J[l̉e͈

	// V~[Vp[^
	RXREAL m_fGasStiffness;			//!< KX萔
	RXREAL m_fViscosity;			//!< SW

	// J[l֐̌vZ̍ۂɗp萔W
	double m_fAw;
	double m_fAg;
	double m_fAl;

	// J[l֐
	double (*m_fpW)(double, double, double);
	Vec3 (*m_fpGW)(double, double, double, Vec3);
	double (*m_fpLW)(double, double, double, double);

protected:
	rxSPH(){}

public:
	//! RXgN^
	rxSPH(bool use_opengl);

	//! fXgN^
	virtual ~rxSPH();

	// p[eBNa
	float GetEffectiveRadius(){ return m_fEffectiveRadius; }

	// ߖTp[eBN
	uint* GetNeighborList(const int &i, int &n);

public:
	//
	// z֐
	//
	// p[eBNf[^
	virtual RXREAL* GetParticle(void);

	// V~[VXebv
	virtual bool Update(RXREAL dt, int step = 0);

	// V[̐ݒ
	virtual void SetBoxObstacle(Vec3 cen, Vec3 ext, Vec3 ang, Vec3 vel, int flg);
	virtual void SetSphereObstacle(Vec3 cen, double rad, Vec3 vel, int flg);

	// zXg<->VBOԓ]
	virtual RXREAL* GetArrayVBO(rxParticleArray type, bool d2h = true, int num = -1);
	virtual void SetArrayVBO(rxParticleArray type, const RXREAL* data, int start, int count);
	virtual void SetColorVBO(int type);

	// A֐lvZ
	virtual double GetImplicit(double x, double y, double z);

	// `֐
	virtual void DrawCell(int i, int j, int k);
	virtual void DrawCells(Vec3 col, Vec3 col2, int sel = 0);
	virtual void DrawObstacles(void);


public:
	// SPH
	void Initialize(const rxEnviroment &env);
	void Allocate(int max_particles);
	void Finalize(void);

	// ߖT擾
	void GetNearestNeighbors(Vec3 pos, vector<rxNeigh> &neighs, RXREAL h = -1.0);
	void GetNearestNeighbors(int idx, RXREAL *p, vector<rxNeigh> &neighs, RXREAL h = -1.0);

	// ZɃp[eBNi[
	virtual void SetParticlesToCell(void);
	virtual void SetParticlesToCell(RXREAL *prts, int n, RXREAL h);

	// @vZ
	void CalNormal(void);



protected:
	// CPUɂSPHvZ
	void calDensity(const RXREAL *pos, RXREAL *dens, RXREAL h);
	void calForce(const RXREAL *ppos, const RXREAL *pvel, const RXREAL *pdens, RXREAL *ppres, RXREAL *pfrc, RXREAL h);

	// ʒuƑx̍XV
	void integrate(const RXREAL *pos, const RXREAL *vel, const RXREAL *dens, const RXREAL *acc, 
				   RXREAL *pos_new, RXREAL *vel_new, RXREAL dt);

	// rest density̌vZ
	RXREAL calRestDensity(RXREAL h);

	// Փ˔
	int calCollisionSolid(Vec3 &pos0, Vec3 &pos1, Vec3 &vel, RXREAL dt);
};



#endif	// _SPH_H_

