/*!
  @file rx_fluid_solver.cpp
	
  @brief _V~[V
 
  @author Makoto Fujisawa
  @date 2013-10
*/


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_fluid_solver.h"
#include "rx_numericalf.h"

#include <GL/glut.h>


//-----------------------------------------------------------------------------
// rxFluidSolver3DNX̎
//-----------------------------------------------------------------------------
/*!
 * RXgN^
 */
rxFluidSolver3D::rxFluidSolver3D(rxEnviroment scn)
{
	m_iNx = scn.Nx-2;
	m_iNy = scn.Ny-2;
	m_iNz = scn.Nz-2;

	m_fH = scn.H;
	m_fDt = scn.Dt;

	m_Par = scn;

	m_iStep = 0;
	m_iMaxStep = 1000000;

	m_bUseVC = true;

	Allocate();
	Clear();
}

/*!
 * fXgN^
 */
rxFluidSolver3D::~rxFluidSolver3D()
{
	Finalize();
}

/*!
 * z̈m
 */
int rxFluidSolver3D::Allocate(void)
{
	int size = (m_iNx+2)*(m_iNy+2)*(m_iNz+2);

	m_pU = RX_ALLOC<RXREAL>((m_iNx+3)*(m_iNy+2)*(m_iNz+2));
	m_pV = RX_ALLOC<RXREAL>((m_iNx+2)*(m_iNy+3)*(m_iNz+2));
	m_pW = RX_ALLOC<RXREAL>((m_iNx+2)*(m_iNy+2)*(m_iNz+3));

	m_pC = RX_ALLOC<RXREAL>(size);

	m_pT = RX_ALLOC<RXREAL>(size);
	m_pP = RX_ALLOC<RXREAL>(size);

	m_pQv = RX_ALLOC<RXREAL>(size);
	m_pQc = RX_ALLOC<RXREAL>(size);

	m_pX = RX_ALLOC<RXREAL>((m_iNx+3)*(m_iNy+3)*(m_iNz+3));

	m_pPois  = RX_ALLOC2D<RXREAL>(size, 7);
	m_pIC    = RX_ALLOC2D<RXREAL>(size, 7);
	m_pNeigh = RX_ALLOC2D<int>(size, 7);

	return 1;
}

/*!
 * f[^
 */
void rxFluidSolver3D::Finalize(void)
{
	int size = (m_iNx+2)*(m_iNy+2)*(m_iNz+2);

	RX_FREE(m_pU);
	RX_FREE(m_pV);
	RX_FREE(m_pW);

	RX_FREE(m_pC);
	RX_FREE(m_pT);
	RX_FREE(m_pP);

	RX_FREE(m_pQv);
	RX_FREE(m_pQc);

	RX_FREE(m_pX);

	RX_FREE2D(m_pPois, size);
	RX_FREE2D(m_pIC, size);
	RX_FREE2D(m_pNeigh, size);
}

/*!
 * f[^
 */
void rxFluidSolver3D::Clear(void)
{
	int size = (m_iNx+2)*(m_iNy+2)*(m_iNz+2);

	for(int i = 0; i < size; ++i){
		m_pC[i] = 0.0;
		m_pT[i] = m_Par.Tamb;
		m_pP[i] = 0.0;

		m_pQv[i] = 0.0;
		m_pQc[i] = 0.0;

		for(int j = 0; j < 7; ++j){
			m_pPois[i][j] = 0.0;
			m_pIC[i][j] = 0.0;
		}
	}

	RXFOR3(0, m_iNx+2, 0, m_iNy+2, 0, m_iNz+2){
		RXREAL h = j/(RXREAL)(m_iNy+2);
		m_pT[ix(i, j, k, m_iNx+2, m_iNy+2)] = (1.0-h)*m_Par.Tmax+h*m_Par.Tamb;
	}

	for(int i = 0; i < (m_iNx+3)*(m_iNy+2)*(m_iNz+2); ++i) m_pU[i] = 0.0;
	for(int i = 0; i < (m_iNx+2)*(m_iNy+3)*(m_iNz+2); ++i) m_pV[i] = 0.0;
	for(int i = 0; i < (m_iNx+2)*(m_iNy+2)*(m_iNz+3); ++i) m_pW[i] = 0.0;
	for(int i = 0; i < (m_iNx+3)*(m_iNy+3)*(m_iNz+3); ++i) m_pX[i] = 0.0;
}


/*!
 * ǖʋE
 * @param[in] b E(0:XJ,1:xx,2:yx,3:zx)
 * @param[in] x EK
 */
void rxFluidSolver3D::setBnd(int b, RXREAL* x, int nx, int ny, int nz)
{
	RXFOR2(0, nz, 0, nx){
		x[ix(j,    0, i, nx, ny)] = b==2 ? -x[ix(j,    2, i, nx, ny)] : x[ix(j,    1, i, nx, ny)];
		x[ix(j, ny-1, i, nx, ny)] = b==2 ? x[ix(j, ny-3, i, nx, ny)] : x[ix(j, ny-2, i, nx, ny)];
		if(b == 2){ x[ix(j, 1, i, nx, ny)] = 0.0; x[ix(j, ny-2, i, nx, ny)] = 0.0; }
	}
	RXFOR2(0, nx, 0, ny){
		x[ix(i, j,    0, nx, ny)] = b==3 ? x[ix(i, j,    2, nx, ny)] : x[ix(i, j,    1, nx, ny)];
		x[ix(i, j, nz-1, nx, ny)] = b==3 ? x[ix(i, j, nz-3, nx, ny)] : x[ix(i, j, nz-2, nx, ny)];
		if(b == 3) x[ix(i, j, 1, nx, ny)] = x[ix(i, j, nz-2, nx, ny)] = 0.0;
	}
	RXFOR2(0, ny, 0, nz){
		x[ix(   0, i, j, nx, ny)] = b==1 ? x[ix(   2, i, j, nx, ny)] : x[ix(   1, i, j, nx, ny)];
		x[ix(nx-1, i, j, nx, ny)] = b==1 ? x[ix(nx-3, i, j, nx, ny)] : x[ix(nx-2, i, j, nx, ny)];
		if(b == 1) x[ix(1, i, j, nx, ny)] = x[ix(nx-2, i, j, nx, ny)] = 0.0;
	}
}
/*!
 * ǖʋE
 * @param[in] b E(0:XJ,1:xx,2:yx)
 * @param[in] x EK
 */
void rxFluidSolver3D::setBnd(RXREAL* x, int nx, int ny, int nz, RXREAL val)
{
	RXFOR2(0, nx, 0, ny){
		x[ix(i, j,    0, nx, ny)] = val;
		x[ix(i, j, nz-1, nx, ny)] = val;
	}
	RXFOR2(0, ny, 0, nz){
		x[ix(   0, i, j, nx, ny)] = val;
		x[ix(nx-1, i, j, nx, ny)] = val;
	}
	RXFOR2(0, nz, 0, nx){
		x[ix(j,    0, i, nx, ny)] = val;
		x[ix(j, ny-1, i, nx, ny)] = val;
	}
}


/*!
 * Qx̌vZ
 * @param[in] i,j,k Obhʒu
 * @return QxxNg
 */
Vec3 rxFluidSolver3D::calCurl(int i, int j, int k)
{
	RXREAL w0, w1, v0, v1, u0, u1;

	u0 = 0.5*(m_pU[ix(i, j-1, k, m_iNx+3, m_iNy+2)]+m_pU[ix(i+1, j-1, k, m_iNx+3, m_iNy+2)]);
	u1 = 0.5*(m_pU[ix(i, j+1, k, m_iNx+3, m_iNy+2)]+m_pU[ix(i+1, j+1, k, m_iNx+3, m_iNy+2)]);
	RXREAL dudy = 0.5*(u1-u0);
	u0 = 0.5*(m_pU[ix(i, j, k-1, m_iNx+3, m_iNy+2)]+m_pU[ix(i+1, j, k-1, m_iNx+3, m_iNy+2)]);
	u1 = 0.5*(m_pU[ix(i, j, k+1, m_iNx+3, m_iNy+2)]+m_pU[ix(i+1, j, k+1, m_iNx+3, m_iNy+2)]);
	RXREAL dudz = 0.5*(u1-u0);

	v0 = 0.5*(m_pV[ix(i-1, j, k, m_iNx+2, m_iNy+3)]+m_pV[ix(i-1, j+1, k, m_iNx+2, m_iNy+3)]);
	v1 = 0.5*(m_pV[ix(i+1, j, k, m_iNx+2, m_iNy+3)]+m_pV[ix(i+1, j+1, k, m_iNx+2, m_iNy+3)]);
	RXREAL dvdx = 0.5*(u1-u0);
	v0 = 0.5*(m_pV[ix(i, j, k-1, m_iNx+2, m_iNy+3)]+m_pV[ix(i, j+1, k-1, m_iNx+2, m_iNy+3)]);
	v1 = 0.5*(m_pV[ix(i, j, k+1, m_iNx+2, m_iNy+3)]+m_pV[ix(i, j+1, k+1, m_iNx+2, m_iNy+3)]);
	RXREAL dvdz = 0.5*(u1-u0);

	w0 = 0.5*(m_pW[ix(i, j-1, k, m_iNx+2, m_iNy+2)]+m_pW[ix(i, j-1, k+1, m_iNx+2, m_iNy+2)]);
	w1 = 0.5*(m_pW[ix(i, j+1, k, m_iNx+2, m_iNy+2)]+m_pW[ix(i, j+1, k+1, m_iNx+2, m_iNy+2)]);
	RXREAL dwdy = 0.5*(w1-w0);
	w0 = 0.5*(m_pW[ix(i-1, j, k, m_iNx+2, m_iNy+2)]+m_pW[ix(i-1, j, k+1, m_iNx+2, m_iNy+2)]);
	w1 = 0.5*(m_pW[ix(i+1, j, k, m_iNx+2, m_iNy+2)]+m_pW[ix(i+1, j, k+1, m_iNx+2, m_iNy+2)]);
	RXREAL dwdx = 0.5*(w1-w0);

	return Vec3(dwdy-dvdz, dudz-dwdx, dvdx-dudy);
}


/*!
 * Vorticity ConfinemenťvZ
 * @param[out] u,v Fconf
 */
void rxFluidSolver3D::vorticityConfinement(RXREAL dt)
{
	RXREAL e = m_Par.VC;

	// Qx(=ށ~v)̌vZ
	RXFOR3(1, m_iNx+1, 1, m_iNy+1, 1, m_iNz+1){
		m_pC[ix(i, j, k, m_iNx+2, m_iNy+2)] = (RXREAL)(norm(calCurl(i, j, k)));
	}

	setBnd(0, m_pC, m_iNx+2, m_iNy+2, m_iNz+2);

	RXFOR3(1, m_iNx+1, 1, m_iNy+1, 1, m_iNz+1){
		// ||
		RXREAL dcdx = 0.5*(m_pC[ix(i+1, j, k, m_iNx+2, m_iNy+2)]-m_pC[ix(i-1, j, k, m_iNx+2, m_iNy+2)])/m_fH;
		RXREAL dcdy = 0.5*(m_pC[ix(i, j+1, k, m_iNx+2, m_iNy+2)]-m_pC[ix(i, j-1, k, m_iNx+2, m_iNy+2)])/m_fH;
		RXREAL dcdz = 0.5*(m_pC[ix(i, j, k+1, m_iNx+2, m_iNy+2)]-m_pC[ix(i, j, k-1, m_iNx+2, m_iNy+2)])/m_fH;

		// |n|
		RXREAL len = sqrt(dcdx*dcdx+dcdy*dcdy+dcdz*dcdz)+0.000001;

		// N = ( n/|n| )
		RXREAL nx = dcdx/len;
		RXREAL ny = dcdy/len;
		RXREAL nz = dcdz/len;

		Vec3 omega = calCurl(i, j, k);

		// N x w
		RXREAL nwx = ny*omega[2]-nz*omega[1];
		RXREAL nwy = nz*omega[0]-nx*omega[2];
		RXREAL nwz = nx*omega[1]-ny*omega[0];

		m_pU[ix(i,  j,  k, m_iNx+3, m_iNy+2)] += 0.5*e*m_fH*nwx;
		m_pV[ix(i,  j,  k, m_iNx+2, m_iNy+3)] += 0.5*e*m_fH*nwy;
		m_pW[ix(i,  j,  k, m_iNx+2, m_iNy+2)] += 0.5*e*m_fH*nwz;
		m_pU[ix(i+1, j, k, m_iNx+3, m_iNy+2)] += 0.5*e*m_fH*nwx;
		m_pV[ix(i, j+1, k, m_iNx+2, m_iNy+3)] += 0.5*e*m_fH*nwy;
		m_pW[ix(i, j, k+1, m_iNx+2, m_iNy+2)] += 0.5*e*m_fH*nwz;
	}
}


/*!
 * d͂ƕ͂̒ǉ
 * @param[in] x ǉ
 * @param[in] dt ԃXebv
 */
void rxFluidSolver3D::addGravity(RXREAL dt)
{
	// yx
	RXFOR3(0, m_iNx+2, 1, m_iNy+2, 0, m_iNz+2){
		RXREAL T = 0.5*(m_pT[ix(i, j-1, k, m_iNx+2, m_iNy+2)]+m_pT[ix(i, j, k, m_iNx+2, m_iNy+2)]);

		RXREAL h = j/(RXREAL)(m_iNy+2);
		RXREAL Tamb = (1.0-h)*m_Par.Tmax+h*m_Par.Tamb;

		m_pV[ix(i, j, k, m_iNx+2, m_iNy+3)] += -dt*m_Par.Gravity+m_Par.Beta*(T-Tamb)/Tamb;
	}

	setBnd(1, m_pU, m_iNx+3, m_iNy+2, m_iNz+2);
	setBnd(2, m_pV, m_iNx+2, m_iNy+3, m_iNz+2);
	setBnd(3, m_pW, m_iNx+2, m_iNy+2, m_iNz+3);
}

/*!
 * O
 */
void rxFluidSolver3D::addExternalForce(RXREAL dt)
{
	// C̋
	vector<Vec3> c;
	c.push_back(Vec3(0.4,  0.5,  0.2));	// ~󋟋̒SWƔa
	c.push_back(Vec3(0.65, 0.5, 0.12));

	double sl = 0.35;
	RXFOR2(0, m_iNx+1, 0, m_iNz+1){
		Vec2 pos = Vec2((i+0.5)*m_fH, (j+0.5)*m_fH);
		vector<Vec3>::iterator itr = c.begin();
		while(itr != c.end()){
			if(norm2(pos-Vec2(*itr)) < (*itr)[2]*(*itr)[2]){
				m_pQv[ix(i, 0, j, m_iNx+2, m_iNy+2)] = m_Par.Sv;
				m_pQv[ix(i, 1, j, m_iNx+2, m_iNy+2)] = m_Par.Sv;
				m_pQv[ix(i, 2, j, m_iNx+2, m_iNy+2)] = m_Par.Sv;
				break;
			}
			itr++;
		}
	}
}


/*!
 * SgUvZ
 * @param[in] b E
 * @param[inout] f gU
 * @param[in] nx,ny,nz Obh
 * @param[in] diff gUW
 * @param[in] dt ^CXebv
 */
void rxFluidSolver3D::diffuse(int b, RXREAL *f, int nx, int ny, int nz, RXREAL diff, RXREAL dt)
{
	copy(f, m_pX, nx*ny*nz);

	RXREAL a = dt*diff/(m_fH*m_fH);
	RXREAL c = 1+6*a;

	// KEXEUCf
	for(int k = 0; k < 10; ++k){
		RXFOR3(1, nx-1, 1, ny-1, 1, nz-1){
			f[ix(i, j, k, nx, ny)] = (f[ix(i, j, k, nx, ny)]+a*(f[ix(i-1, j, k, nx, ny)]+f[ix(i+1, j, k, nx, ny)]+
																f[ix(i, j-1, k, nx, ny)]+f[ix(i, j+1, k, nx, ny)]+
																f[ix(i, j, k-1, nx, ny)]+f[ix(i, j, k+1, nx, ny)]))/c;
		}
		setBnd(b, f, nx, ny, nz);
	}
}

/*!
 * ICCG@ɂ|A\\o - s̐
 * @param[in] i,j,k ObhCfbNX
 * @param[in] nx,ny,nz Obh
 * @param[in] h Obh 
 * @param[out] A s
 * @param[in] neigh ߖTCfbNX
 */
void rxFluidSolver3D::setMatrix(int i, int j, int k, int nx, int ny, int nz, RXREAL h, RXREAL **A, int **neigh)
{
	int idx = ix(i, j, k, nx, ny);
	RXREAL dens = 1.0;	// x
	RXREAL h2 = h*h;

	RXREAL a = 0.0;
	for(int l = 1; l <= neigh[idx][0]; ++l){
		int jn = neigh[idx][l];
		if(jn >= 0){
			// Obh
			A[idx][l] = -1.0/dens/h2;
			a += 1.0;
		}
		else{
			// ObhO
			A[idx][l] = 0.0;
		}
	}
	A[idx][0] = a/dens/h2;

}

/*!
 * ͂̃|A\ɂ񈳏k(A@,sSRXL[Otz@)
 * @param[inout] u,v x
 * @param[inout] p xi[ꎞIȔz(m_iNx+2)~(m_iNx+2)
 * @param[inout] div xNg̔U(divergence)i[ꎞIȔz
 * @param[in] dt ԃXebv
 */
void rxFluidSolver3D::project(RXREAL dt)
{
	RXREAL *div = m_pX;

	// |A\̉EӍ(Divergence)̌vZƈ͂̏
	RXFOR3(0, m_iNx+2, 0, m_iNy+2, 0, m_iNz+2){
		int idx = ix(i, j, k, m_iNx+2, m_iNy+2);

		RXREAL divx, divy, divz;
		divx = (m_pU[ix(i+1, j, k, m_iNx+3, m_iNy+2)]-m_pU[ix(i, j, k, m_iNx+3, m_iNy+2)])/(m_fH);
		divy = (m_pV[ix(i, j+1, k, m_iNx+2, m_iNy+3)]-m_pV[ix(i, j, k, m_iNx+2, m_iNy+3)])/(m_fH);
		divz = (m_pW[ix(i, j, k+1, m_iNx+2, m_iNy+2)]-m_pW[ix(i, j, k, m_iNx+2, m_iNy+2)])/(m_fH);

		div[idx] = -(divx+divy+divz);	// s̑ΊpvX̒lɂ邽߂ɗӂ"-"Ă
		m_pP[idx] = 0.0;
	}
	// ͂̓̂߂Ɏ͋Ȇxz0ɂĂ
	setBnd(div, m_iNx+2, m_iNy+2, m_iNz+2, 0.0);

	// ߖTiq̍XV
	RXNumerical::SetNeighbors3D(m_pNeigh, m_iNx+2, m_iNy+2, m_iNz+2);

	// s̐
	RXFOR3(0, m_iNx+2, 0, m_iNy+2, 0, m_iNz+2){
		setMatrix(i, j, k, m_iNx+2, m_iNy+2, m_iNz+2, m_fH, m_pPois, m_pNeigh);
	}

	int n = (m_iNx+2)*(m_iNy+2)*(m_iNz+2);

	// sSRXL[
	RXNumerical::IcDecomp<RXREAL>(m_pPois, m_pIC, m_pNeigh, n);

	// ͎z@ŉ
	int max_iter = 100;	// CG@̍ő唽
	RXREAL tol = 1.0e-6;	// CG@̍ő勖e덷
	RXNumerical::pcgSolver<RXREAL>(m_pPois, m_pP, div, m_pIC, m_pNeigh, n, max_iter, tol);
	//RXNumerical::cgSolver<RXREAL>(m_pPois, m_pP, div, m_pNeigh, n, max_iter, tol);
	
	//cout << "iter : " << max_iter << endl;
}

/*!
 * ͍̌vZ
 * @param[in] dt ԃXebv
 */
void rxFluidSolver3D::pressure(RXREAL dt)
{
	RXFOR3(1, m_iNx+2, 0, m_iNy+2, 0, m_iNz+2){
		m_pU[ix(i, j, k, m_iNx+3, m_iNy+2)] -= (m_pP[ix(i, j, k, m_iNx+2, m_iNy+2)]-m_pP[ix(i-1, j, k, m_iNx+2, m_iNy+2)])/m_fH;
	}

	RXFOR3(0, m_iNx+2, 1, m_iNy+2, 0, m_iNz+2){
		m_pV[ix(i, j, k, m_iNx+2, m_iNy+3)] -= (m_pP[ix(i, j, k, m_iNx+2, m_iNy+2)]-m_pP[ix(i, j-1, k, m_iNx+2, m_iNy+2)])/m_fH;
	}

	RXFOR3(0, m_iNx+2, 0, m_iNy+2, 1, m_iNz+2){
		m_pW[ix(i, j, k, m_iNx+2, m_iNy+2)] -= (m_pP[ix(i, j, k, m_iNx+2, m_iNy+2)]-m_pP[ix(i, j, k-1, m_iNx+2, m_iNy+2)])/m_fH;
	}

	setBnd(1, m_pU, m_iNx+3, m_iNy+2, m_iNz+2);
	setBnd(2, m_pV, m_iNx+2, m_iNy+3, m_iNz+2);
	setBnd(3, m_pW, m_iNx+2, m_iNy+2, m_iNz+3);
}

/*!
 * ڗvZ
 * @param[out] f ڗ
 * @param[in] nx,ny,nz ̃Obh
 * @param[in] dx,dy,dz l`Obh̑Έʒu(ObhS(0.5,0.5,0.5))
 * @param[in] dt ^CXebv
 */
void rxFluidSolver3D::advect(RXREAL *f, int nx, int ny, int nz, RXREAL dx, RXREAL dy, RXREAL dz, RXREAL dt)
{
	copy(f, m_pX, nx*ny*nz);

	RXREAL x, y, z;	// vZ_ʒu
	RXFOR3(0, nx, 0, ny, 0, nz){
		// CfbNX
		int idx = ix(i, j, k, nx, ny);

		// vZ_ʒu
		x = (RXREAL)(i+dx)*m_fH;
		y = (RXREAL)(j+dy)*m_fH;
		z = (RXREAL)(k+dz)*m_fH;

		// dtԑÖʒuobNg[X
		Vec3 vel = Velocityf(x, y, z);
		x -= dt*vel[0];
		y -= dt*vel[1];
		z -= dt*vel[2];

		// `ԂŒlXV
		f[idx] = LERP3D(x-dx*m_fH, y-dy*m_fH, z-dz*m_fH, m_pX, m_fH, m_fH, m_fH, nx, ny, nz);
	}
}


/*!
 * C_ւ̑]ڗʂ̌vZ
 * @param[in] T x
 * @param[in] Qv C
 * @param[in] Qc _
 * @return ]ڗ
 */
RXREAL rxFluidSolver3D::calTransitionAmount(int i, int j, int k)
{
	int idx = ix(i, j, k, m_iNx+2, m_iNy+2);

	// ݂̍xɂ鉷x(m_Par.C̒lɂĂqsÛŁCŏCő剷xŃNvĂ)
	RXREAL T = RX_MIN(RX_MAX(m_pT[idx], m_Par.Tamb), m_Par.Tmax);

	// OaC
	//RXREAL qs = RX_MAX(m_Par.A*exp(-m_Par.B/(T+m_Par.C)), m_pQv[idx]+m_pQc[idx]);
	RXREAL qs = m_Par.A*exp(-m_Par.B/(T+m_Par.C));

	// ]ڗ
	return RX_MAX<RXREAL>(m_Par.Alpha*(m_pQv[idx]-qs), 0.0);
}

/*!
 * _ƐC̑ω
 * @param[in] dt ԃXebv
 */
void rxFluidSolver3D::transition(RXREAL dt)
{
	RXFOR3(0, m_iNx+2, 0, m_iNy+2, 0, m_iNz+2){
		int idx = ix(i, j, k, m_iNx+2, m_iNy+2);
		RXREAL cc = calTransitionAmount(i, j, k);

		m_pQc[idx] += dt*cc;
		m_pQv[idx] -= dt*cc;
	}
}

/*!
 * x̐Mɂω
 * @param[in] dt ԃXebv
 */
void rxFluidSolver3D::latent(RXREAL dt)
{
	RXFOR3(0, m_iNx+2, 0, m_iNy+2, 0, m_iNz+2){
		int idx = ix(i, j, k, m_iNx+2, m_iNy+2);
		
		RXREAL cc = calTransitionAmount(i, j, k);
		RXREAL v = 0.5*(m_pV[ix(i, j, k, m_iNx+2, m_iNy+3)]+m_pV[ix(i, j+1, k, m_iNx+2, m_iNy+3)]);

		// MƒfMωɂ鉷xω
		m_pT[idx] += dt*(m_Par.Q*cc-m_Par.GammaD*v);
		
		// nʂɉxCʂ̉xɂĂ
		if(j == 0 || j == 1){
			m_pT[idx] += dt*m_Par.St;
		}
		else if(j == m_iNy || j == m_iNy+1){
			m_pT[idx] = m_Par.Tamb;
		}
	}
}


void rxFluidSolver3D::update(RXREAL dt)
{
	// d͍,O͍
	addGravity(dt);
	addExternalForce(dt);

	// Q
	if(m_bUseVC) vorticityConfinement(dt);

	// SgU
	diffuse(1, m_pU, m_iNx+3, m_iNy+2, m_iNz+2, m_Par.Visc, dt);
	diffuse(2, m_pV, m_iNx+2, m_iNy+3, m_iNz+2, m_Par.Visc, dt);
	diffuse(3, m_pW, m_iNx+2, m_iNy+2, m_iNz+3, m_Par.Visc, dt);
	
	// ͍(񈳏k)
	project(dt);	// ͎Zo
	pressure(dt);	// ͍̌vZ

	// xڗ
	advect(m_pU, m_iNx+3, m_iNy+2, m_iNz+2, 0.0, 0.5, 0.5, dt); setBnd(1, m_pU, m_iNx+3, m_iNy+2, m_iNz+2);
	advect(m_pV, m_iNx+2, m_iNy+3, m_iNz+2, 0.5, 0.0, 0.5, dt); setBnd(2, m_pV, m_iNx+2, m_iNy+3, m_iNz+2);
	advect(m_pW, m_iNx+2, m_iNy+2, m_iNz+3, 0.5, 0.5, 0.0, dt); setBnd(3, m_pW, m_iNx+2, m_iNy+2, m_iNz+3);

	// xgUƈڗ
	latent(dt);
	diffuse(0, m_pT, m_iNx+2, m_iNy+2, m_iNz+2, m_Par.DiffA, dt);

	// nƉẍڗĉ_̃V~[Vɂ͌Ȃ̂ňڗRgAEgĂ
	//advect(m_pT, m_iNx+2, m_iNy+2, m_iNz+2, 0.5, 0.5, 0.5, dt); setBnd(0, m_pT, m_iNx+2, m_iNy+2, m_iNz+2);

	// CƉ_
	transition(dt);
	advect(m_pQv, m_iNx+2, m_iNy+2, m_iNz+2, 0.5, 0.5, 0.5, dt); setBnd(0, m_pQv, m_iNx+2, m_iNy+2, m_iNz+2);
	advect(m_pQc, m_iNx+2, m_iNy+2, m_iNz+2, 0.5, 0.5, 0.5, dt); setBnd(0, m_pQc, m_iNx+2, m_iNy+2, m_iNz+2);

	m_iStep++;
}


/*!
 * steps
 * @param[in] step sXebv 
 * @return 
 */
int rxFluidSolver3D::Run(const int step)
{
	if(m_iStep >= m_iMaxStep) return false;

	for(int k = 0; k < step; ++k){
		update(m_fDt);

		if(m_iStep >= m_iMaxStep) return false;
		m_iStep++;
	}

	return 1;
}

/*!
 * ԂŎs
 * @param[in] t 
 */
int rxFluidSolver3D::RunInTime(const RXREAL t)
{
	return Run((int)(t/m_fDt));
}




//-----------------------------------------------------------------------------
// OpenGLɂ`֐
//-----------------------------------------------------------------------------
/*!
 * x̕`(`͈[0,1])
 */
void rxFluidSolver3D::DrawVelocity(RXREAL dl)
{
	RXREAL dx = 1.0/(m_iNx+2);
	RXREAL dy = 1.0/(m_iNy+2);
	RXREAL dz = 1.0/(m_iNz+2);

	double x, y, z;
	int k = (m_iNz+2)/2;
	z = (k-0.5)*dz;

	glLineWidth(1.0);
	glBegin(GL_LINES);

	for(int i = 0; i < m_iNx+2; ++i){
		x = (double)((i+0.5)*dx);
		for(int j = 0; j < m_iNy+2; ++j){
			y = (double)((j+0.5)*dy);

			double u = (double)(m_pU[ix(i, j, k, m_iNx+3, m_iNy+2)]+m_pU[ix(i+1, j, k, m_iNx+3, m_iNy+2)])/2.0;
			double v = (double)(m_pV[ix(i, j, k, m_iNx+2, m_iNy+3)]+m_pV[ix(i, j+1, k, m_iNx+2, m_iNy+3)])/2.0;
			double w = (double)(m_pW[ix(i, j, k, m_iNx+2, m_iNy+2)]+m_pW[ix(i, j, k+1, m_iNx+2, m_iNy+2)])/2.0;

			glVertex3d(x, y, z);
			glVertex3d(x+dl*u, y+dl*v, z+dl*w);
		}
	}

	glEnd();

}

/*!
 * x̕`(`͈[0,1])
 */
void rxFluidSolver3D::DrawTemperature(void)
{
	RXREAL dx = 1.0/(m_iNx+2);
	RXREAL dy = 1.0/(m_iNy+2);
	RXREAL dz = 1.0/(m_iNz+2);

	glBegin(GL_QUADS);
	RXREAL x, y, z;
	Vec3 c[4];

	RXREAL tmin = m_Par.Tamb;
	RXREAL tmax = m_Par.Tmax;
	
	int k = (m_iNz+2)/2;
	z = (k-0.5)*dz;

	for(int i = 0; i < m_iNx+1; ++i){
		x = (double)(i*dx);
		for(int j = 0; j < m_iNy+1; ++j){
			y = (double)(j*dy);
			
			c[0] = RXFunc::Thermograph(m_pT[ix(i,   j,   k, m_iNx+2, m_iNy+2)], tmin, tmax);
			c[1] = RXFunc::Thermograph(m_pT[ix(i+1, j,   k, m_iNx+2, m_iNy+2)], tmin, tmax);
			c[2] = RXFunc::Thermograph(m_pT[ix(i+1, j+1, k, m_iNx+2, m_iNy+2)], tmin, tmax);
			c[3] = RXFunc::Thermograph(m_pT[ix(i,   j+1, k, m_iNx+2, m_iNy+2)], tmin, tmax);

			glColor3dv(c[0]); glVertex3d(x,    y,    z);
			glColor3dv(c[1]); glVertex3d(x+dx, y,    z);
			glColor3dv(c[2]); glVertex3d(x+dx, y+dy, z);
			glColor3dv(c[3]); glVertex3d(x,    y+dy, z);
		}
	}

	glEnd();
}

/*!
 * C,_̕`(`͈[0,1])
 */
void rxFluidSolver3D::DrawVapor(int vapor)
{
	RXREAL dx = 1.0/(m_iNx+2);
	RXREAL dy = 1.0/(m_iNy+2);
	RXREAL dz = 1.0/(m_iNz+2);

	RXREAL x, y, z;
	Vec3 c[4];

	RXREAL qmin = 0.0;
	RXREAL qmax = 1.0;
	
	int k = (m_iNz+2)/2;
	z = (k-0.5)*dz;

	RXREAL *q = vapor ? m_pQv : m_pQc;
	Vec3 col1(0.0), col2(1.0);

	glBegin(GL_QUADS);
	for(int i = 0; i < m_iNx+1; ++i){
		x = (double)(i*dx);
		for(int j = 0; j < m_iNy+1; ++j){
			y = (double)(j*dy);
			
			RXFunc::Gradation(c[0], col1, col2, q[ix(i,   j,   k, m_iNx+2, m_iNy+2)], qmin, qmax);
			RXFunc::Gradation(c[1], col1, col2, q[ix(i+1, j,   k, m_iNx+2, m_iNy+2)], qmin, qmax);
			RXFunc::Gradation(c[2], col1, col2, q[ix(i+1, j+1, k, m_iNx+2, m_iNy+2)], qmin, qmax);
			RXFunc::Gradation(c[3], col1, col2, q[ix(i,   j+1, k, m_iNx+2, m_iNy+2)], qmin, qmax);

			glColor3dv(c[0]); glVertex3d(x,    y,    z);
			glColor3dv(c[1]); glVertex3d(x+dx, y,    z);
			glColor3dv(c[2]); glVertex3d(x+dx, y+dy, z);
			glColor3dv(c[3]); glVertex3d(x,    y+dy, z);
		}
	}

	glEnd();
}




//-----------------------------------------------------------------------------
// fobO֐
//-----------------------------------------------------------------------------
template<class T> 
void rxFluidSolver3D::dump(const char *fn, int nx, int ny, int k, T *data, const char *comment, bool app)
{
	if(!data) return;

	ofstream fo;
	if(app){
		fo.open(fn, ios::app);
	}
	else{
		fo.open(fn);
	}

	fo << "[" << comment << "]" << endl;
	for(int j = ny-1; j >= 0; --j){
		for(int i = 0; i < nx; ++i){
			int idx = k*nx*ny+j*nx+i;
			fo << data[idx] << " ";
		}
		fo << endl;
	}
	fo << endl;

	fo.close();
}

