/*!
  @file rx_fluid_solver.
	
  @brief V~[V
 
  @author Makoto Fujisawa
  @date 2013-10
*/


#ifndef _RX_FLUID_SOLVER_H_
#define _RX_FLUID_SOLVER_H_



//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_common.h"

using namespace std;



//-----------------------------------------------------------------------------
// V[p[^\
//-----------------------------------------------------------------------------
struct rxEnviroment
{
	int Nx, Ny, Nz;				//!< Obh
	RXREAL H;					//!< Obh
	RXREAL Dt;					//!< ^CXebv

	RXREAL Gravity;				//!< d͉x

	// xɊւp[^
	RXREAL Visc;				//!< SW()
	RXREAL Frc;					//!< }EXɂO͐ݒ̂߂̌W
	RXREAL VC;					//!< Qx̌W

	// xɊւp[^
	RXREAL Diff;				//!< gUW(̖x)
	RXREAL Src;					//!< }EXɂ鉌x̒ǉ

	// xɊւp[^
	RXREAL DiffA;				//!< gUW(x)
	RXREAL Beta;				//!< ͌W
	RXREAL Tamb;				//!< x
	RXREAL Tmax;				//!< ő剷x(Mx)
	RXREAL Radians;				//!< MˌW

	rxEnviroment()
	{
		Nx = Ny = Nz = 64;
		H = 1.0/Nx;
		Dt = 0.03;
		Gravity = 9.8;

		Visc = 0.0;
		Frc = 300.0;
		VC = 0.25;

		Diff = 0.0;
		Src = 10.0;

		DiffA = 0.01;
		Beta = 1.0;
		Tamb = 0.0;
		Tmax = 100.0;
		Radians = 1.0;
	}
};



//-----------------------------------------------------------------------------
// rxFluidSolver3DNX
//-----------------------------------------------------------------------------
class rxFluidSolver3D
{
protected:
	// oϐ
	RXREAL *m_pU, *m_pV, *m_pW;		//!< x
	RXREAL *m_pD;					//!< x
	RXREAL *m_pC;					//!< Qx - Curl

	RXREAL *m_pP;					//!< ͏
	RXREAL *m_pT;					//!< x

	RXREAL *m_pX;					//!< ꎞIȒl̊mۏꏊ

	int m_iStep, m_iMaxStep;		//!< vZXebvpϐ
	RXREAL m_fDt;					//!< ^CXebv

	RXREAL **m_pPois, **m_pIC;		//!< Poisson̍si[
	int **m_pNeigh;					//!< ߖTObh̃CfbNX

	int m_iNx, m_iNy, m_iNz;		//!< Obh
	RXREAL m_fH;					//!< Obh

	rxEnviroment m_Par;				//!< ̂Ɋւp[^

	bool m_bUseVC;					//!< Vorticity confinementgptO

protected:
	//! ftHgRXgN^
	rxFluidSolver3D(){}

public:
	//! RXgN^
	rxFluidSolver3D(rxEnviroment scn);

	//! fXgN^
	~rxFluidSolver3D();

public:
	// ANZX\bh
	bool IsVortex(void) const { return m_bUseVC; }
	void SetVortex(bool vc){ m_bUseVC = vc; }

	int GetGridNum(int d = 0){ return (d == 0) ? m_iNx+2 : (d == 1 ? m_iNy+2 : m_iNz+2); }

public:
	// V~[VXebv
	int Run(const int step = 1);
	int RunInTime(const RXREAL t);

	void SetUI(int target, Vec2 mx, Vec2 mx_prev);

	// f[^
	int  Allocate(void);
	void Finalize(void);
	void Clear(void);

	// `֐
	void DrawDensity(void);
	void DrawTemperature(void);
	void DrawVelocity(RXREAL dl = 0.1);

	// x擾
	RXREAL U(RXREAL x, RXREAL y, RXREAL z){ return LERP3D(x, y-0.5*m_fH, z-0.5*m_fH, m_pU, m_fH, m_fH, m_fH, m_iNx+3, m_iNy+2, m_iNz+2); }
	RXREAL V(RXREAL x, RXREAL y, RXREAL z){ return LERP3D(x-0.5*m_fH, y, z-0.5*m_fH, m_pV, m_fH, m_fH, m_fH, m_iNx+2, m_iNy+3, m_iNz+2); }
	RXREAL W(RXREAL x, RXREAL y, RXREAL z){ return LERP3D(x-0.5*m_fH, y-0.5*m_fH, z, m_pW, m_fH, m_fH, m_fH, m_iNx+2, m_iNy+2, m_iNz+3); }
	Vec3 Velocityf(RXREAL x, RXREAL y, RXREAL z){ return Vec3(U(x, y, z), V(x, y, z), W(x, y, z)); }

	// x擾
	static RXREAL GetDensityf_s(RXREAL x, RXREAL y, RXREAL z, void* ptr);
	inline RXREAL GetDensityf(RXREAL x, RXREAL y, RXREAL z);

	// {[f[^
	void GetGridSize(int &nx, int &ny, int &nz) const { nx = m_iNx+2; ny = m_iNy+2; nz = m_iNz+2; }
	RXREAL* GetVolume(void){ return &m_pT[0]; }


protected:
	inline int ix(int i, int j, int k, int nx, int ny){ return (i+(j+k*ny)*nx); }
	inline void copy(RXREAL *src, RXREAL *dst, int n){ for(int i = 0; i < n; ++i) dst[i] = src[i]; }

	// vZXebv
	void update(RXREAL dt);

	// Eݒ
	void setBnd(int b, RXREAL *x, int nx, int ny, int nz);
	void setBnd(RXREAL* x, int nx, int ny, int nz, RXREAL val);

	// ÓEd
	void addExternalForce(RXREAL dt);
	void addGravity(RXREAL dt);

	// gUCڗ
	void diffuse(int b, RXREAL *f, int nx, int ny, int nz, RXREAL diff, RXREAL dt);
	void advect(RXREAL *f, int nx, int ny, int nz, RXREAL dx, RXREAL dy, RXREAL dz, RXREAL dt);

	// 
	void setMatrix(int i, int j, int k, int nx, int ny, int nz, RXREAL h, RXREAL **poiss, int **neigh);
	void project(RXREAL dt);
	void pressure(RXREAL dt);

	// Qx
	void vorticityConfinement(RXREAL dt);
	Vec3 calCurl(int i, int j, int k);

	// Mt
	void radiation(RXREAL dt);
	 
	// o
	template<class T> void dump(const char *fn, int nx, int ny, int k, T *data, const char *comment, bool app);
};





#endif // #ifndef _RX_FLUID_SOLVER_H_