/*!
  @file rx_volume_renderer.cpp
	
  @brief XCXp{[_O
 
  @author Makoto Fujisawa
  @date 2012-12
*/



//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include "rx_volume_renderer.h"

// Nbv : ax+by+cz+d >= 0`悳
const GLdouble CLIP_PLANE[6][4] = { {  1,  0,  0,  0 },		// x = 0
									{ -1,  0,  0,  1 },		// x = 1
									{  0,  1,  0,  0 },		// y = 0
									{  0, -1,  0,  1 },		// y = 1
									{  0,  0,  1,  0 },		// z = 0
									{  0,  0, -1,  1 } };	// z = 1


//-----------------------------------------------------------------------------
// rxVolumeRendererNX̎
//-----------------------------------------------------------------------------
/*!
 * RXgN^
 * @param[in] near,far Nbv
 * @param[in] num_layer XCX̐
 */
rxVolumeRenderer::rxVolumeRenderer(double near, double far, int num_layer)
	: m_fNear(near), m_fFar(far), m_iNumLayers(num_layer) 
{
	m_uVolumeTex = 0;
	m_iNx = m_iNy = m_iNz = 0;
	m_v3DensColor = Vec3(1.0, 1.0, 1.0);
	m_fOpacity = 60.0;
	m_iShaderInit = 0;

	m_uColorTex = 0;
	m_iW = m_iH = m_iC = 0;
	m_fMinT = 0.0;
	m_fMaxT = 100.0;
}

/*!
 * fXgN^
 */
rxVolumeRenderer::~rxVolumeRenderer()
{
	if(glIsTexture(m_uVolumeTex)) glDeleteTextures(1, &m_uVolumeTex);
}


/*!
 * GLSLVF[_̏
 */
void rxVolumeRenderer::initShaders(void)
{
	m_glslVolume = CreateGLSL(volume_image_vs, volume_image_fs, "Volume Renderer");
	m_iShaderInit = 1;
}


/*!
 * FeNX`̃Zbg
 * @param[in] data  RGB摜f[^
 * @param[in] w,h,c 摜̃TCYƐF
 */
int rxVolumeRenderer::SetColorData(unsigned char *data, int w, int h, int c)
{
	int update = true;
	if((w != m_iW || h != m_iH || c != m_iC)){
		// ̃eNX`f[^̔j
		if(glIsTexture(m_uColorTex)) glDeleteTextures(1, &m_uColorTex);
		glGenTextures(1, &m_uColorTex);
		update = false;
	}

	// 摜tH[}bg
	GLuint iformat, format;
	format = GL_RGBA;
	iformat = GL_RGBA;
	if(c == 1){
		format = GL_LUMINANCE;
		iformat = GL_LUMINANCE;
	}
	else if(c == 3){
		format = GL_RGB;
		iformat = GL_RGB;
	}

	glBindTexture(GL_TEXTURE_3D, m_uColorTex);

	if(!update){
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
	}

	glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
	glTexImage2D(GL_TEXTURE_2D, 0, iformat, w, h, 0, format, GL_UNSIGNED_BYTE, data);

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);

	glBindTexture(GL_TEXTURE_2D, 0);

	m_iW = w;
	m_iH = h;
	m_iC = c;

	return 0;
}


/*!
 * {[f[^eNX`ɐݒ(unsigned char)
 * @param[in] data {[f[^
 * @param[in] nx,ny,nz f[^TCY
 */
int rxVolumeRenderer::SetVolumeData(unsigned char *data, int nx, int ny, int nz)
{
	int update = true;
	if((nx != m_iNx || ny != m_iNy || nz != m_iNz)){
		// ̃eNX`f[^̔j
		if(glIsTexture(m_uVolumeTex)) glDeleteTextures(1, &m_uVolumeTex);
		glGenTextures(1, &m_uVolumeTex);
		update = false;
	}

	glBindTexture(GL_TEXTURE_3D, m_uVolumeTex);

	if(update){
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, 0, nx, ny, nz, GL_LUMINANCE, GL_UNSIGNED_BYTE, data);
	}
	else{
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE, nx, ny, nz, 0, GL_LUMINANCE, GL_UNSIGNED_BYTE, data);

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	m_iNx = nx;
	m_iNy = ny;
	m_iNz = nz;

	return 0;
}

/*!
 * {[f[^eNX`ɐݒ(RXREAL)
 * @param[in] data {[f[^
 * @param[in] nx,ny,nz f[^TCY
 */
int rxVolumeRenderer::SetVolumeDataf(RXREAL *data, int nx, int ny, int nz, Vec3 col)
{
	int update = true;
	if((nx != m_iNx || ny != m_iNy || nz != m_iNz)){
		// ̃eNX`f[^̔j
		if(glIsTexture(m_uVolumeTex)) glDeleteTextures(1, &m_uVolumeTex);
		glGenTextures(1, &m_uVolumeTex);
		update = false;
	}

	glBindTexture(GL_TEXTURE_3D, m_uVolumeTex);
	int type = (sizeof(RXREAL) == sizeof(double) ? GL_DOUBLE : GL_FLOAT);

	if(update){
		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glTexSubImage3D(GL_TEXTURE_3D, 0, 0, 0, 0, nx, ny, nz, GL_LUMINANCE, type, data);
	}
	else{
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
		glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE, nx, ny, nz, 0, GL_LUMINANCE, type, data);

		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
	}

	m_v3DensColor = col;

	m_iNx = nx;
	m_iNy = ny;
	m_iNz = nz;

	return 0;
}

/*!
 * {[`͈̈͂̌vZ
 * @param[out] minp,maxp ͈
 */
void rxVolumeRenderer::calBBox(Vec3 &minp, Vec3 &maxp)
{
	// OpenGL̃fr[s
	double m[16];
	glGetDoublev(GL_MODELVIEW_MATRIX, m);

	minp[0] = m[3*4+0];
	maxp[0] = m[3*4+0];
	for(int i = 0; i < 3; i++){
		if(m[i*4+0] < 0){
			minp[0] += m[i*4+0];
		} else{
			maxp[0] += m[i*4+0];
		}
	}

	minp[1] = m[3*4+1];
	maxp[1] = m[3*4+1];
	for(int i = 0; i < 3; i++){
		if(m[i*4+1] < 0){
			minp[1] += m[i*4+1];
		} else{
			maxp[1] += m[i*4+1];
		}
	}

	minp[2] = m[3*4+2];
	maxp[2] = m[3*4+2];
	for(int i = 0; i < 3; i++){
		if(m[i*4+2] < 0){
			minp[2] += m[i*4+2];
		} else{
			maxp[2] += m[i*4+2];
		}
	}
}

/*!
 * {[f[^_O
 *  - {[f[^(0, 0, 0)-(1, 1, 1)̗̓ɕ`悳
 */
void rxVolumeRenderer::Draw(void)
{
	if(!m_iShaderInit){
		initShaders();
	}
	
	glPushMatrix();

	// XCX̕`͈͂fr[s񂩂vZ
	Vec3 minp, maxp;
	calBBox(minp, maxp);

	// NbvʂL
	for(int i = 0; i < 6; i++){
		glClipPlane(GL_CLIP_PLANE0+i, CLIP_PLANE[i]);
		glEnable(GL_CLIP_PLANE0+i);
	}
	
	// XCXԋ
	double thickness = (maxp[2]-minp[2])/m_iNumLayers;

	// JOƃCeBOOFF
	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	
	// At@uhON
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// {[eNX`Lɂ
	glActiveTexture(GL_TEXTURE0);
	glEnable(GL_TEXTURE_3D);
	glBindTexture(GL_TEXTURE_3D, m_uVolumeTex);
	
	// GLSL
	glUseProgram(m_glslVolume.Prog);
	glUniform1i(glGetUniformLocation(m_glslVolume.Prog, "volume_tex"), 0);
	glUniform1i(glGetUniformLocation(m_glslVolume.Prog, "color_tex"), 1);
	glUniform1f(glGetUniformLocation(m_glslVolume.Prog, "thickness"), thickness);
	glUniform1f(glGetUniformLocation(m_glslVolume.Prog, "opacity"), (double)m_fOpacity);
	glUniform3f(glGetUniformLocation(m_glslVolume.Prog, "dens_col"), (double)m_v3DensColor[0], (double)m_v3DensColor[1], (double)m_v3DensColor[2]);

	// XCX`
	glPushMatrix();
	glColor4d(1.0, 0.0, 0.0, 1.0);
	glBegin(GL_QUADS);
	for(int i = 0; i < m_iNumLayers; i++){
		double z = minp[2]+thickness*i;
		glVertex3f(minp[0], minp[1], z);
		glVertex3f(maxp[0], minp[1], z);
		glVertex3f(maxp[0], maxp[1], z);
		glVertex3f(minp[0], maxp[1], z);
	}
	glEnd();
	glPopMatrix();

	glUseProgram(0);

	glActiveTexture(GL_TEXTURE0);
	glDisable(GL_TEXTURE_3D);

	glDisable(GL_BLEND);

	// Nbvʂ𖳌
	for(int i = 0; i < 6; i++){
		glDisable(GL_CLIP_PLANE0+i);
	}

	glPopMatrix();
}
