/*!
  @file glmain.cpp
	
  @brief V~[V
 
  @author Makoto Fujisawa
  @date   2013-10
*/

#pragma comment(lib, "glew32.lib")


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <GL/glew.h>

#include "rx_common.h"

#include "rx_bitmap.h"
#include "rx_vld.h"

#include "rx_trackball.h"
#include "rx_atom_ini.h"

#include "rx_fluid_solver.h"
#include "rx_volume_renderer.h"

#include "rx_config.h"


//-----------------------------------------------------------------------------
// `/萔
//-----------------------------------------------------------------------------
const string RX_PROGRAM_NAME = "rx_smoke3d";

const GLfloat RX_FOV = 45.0f;

const GLfloat RX_NEAR_CLIP = 0.1f;
const GLfloat RX_FAR_CLIP  = 100.0f;

//! etO
enum{
	RXD_BOUNDARY	= (1 << 0), 	//!< V~[VԋE`
	RXD_VELOCITY	= (1 << 1), 	//!< x`
	RXD_DENSITY		= (1 << 2),		//!< x`
	RXD_TEMPERATURE = (1 << 3),		//!< x`
	RXD_VOLUME		= (1 << 4),		//!< {[f[^`
	RXD_OUT			= (1 << 5),		//!< {[f[^̃t@Co
};


//-----------------------------------------------------------------------------
// MARK:O[oϐ
//-----------------------------------------------------------------------------
bool g_bIdle = false;

// }EX
int g_iWinW = 720;				//!< `EBhE̕
int g_iWinH = 720;				//!< `EBhE̍
int g_iWinX = 100;				//!< `EBhEʒux
int g_iWinY = 100;				//!< `EBhEʒuy

int g_iWinPos[2] = {100, 100};	//!< EBhEʒu(tXN[ؑ֎̋Lp)
int g_iWinSize[2] = {640, 640};	//!< EBhETCY(tXN[ؑ֎̋Lp)
bool g_bFullScreen = false;		//!< tXN[\

//! `tO
int g_iDraw = RXD_BOUNDARY | RXD_VOLUME;

// 摜o
int g_iSaveImageSpacing = -1;	//!< 摜ۑԊu(=-1ȂۑȂ)
bool g_bSavedImage = false;		//!< dۑh~ptO

// ݒt@C
rxINI *g_pINI = NULL;

// V[ݒ
rxSceneConfig g_Scene;
int g_iCurrentSceneIdx = 0;

// ̃\o
rxFluidSolver3D *g_pSolver = 0;

// vZXebv
int g_iStep = 0;

// V~[V̈
Vec3 g_v3Range[2] = {Vec3(0.0), Vec3(1.0)};

// gbN{[
rxTrackball g_tbView;			//!< _

//! {[_[
rxVolumeRenderer *g_pRender = 0;


//-----------------------------------------------------------------------------
// ֐vg^Cv錾
//-----------------------------------------------------------------------------
void OnExitApp(int id = -1);

void Idle(void);
void CleanGL(void);



//-----------------------------------------------------------------------------
// ֐̎
//-----------------------------------------------------------------------------

/*!
 * ACh֐ON/OFF
 * @param[in] on 1ON, 0OFF, -1ON/OFF؂ւ
 */
void SwitchIdle(int on = -1)
{
	g_bIdle = (on == -1) ? !g_bIdle : (on ? true : false);
	glutIdleFunc((g_bIdle ? Idle : 0));
	cout << "idle " << (g_bIdle ? "on" : "off") << endl;
}

/*!
 * tXN[/EBhE\̐؂ւ
 */
void SwitchFullScreen(void)
{
	if(g_bFullScreen){
		glutPositionWindow(g_iWinPos[0], g_iWinPos[1]);
		glutReshapeWindow(g_iWinSize[0], g_iWinSize[1]);
		g_bFullScreen = false;
	}
	else{
		g_iWinPos[0] = glutGet(GLUT_WINDOW_X);
		g_iWinPos[1] = glutGet(GLUT_WINDOW_Y);
		g_iWinSize[0] = glutGet(GLUT_WINDOW_WIDTH);
		g_iWinSize[1] = glutGet(GLUT_WINDOW_HEIGHT);

		glutFullScreen();
		g_bFullScreen = true;
	}
}

/*!
 * AvP[VI  Quit{^̃Cxgnh
 * @param[in] id {^̃CxgID
 */
void OnExitApp(int id)
{
	CleanGL();

	exit(1);
}

/*!
 * ݒt@Cǂݍ
 * @param[in] fn ݒt@C(gq)
 */
void ReadConfig(const string &fn)
{
	// AvP[VǗݒt@C
	g_pINI = new rxINI();

	g_pINI->Set("window", "width",  &g_iWinW, 720);
	g_pINI->Set("window", "height", &g_iWinH, 720);
	g_pINI->Set("window", "pos_x",  &g_iWinX, 100);
	g_pINI->Set("window", "pos_y",  &g_iWinY, 100);
	g_pINI->Set("scene", "idx", &g_iCurrentSceneIdx, 0);

	if(!(g_pINI->Load(fn+".ini"))){
		RXCOUT << "Failed opening the " << fn << ".ini file!" << endl;
	}
	
}

/*!
 * ݒt@C
 * @param[in] fn ݒt@C(gq)
 */
void WriteConfig(const string &fn)
{
	g_iWinX = glutGet(GLUT_WINDOW_X);
	g_iWinY = glutGet(GLUT_WINDOW_Y);
	g_iCurrentSceneIdx = g_Scene.GetCurrentSceneIdx();

	if(g_pINI->Save(fn+".ini")){
		RXCOUT << "save : " << fn << ".ini" << endl;
	}
}

/*!
 * EBhE^Cgݒ
 */
void SetTitle(void)
{
	stringstream ss;
	ss << "step " << g_iStep;
	ss << ((g_iDraw & RXD_OUT) ? " - output" : "");
	ss << " - scene " << g_iCurrentSceneIdx+1;
	glutSetWindowTitle(ss.str().c_str());
}

/*!
 * ݂̉ʕ`摜t@CƂĕۑ
 * @param[in] fn t@CpX
 */
void SaveDisplay(const string &fn)
{
	int w_ = g_iWinW;
	int h_ = g_iWinH;
	int c_ = 4;

	unsigned char* data = new unsigned char[w_*h_*c_];

	glReadPixels(0, 0, w_, h_, GL_RGBA, GL_UNSIGNED_BYTE, data);

	// ㉺]
	int stride = w_*c_;
	for(int j = 0; j < h_/2; ++j){
		for(int i = 0; i < stride; ++i){
			unsigned char tmp = data[j*stride+i];
			data[j*stride+i] = data[(h_-j-1)*stride+i];
			data[(h_-j-1)*stride+i] = tmp;
		}
	}

	string ext = GetExtension(fn);
	if(ext == "bmp"){
		WriteBitmapFile(fn, data, w_, h_, c_, RX_BMP_WINDOWS_V3);
		RXCOUT << "saved the screen image to " << fn << endl;
	}

	delete [] data;
}


//-----------------------------------------------------------------------------
// {[_O
//-----------------------------------------------------------------------------
/*!
 * {[_̏
 */
void InitVolume(void)
{
	if(!g_pSolver) return;

	int nx, ny, nz;
	RXREAL *data;
	g_pSolver->GetGridSize(nx, ny, nz);
	data = g_pSolver->GetVolume();

	//
	// _
	//
	g_pRender = new rxVolumeRenderer(RX_NEAR_CLIP, RX_FAR_CLIP, 128);
	g_pRender->SetVolumeDataf(data, nx, ny, nz, Vec3(0.9));
}

/*!
 * {[f[^̍XV
 * @param[in] output t@Co̓tO
 */
void UpdateVolume(int output = 0)
{
	if(!g_pRender) return;

	int nx, ny, nz;
	RXREAL *data;

	g_pSolver->GetGridSize(nx, ny, nz);
	data = g_pSolver->GetVolume();

	g_pRender->SetVolumeDataf(data, nx, ny, nz, Vec3(0.9));

	// {[f[^̃t@Co
	if(output){
		string fn = CreateFileName("vol/volume_", ".vld", g_iStep, 5);
		rxVLD<RXREAL> vld;
		vld.Save(fn, nx, ny, nz, data);
	}
}


//-----------------------------------------------------------------------------
// ̃V~[V
//-----------------------------------------------------------------------------
/*!
 * ̃V~[^̏
 */
void InitSim(void)
{
	g_Scene.ReadSceneFiles();					// V[t@C̓ǂݎ
	g_Scene.SetCurrentScene(g_iCurrentSceneIdx);
	if(g_Scene.LoadSpaceFromFile()){
		// V[t@CV~[Vԏ񂪓ǂݍ߂\o\z
		if(g_pSolver) delete g_pSolver;
		g_pSolver = new rxFluidSolver3D(g_Scene.GetEnv());

		// \oݒ肵āCV[擾
		g_Scene.Set(g_pSolver);
		g_Scene.LoadSceneFromFile();

		// Xebvƃ{[f[^XV
		g_iStep = 0;
		UpdateVolume();

		SetTitle();
		glutPostRedisplay();
	}
	else{
		RXCOUT << "there is no scene file." << endl;
		exit(0);
	}
}



//------------------------------------------------------------------
// OpenGLLoXp̃Cxgnh
//------------------------------------------------------------------
/*!
 * ĕ`Cxg֐
 */
void Display(void)
{
	// t[obt@ƃfvXobt@NA
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glDisable(GL_CULL_FACE);
	glDisable(GL_LIGHTING);
	glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

	glPushMatrix();

	// }EXɂ]Esړ̓Kp
	g_tbView.Apply();
		
	glTranslatef(-0.5, -0.5, -0.5);

	// x̎o(z=nz/2̖ʂ̂)
	if(g_iDraw & RXD_VELOCITY){
		glDisable(GL_LIGHTING);
		glColor3d(0.8, 0.8, 0.8);
		g_pSolver->DrawVelocity();
	}

	// E
	if(g_iDraw & RXD_BOUNDARY){
		glColor3d(0.4, 1.0, 0.4);
		glLineWidth(2.0);
		DrawWireCuboid(g_v3Range[0], g_v3Range[1]);
		glLineWidth(1.0);
	}

	// x̕`(z=nz/2̖ʂ̂)
	if(g_iDraw & RXD_DENSITY){
		glDisable(GL_LIGHTING);
		g_pSolver->DrawDensity();
	}

	// x̕`(z=nz/2̖ʂ̂)
	if(g_iDraw & RXD_TEMPERATURE){
		glDisable(GL_LIGHTING);
		g_pSolver->DrawTemperature();
	}

	// {[f[^̕`
	if((g_iDraw & RXD_VOLUME) && g_pRender){
		glPushMatrix();
		g_pRender->Draw();
		glPopMatrix();
	}


	glPopMatrix();

	glutSwapBuffers();
}


/*! 
 * TCYCxg֐
 * @param[in] w LoX(sNZ)
 * @param[in] h LoX(sNZ)
 */
void Resize(int w, int h)
{
	g_iWinW = w;
	g_iWinH = h;

	glViewport(0, 0, w, h);
	g_tbView.SetRegion(w, h);

	// ϊs̐ݒ
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(RX_FOV, (float)g_iWinW/(float)g_iWinH, RX_NEAR_CLIP, RX_FAR_CLIP);

	// fr[ϊs̐ݒ
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

/*!
 * }EXCxg֐
 * @param[in] button }EX{^(GLUT_LEFT_BUTTON,GLUT_MIDDLE_BUTTON,GLUT_RIGHT_BUTTON)
 * @param[in] state }EX{^̏(GLUT_UP, GLUT_DOWN)
 * @param[in] x,y }EXW(XN[Wn)
 */
void Mouse(int button, int state, int x, int y)
{
	if(x < 0 || y < 0 || x > g_iWinW || y > g_iWinH) return;

	int mod = glutGetModifiers();	// SHIFT,CTRL,ALT̏Ԏ擾

	if(button == GLUT_LEFT_BUTTON){
		if(state == GLUT_DOWN){
			g_tbView.Start(x, y, mod+1);
		}
		else if(state == GLUT_UP){
			g_tbView.Stop(x, y);
		}
	}

	glutPostRedisplay();
}

/*!
 * [VCxg֐(}EX{^܂܃hbO)
 * @param[in] x,y }EXW(XN[Wn)
 */
void Motion(int x, int y)
{
	if(x < 0 || y < 0 || x > g_iWinW || y > g_iWinH) return;
	
	g_tbView.Motion(x, y);

	glutPostRedisplay();
}


/*! 
 * AChCxg֐
 */
void Idle(void)
{
	if(g_pSolver){
		// V~[VXebvi߂
		g_pSolver->Run();

		// {[f[^XV
		UpdateVolume((g_iDraw & RXD_OUT));

		// `ʂ̉摜o
		if(g_iSaveImageSpacing > 0 && (g_iStep%g_iSaveImageSpacing == 0)){
			SaveDisplay(CreateFileName("img/display_", ".bmp", g_iStep, 5));
		}

		// EBhE^CgXV
		SetTitle();
		
		g_iStep++;
		glutPostRedisplay();
	}
}

/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void Keyboard(unsigned char key, int x, int y)
{
	switch(key){
	case 'q':
	case 'Q':
	case '\033':
		OnExitApp(0);
		break;

	case ' ':	// V~[V1Xebvi߂
		Idle();
		break;

	case 's':	// Aj[VON/OFF
		SwitchIdle(-1);
		break;

	case 'F':	// tXN[ON/OFF
		SwitchFullScreen();
		break;

	case 'b':	// E`
		g_iDraw ^= RXD_BOUNDARY;
		break;

	case 'v':	// x`
		g_iDraw ^= RXD_VELOCITY;
		break;

	case 'd':	// x`
		g_iDraw ^= RXD_DENSITY;
		break;

	case 't':	// x`
		g_iDraw ^= RXD_TEMPERATURE;
		break;

	case 'V':	// {[_O
		g_iDraw ^= RXD_VOLUME;
		break;

	case 'w':	// QxꋭON/OFF
		g_pSolver->SetVortex(!g_pSolver->IsVortex());
		cout << "vorticity confinement : " << (g_pSolver->IsVortex() ? "on" : "off") << endl;
		break;

	case 'o':	// t@Co
		g_iDraw ^= RXD_OUT;
		SetTitle();
		break;

	default:
		break;
	}

	glutPostRedisplay();
}

/*!
 * L[{[hCxg֐
 * @param[in] key L[̎
 * @param[in] x,y L[ꂽƂ̃}EXW(XN[Wn)
 */
void SpecialKey(int key, int x, int y)
{
	// F1`F12ŃV[ݒǂݍ
	if(key >= GLUT_KEY_F1 && key <= GLUT_KEY_F12){
		if(g_Scene.SetCurrentScene(key-GLUT_KEY_F1)){
			g_iCurrentSceneIdx = g_Scene.GetCurrentSceneIdx();
			InitSim();
		}
	}
}



/*! 
 * GL̏֐
 */
void InitGL(void)
{
	RXCOUT << "OpenGL Ver. " << glGetString(GL_VERSION) << endl;
	glewInit();

	glClearColor(0.0, 0.0, 0.0, 1.0);
	glClearDepth(1.0f);

	glEnable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);

	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	// ݒ
	static const GLfloat lpos[] = { 0.0f, 0.0f, 1.0f, 1.0f }; // ʒu
	static const GLfloat difw[] = { 0.7f, 0.7f, 0.7f, 1.0f }; // gUF : 
	static const GLfloat spec[] = { 0.2f, 0.2f, 0.2f, 1.0f }; // ʔːF
	static const GLfloat ambi[] = { 0.1f, 0.1f, 0.1f, 1.0f }; // 
	glLightfv(GL_LIGHT0, GL_POSITION, lpos);
	glLightfv(GL_LIGHT0, GL_DIFFUSE,  difw);
	glLightfv(GL_LIGHT0, GL_SPECULAR, spec);
	glLightfv(GL_LIGHT0, GL_AMBIENT,  ambi);
	glEnable(GL_LIGHT0);

	glEnable(GL_LIGHTING);
	glShadeModel(GL_SMOOTH);

	glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);

	// gbN{[p
	g_tbView.SetScaling(-2.0);
	g_tbView.SetTranslation(0.0, 0.0);
}



/*! 
 * GL̏I֐
 */
void CleanGL(void)
{
	// ݒt@C
	WriteConfig(RX_PROGRAM_NAME);
	if(g_pINI != NULL) delete g_pINI;

	if(g_pSolver) delete g_pSolver;
}



/*!
 * Cj[
 * @param[in] id j[ID
 */
void OnMainMenu(int id)
{
	Keyboard((unsigned char)id, 0, 0);
}

/*!
 * GLUT̉ENbNj[쐬
 */
void InitMenu(void)
{
	// Cj[
	glutCreateMenu(OnMainMenu);
	glutAddMenuEntry("Toggle velocity drawing [v]",	'v');
	glutAddMenuEntry("Toggle density drawing [d]",	'd');
	glutAddMenuEntry("Toggle temperature drawing [t]",	't');
	glutAddMenuEntry("Toggle volume drawing [V]",	'V');
	glutAddMenuEntry("Toggle boundary drawing [b]",	'b');
	glutAddMenuEntry("------------", -1);
	glutAddMenuEntry("Toggle vortex conf. [w]",		'w');
	glutAddMenuEntry("Toggle volume output [o]",	'o');
	glutAddMenuEntry("------------", -1);
	glutAddMenuEntry("Toggle Fullscreen [F]",		'F');
	glutAddMenuEntry("Toggle Animation [s]",		's');
	glutAddMenuEntry("Step Animation [ ]",			' ');
	glutAddMenuEntry("------------", -1);
	glutAddMenuEntry("Exit [q]",					'q');

	glutAttachMenu(GLUT_RIGHT_BUTTON);
}



/*!
 * C[`
 * @param[in] argc R}hC̐
 * @param[in] argv R}hC
 */
int main(int argc, char *argv[])
{
	ReadConfig(RX_PROGRAM_NAME);

	MkDir("img");
	MkDir("vol");

	// GLUTEBhE
	glutInit(&argc, argv);

	glutInitWindowPosition(g_iWinX, g_iWinY);
	glutInitWindowSize(g_iWinW, g_iWinH);

	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutCreateWindow(argv[0]);

	InitSim();

	glutDisplayFunc(Display);
	glutReshapeFunc(Resize);
	glutMouseFunc(Mouse);
	glutMotionFunc(Motion);
	glutKeyboardFunc(Keyboard);
	glutSpecialFunc(SpecialKey);

	SwitchIdle(0);

	InitGL();
	InitMenu();

	InitVolume();

	glutMainLoop();

	CleanGL();

	return 0;
}


