/*!
  @file rx_config.h
	
  @brief V[ݒt@Cǂݍ
 
  @author Makoto Fujisawa
  @date 2013-10
*/

#ifndef _RX_FLUID_CONFIG_H_
#define _RX_FLUID_CONFIG_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// STL
#include <vector>
#include <string>

// [eBeB
#include "rx_utility.h"

// V~[V
#include "rx_common.h"
#include "rx_fluid_solver.h"

// ݒt@C
#include "rx_atom_ini.h"

using namespace std;


//-----------------------------------------------------------------------------
// 񏈗֐
//-----------------------------------------------------------------------------
/*!
 * "(x, y, z)"̌`̕񂩂Vec3^֕ϊ
 *  - (x)ƂȂĂ(x, x, x)ƂD
 * @param[in] s 
 * @param[out] v l
 * @return vfLq
 */
static inline int StringToVec3(const string &s, Vec3 &v)
{
	int vcount = 0;
	size_t pos;
	v = Vec3(0.0);
	if((pos = s.find('(')) != string::npos){
		while(pos != string::npos && vcount < 3){
			size_t pos1 = pos;
			if((pos1 = s.find(',', pos+1)) != string::npos){
				v[vcount] = atof(s.substr(pos+1, (pos1-(pos+1))).c_str());
				vcount++;
				pos = pos1;
			}
			else if((pos1 = s.find(')', pos+1)) != string::npos){
				v[vcount] = atof(s.substr(pos+1, (pos1-(pos+1))).c_str());
				vcount++;
				break;
			}
			else{
				break;
			}
		}
	}
	if(vcount < 3){
		for(int i = vcount; i < 3; ++i){
			v[i] = v[vcount-1];
		}
	}

	return vcount;
}

/*!
 * 񂩂l(Vec3)擾
 * @param[out] val l
 * @param[in] str 
 * @param[in] rel trueŃV~[VԂ̑傫ɑ΂WƂČvZ
 * @param[in] cen V~[VԂ̒SW
 * @param[in] ext V~[VԂ̑傫(eӂ̒1/2)
 */
static inline void GetValueFromString(Vec3 &val, const string &str, bool rel = false, Vec3 cen = Vec3(0.0), Vec3 ext = Vec3(1.0))
{
	int n = StringToVec3(str, val);
	if(rel){
		val = cen+(n == 1 ? RXFunc::Min3(ext) : ext)*val;
	}
}
static inline void GetValueFromString(Vec2 &val, const string &str, bool rel = false, Vec2 cen = Vec2(0.0), Vec2 ext = Vec2(1.0))
{
	Vec3 val3;
	int n = StringToVec3(str, val3);
	val[0] = val3[0]; val[1] = val3[1];
	if(rel){
		val = cen+(n == 1 ? RXFunc::Min2(ext) : ext)*val;
	}
}

/*!
 * 񂩂l(double)擾
 * @param[out] val l
 * @param[in] str 
 * @param[in] rel trueŃV~[VԂ̑傫ɑ΂WƂČvZ
 * @param[in] cen V~[VԂ̒SW
 * @param[in] ext V~[VԂ̑傫(eӂ̒1/2)
 */
static inline void GetValueFromString(double &val, const string &str, bool rel = false, Vec3 cen = Vec3(0.0), Vec3 ext = Vec3(1.0))
{
	val = atof(str.c_str());
	if(rel){
		val = RXFunc::Min3(ext)*val;
	}
}

/*!
 * t@C
 * @param head : {t@C
 * @param ext  : gq
 * @param n    : A
 * @param d    : AԌ
 * @return t@C
 */
static inline string CreateFileName(const string &head, const string &ext, int n, const int &d)
{
	string file_name = head;
	int dn = d-1;
	if(n > 0){
		dn = (int)(log10((double)n))+1;
	}
	else if(n == 0){
		dn = 1;
	}
	else{
		n = 0;
		dn = 1;
	}

	for(int i = 0; i < d-dn; ++i){
		file_name += "0";
	}

	file_name += RX_TO_STRING(n);
	if(!ext.empty() && ext[0] != '.') file_name += ".";
	file_name += ext;

	return file_name;
}

static inline int ExistFile(const char *fn)
{
	FILE *fp;
	if( (fp = fopen(fn, "r")) == NULL ){
		return 0;
	}
	fclose(fp);
	return 1;
}


//-----------------------------------------------------------------------------
//! rxSceneConfigNX - V[ݒt@Cǂݍ
//-----------------------------------------------------------------------------
class rxSceneConfig
{
protected:
	rxFluidSolver3D *m_pSolver;			//!< ̃\o
	rxEnviroment m_Env;					//!< ݒ
	
	string m_strCurrentScene;			//!< ݂̃V[̖O
	vector<string> m_vSceneFiles;		//!< V[t@CXg
	int m_iSceneFileNum;				//!< V[t@C̐

	vector<string> m_vSceneTitles;		//!< V[t@CXg
	int m_iCurrentSceneIdx;				//!< ݂̃V[t@C

	// V[IuWFNgǉpϐ
	Vec3 m_v3Cen, m_v3Ext;
	RXREAL m_fRad;

public:
	//! ftHgRXgN^
	rxSceneConfig() : m_pSolver(0)
	{
		m_vSceneFiles.resize(12, "");	// V[t@CXg
		m_vSceneTitles.resize(12, "");	// V[^CgXg
		m_iCurrentSceneIdx = -1;		// ݂̃V[t@C
		m_strCurrentScene = "null";		// ݂̃V[̖O
		m_iSceneFileNum = 0;

		Clear();
	}

	//! fXgN^
	~rxSceneConfig(){}
	
	//! ݒ菉
	void Clear(void)
	{
		m_pSolver = 0;
		m_Env = rxEnviroment();
	}


	//! V[^CgXg
	vector<string> GetSceneTitles(void) const { return m_vSceneTitles; }

	//! ݂̃V[
	int GetCurrentSceneIdx(void) const { return m_iCurrentSceneIdx; }

	//! V[̎擾
	rxEnviroment GetEnv(void) const { return m_Env; }

	//! \o[NX̐ݒ
	void Set(rxFluidSolver3D *x){ m_pSolver = x; }

public:
	//! lp`t̒ǉ֐
	static RXREAL LiquidBox_s(RXREAL x, RXREAL y, RXREAL z, void* ptr)
	{
		return ((rxSceneConfig*)(ptr))->LiquidBox(x, y, z);
	}
	RXREAL LiquidBox(RXREAL x, RXREAL y, RXREAL z)
	{
		Vec3 minp = m_v3Cen-m_v3Ext;
		Vec3 maxp = m_v3Cen+m_v3Ext;
		return (minp[0] < x && x < maxp[0] && minp[1] < y && y < maxp[1] && minp[2] < z && z < maxp[2]) ? -1.0 : 1.0;
	}

	//! `t̒ǉ֐
	static RXREAL LiquidSphere_s(RXREAL x, RXREAL y, RXREAL z, void* ptr)
	{
		return ((rxSceneConfig*)(ptr))->LiquidSphere(x, y, z);
	}
	RXREAL LiquidSphere(RXREAL x, RXREAL y, RXREAL z)
	{
		Vec3 p(x, y, z);
		return (RXREAL)(norm(m_v3Cen-p)-m_fRad);
	}

public:
	/*!
	 * ݒ肩p[^̓ǂݍ
	 * @param[in] names ږXg
	 * @param[in] values lXg
	 * @param[in] n Xg̃TCY
	 * @param[in] header wb_
	 */
	static void SetSpace_s(string *names, string *values, int n, string header, void *x)
	{
		((rxSceneConfig*)x)->SetSpace(names, values, n, header);
	}
	void SetSpace(string *names, string *values, int n, string header)
	{
		rxEnviroment env;
		double len = 1.0;
		for(int i = 0; i < n; ++i){
			if(names[i] == "len")			   len = atof(values[i].c_str());
			else if(names[i] == "nx")		   env.Nx = atoi(values[i].c_str());
			else if(names[i] == "ny")		   env.Ny = atoi(values[i].c_str());
			else if(names[i] == "nz")		   env.Nz = atoi(values[i].c_str());
			else if(names[i] == "dt")		   env.Dt = atof(values[i].c_str());
			else if(names[i] == "gravity") 	   env.Gravity = atof(values[i].c_str());
			else if(names[i] == "v_viscosity") env.Visc = atof(values[i].c_str());
			else if(names[i] == "v_force")     env.Frc = atof(values[i].c_str());
			else if(names[i] == "v_vortex")    env.VC = atof(values[i].c_str());
			else if(names[i] == "d_diffuse")   env.Diff = atof(values[i].c_str());
			else if(names[i] == "d_source")    env.Src = atof(values[i].c_str());
			else if(names[i] == "t_diffuse")   env.DiffA = atof(values[i].c_str());
			else if(names[i] == "t_beta")	   env.Beta = atof(values[i].c_str());
			else if(names[i] == "t_ambient")   env.Tamb = atof(values[i].c_str());
			else if(names[i] == "t_max")       env.Tmax = atof(values[i].c_str());
		}
		env.H = len/env.Nx;
		m_Env = env;

	}

	//! t : `
	static void SetLiquidBox_s(string *names, string *values, int n, string header, void *x)
	{
		((rxSceneConfig*)x)->SetLiquidBox(names, values, n, header);
	}
	void SetLiquidBox(string *names, string *values, int n, string header)
	{
		bool rel = (header.find("(r)") != string::npos);
		Vec3 cen(0.0), ext(0.0);
		for(int i = 0; i < n; ++i){
			if(names[i] == "cen")      GetValueFromString(cen, values[i], rel, Vec3(0.5), Vec3(0.5));
			else if(names[i] == "ext") GetValueFromString(ext, values[i], rel, Vec3(0.5), Vec3(0.5));
		}
		m_v3Cen = cen;
		m_v3Ext = ext;
		//m_pSolver->Add(&rxSceneConfig::LiquidBox_s, this);
		RXCOUT << "set liquid box : " << cen << ", " << ext << endl;
	}

	//! t : 
	static void SetLiquidSphere_s(string *names, string *values, int n, string header, void *x)
	{
		((rxSceneConfig*)x)->SetLiquidSphere(names, values, n, header);
	}
	void SetLiquidSphere(string *names, string *values, int n, string header)
	{
		bool rel = (header.find("(r)") != string::npos);
		Vec3 cen(0.0);
		double rad = 0.0;
		for(int i = 0; i < n; ++i){
			if(names[i] == "cen")      GetValueFromString(cen, values[i], rel, Vec3(0.5), Vec3(0.5));
			else if(names[i] == "rad") GetValueFromString(rad, values[i], rel, Vec3(0.5), Vec3(0.5));
		}
		m_v3Cen = cen;
		m_fRad = rad;
		//m_pSolver->Add(&rxSceneConfig::LiquidSphere_s, this);
		RXCOUT << "set liquid sphere : " << cen << ", " << rad << endl;
	}

	/*!
	 * V~[VԂ̐ݒǂݍ
	 */
	bool LoadSpaceFromFile(void)
	{
		bool ok = true;
		rxINI *cfg = new rxINI();
		cfg->SetHeaderFunc("space", &rxSceneConfig::SetSpace_s, this);
		if(!(cfg->Load(m_strCurrentScene))){
			RXCOUT << "Failed to open the " << m_strCurrentScene << " file!" << endl;
			ok = false;
		}
		delete cfg;
		return ok;
	}
	
	/*!
	 * IuWFNgݒǂݍ
	 */
	bool LoadSceneFromFile(void)
	{
		bool ok = true;
		rxINI *cfg = new rxINI();
		cfg->SetHeaderFunc("liquid box",		&rxSceneConfig::SetLiquidBox_s, this);
		cfg->SetHeaderFunc("liquid box (r)",    &rxSceneConfig::SetLiquidBox_s, this);
		cfg->SetHeaderFunc("liquid sphere",		&rxSceneConfig::SetLiquidSphere_s, this);
		cfg->SetHeaderFunc("liquid sphere (r)", &rxSceneConfig::SetLiquidSphere_s, this);
		if(!(cfg->Load(m_strCurrentScene))){
			RXCOUT << "Failed to open the " << m_strCurrentScene << " file!" << endl;
			ok = false;
		}
		delete cfg;
		return ok;
	}

	/*!
	 * w肵tH_ɂݒt@C̐ƃV[^Cgǂݎ
	 * @param[in] dir ݒt@CtH_(w肵ȂΎstH_)
	 */
	void ReadSceneFiles(string dir = "")
	{
		m_vSceneFiles.resize(12, "");	// V[t@CXg
		m_vSceneTitles.resize(12, "");	// V[^CgXg

		ifstream scene_ifs;
		string scene_fn = "null";
		for(int i = 1; i <= 12; ++i){
			if(ExistFile((scene_fn = CreateFileName(dir+"smoke_scene_", ".cfg", i, 1)).c_str())){
				RXCOUT << "scene " << i << " : " << scene_fn << endl;
				m_vSceneFiles[i-1] = scene_fn;
				m_vSceneTitles[i-1] = scene_fn.substr(0, 11);

				// V[^Cg̓ǂݎ
				scene_ifs.open(scene_fn.c_str(), ios::in);
				string title_buf;
				getline(scene_ifs, title_buf);
				if(!title_buf.empty() && title_buf[0] == '#'){
					m_vSceneTitles[i-1] = title_buf.substr(2, title_buf.size()-2);
				}
				scene_ifs.close();

				m_iSceneFileNum++;
			}
		}

		if(m_iSceneFileNum){
			SetCurrentScene(0);
		}
	}

	/*!
	 * Jg̃V[ݒ
	 * @param[in] V[CfbNX
	 */
	bool SetCurrentScene(int idx)
	{
		if(idx < 0 || idx >= m_iSceneFileNum || m_vSceneFiles[idx] == ""){
			cout << "There is no scene files!" << endl;
			return false;
		}
		m_iCurrentSceneIdx = idx;
		m_strCurrentScene = m_vSceneFiles[m_iCurrentSceneIdx];
		return true;
	}

	/*!
	 * ^CgJg̃V[ݒ
	 * @param[in] label V[^Cg
	 */
	bool SetCurrentSceneFromTitle(const string label)
	{
		int scene = 0;
		while(label.find(m_vSceneTitles[scene]) == string::npos) scene++;

		if(m_iCurrentSceneIdx != -1 && m_vSceneFiles[scene] != ""){
			RXCOUT << "scene " << scene+1 << " : " << m_vSceneFiles[scene] << endl;
			m_iCurrentSceneIdx = scene;
			m_strCurrentScene = m_vSceneFiles[scene];
			return true;
		}

		return false;
	}


	vector<string> GetSceneTitleList(void){ return m_vSceneTitles; }
};


#endif // #ifndef _RX_SPH_CONFIG_H_