/*! @file rx_numerical.h
	
	@brief lvZ[`
 
	@author Makoto Fujisawa
	@date  
*/


#ifndef _RX_NUMERICAL_H_
#define _RX_NUMERICAL_H_

#pragma warning (disable: 4996)
#pragma warning (disable: 4819)
#pragma warning (disable: 4786)
#pragma warning (disable: 4244)


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cstdio>
#include <cstdlib>
#include <cmath>

#include <cmath>
#include <iostream>
#include <fstream>

#include <string>
#include <vector>

using namespace std;

namespace RXNumerical
{
	// e덷Ȃ
	const double RX_FEQ_EPS = 1.0e-10;
	const double RX_EPS = 1.0e-8;
	
	inline int ix3(int i, int j, int k, int nx, int ny){ return k*nx*ny+j*nx+i; }
	inline int ix(int i, int j, int nx){ return i+nx*j; }

	template<class T> 
	void dump(const char *fn, int w, int h, T *data, const char *comment, bool app)
	{
		if(!data) return;

		ofstream fo;
		if(app){
			fo.open(fn, ios::app);
		}
		else{
			fo.open(fn);
		}

		fo << "[" << comment << "]" << endl;
		for(int j = h-1; j >= 0; --j){
			for(int i = 0; i < w; ++i){
				int idx = j*w+i;
				fo << data[idx] << " ";
			}
			fo << endl;
		}
		fo << endl;

		fo.close();
	}

	/*!
	 * ICCG@pאڃObhT[`
	 * @param[out] neigh ߖTʒu\z
	 * @param[in] nx,ny Obh
	 * @return zTCY(size_m)
	 */
	static int SetNeighbors2D(int **neigh, int nx, int ny)
	{
		int i, j, idx;
		for(i = 0; i < nx; ++i){
			for(j = 0; j < ny; ++j){
				idx = ix(i, j, nx);

				int cn = 0;
				if(i > 0   ){ neigh[idx][cn+1] = ix(i-1, j, nx); cn++; }
				if(i < nx-1){ neigh[idx][cn+1] = ix(i+1, j, nx); cn++; }
				if(j > 0   ){ neigh[idx][cn+1] = ix(i, j-1, nx); cn++; }
				if(j < ny-1){ neigh[idx][cn+1] = ix(i, j+1, nx); cn++; }

				neigh[idx][0] = cn;	// אڃObh̐

				for(int l = cn+1; l < 5; ++l){
					neigh[idx][l] = -1;
				}
			}
		}

		return nx*ny;
	}

	/*!
	 * ICCG@pאڃObhT[`
	 * @param[out] neigh ߖTʒu\z
	 * @param[in] nx,ny Obh
	 * @return zTCY(size_m)
	 */
	static int SetNeighbors3D(int **neigh, int nx, int ny, int nz)
	{
		for(int i = 0; i < nx; ++i){
			for(int j = 0; j < ny; ++j){
				for(int k = 0; k < nz; ++k){
					int idx = ix3(i, j, k, nx, ny);

					int cn = 0;
					if(i > 0   ){ neigh[idx][cn+1] = ix3(i-1, j, k, nx, ny); cn++; }
					if(i < nx-1){ neigh[idx][cn+1] = ix3(i+1, j, k, nx, ny); cn++; }
					if(j > 0   ){ neigh[idx][cn+1] = ix3(i, j-1, k, nx, ny); cn++; }
					if(j < ny-1){ neigh[idx][cn+1] = ix3(i, j+1, k, nx, ny); cn++; }
					if(k > 0   ){ neigh[idx][cn+1] = ix3(i, j, k-1, nx, ny); cn++; }
					if(k < nz-1){ neigh[idx][cn+1] = ix3(i, j, k+1, nx, ny); cn++; }

					neigh[idx][0] = cn;	// אڃObh̐

					for(int l = cn+1; l < 7; ++l){
						neigh[idx][l] = -1;
					}
				}
			}
		}

		return nx*ny*nz;
	}

	/*!
	 * sƃxNg̊|Z
	 * @param[out] y ʂi[Rei
	 * @param[in] mat s
	 * @param[in] x xNg
	 * @param[in] f xZbg֐
	 * @param[in] neigh ߖTʒu\z
	 * @param[in] n zTCY
	 */
	template<class T> 
	static void multMV(T *y, T **mat, T *x, int **neigh, int n)
	{
		for(int i = 0; i < n; ++i){
			y[i] = mat[i][0]*x[i];
			for(int l = 1; l <= neigh[i][0]; ++l){
				int j = neigh[i][l];	// ߖTCfbNX
				if(j >= 0) y[i] += mat[i][l]*x[j];
			}			
		}
	}

	/*!
	 * z񓯎m̓
	 * @param[in] x1 z1
	 * @param[in] x2 z2
	 * @param[in] f xZbg֐
	 * @param[in] n zTCY
	 */
	template<class T> 
	static double dotProduct(const T *x1, const T *x2, int n)
	{
		T mul = 0.0;
		for(int i = 0; i < n; ++i){
			mul += x1[i]*x2[i];
		}
		return mul;
	}

	/*!
	 * z̃Rs[
	 * @param[in] dst Rs[z
	 * @param[in] src Rs[z
	 * @param[in] n zTCY
	 */
	template<class T> 
	static void copy(T *dst, const T *src, int n)
	{
		for(int i = 0; i < n; ++i){
			dst[i] = src[i];
		}
	}

	/*!
	 * z̃Rs[
	 * @param[in] dst Rs[z
	 * @param[in] val Rs[l
	 * @param[in] n zTCY
	 */
	template<class T> 
	static void copy(T *dst, const T val, int n)
	{
		for(int i = 0; i < n; ++i){
			dst[i] = val;
		}
	}

	/*!
	 * KEX@(z)
	 * @param[out] y xNg
	 * @param[in] mat s(Ӎ)
	 * @param[in] xx EӍ
	 * @param[in] neigh ߖTʒu\z
	 * @param[in] size_m zTCY
	 */
	template<class T> 
	static void solverLL(T *y, T **mat, T *xx, int **neigh, int n)
	{
		int i, j, l;
		T *x;
		x = new T[n];

		for(i = 0; i < n; ++i) x[i] = xx[i];


		// Oi
		for(i = 0; i < n; ++i){
			for(l = 1; l <= neigh[i][0]; ++l){
				j = neigh[i][l];	// ߖTCfbNX

				if(j < 0 || j > i) continue;

				x[i] = x[i]-mat[i][l]*y[j];
			}
			y[i] = x[i]/mat[i][0];
		}

		for(i = 0; i < n; ++i) x[i] = y[i];

		// ޑ
		for(i = n-1; i >= 0; --i){
			for(l = 1; l <= neigh[i][0]; ++l){
				j = neigh[i][l];	// ߖTCfbNX

				if(j < 0 || j < i) continue;

				x[i] = x[i]-mat[i][l]*y[j];
			}
			y[i] = x[i]/mat[i][0];
		}

		delete [] x;
	}


	/*!
	 * sSRXL[(incomplete Cholesky decomposition)(z)
	 * @param[in] poiss ̍s
	 * @param[out] ic   sSRXL[̍s
	 * @param[in] neigh ߖTʒu\z
	 * @param[in] size_m zTCY
	 */
	template<class T> 
	static void IcDecomp(T **poiss, T **ic, int **neigh, int n)
	{
		int i, j, k, mi, mj, ki, kj;
		T sum;

		for(i = 0; i < n; ++i){
			sum = poiss[i][0];
			for(k = 1; k <= neigh[i][0]; ++k){
				j = neigh[i][k];	// i̋ߖTObh̃CfbNX
				if(j < 0 || j > i) continue;

				sum = sum-ic[i][k]*ic[i][k];
			}
			ic[i][0] = sqrt(sum);

			for(k = 1; k <= neigh[i][0]; ++k){
				j = neigh[i][k];	// i̋ߖTObh̃CfbNX
				if(j < 0 || j < i) continue;

				sum = poiss[i][k];
				for(mj = 1; mj <= neigh[j][0]; ++mj){
					kj = neigh[j][mj];	// j̋ߖTObh̃CfbNX
					if(kj < 0 || kj >= i) continue;

					for(mi = 1; mi <= neigh[i][0]; ++mi){
						ki = neigh[i][mi];	// i̋ߖTObh̃CfbNXmi

						if(ki == kj){
							sum = sum-ic[i][mi]*ic[j][mj];
							break;
						}
					}
				}
				ic[i][k] = sum/ic[i][0];

				for(mj = 1; mj <= neigh[j][0]; ++mj){
					kj = neigh[j][mj];	// j̋ߖTObh̃CfbNX

					if(i == kj){
						ic[j][mj] = ic[i][k];
						break;
					}
				}
			}
		}
	}

	/*!
	 * sSRXL[tz(ICCG)\o(z)
	 * @param[in] poiss ̍s
	 * @param[out] x i[
	 * @param[in] b EӍ
	 * @param[in] ic sSRXL[s
	 * @param[inout] max_iter ő唽
	 * @param[inout] tol e덷
	 * @param[in] neigh ߖTʒu\z
	 * @param[in] size_m zTCY
	 */
	template<class T> 
	static int pcgSolver(T **poiss, T *x, T *b, T **ic, int **neigh, int n, int &max_iter, T &tol)
	{
		int k = 0;
		T *r, *p, *q, *s;

		r = new T[n];
		p = new T[n];
		q = new T[n];
		s = new T[n];
			
		copy(x, (T)(0.0), n);
		copy(s, (T)(0.0), n);
		copy(r, (T)(0.0), n);
		copy(q, (T)(0.0), n);

		multMV(s, poiss, x, neigh, n);

		for(int i = 0; i < n; ++i){
			r[i] = b[i]-s[i];
		}

		solverLL(q, ic, r, neigh, n);
		copy(p, q, n);

		T rqo = dotProduct(r, q, n); // rqo = r*q
		for(k = 0; k < max_iter; k++){
			multMV(s, poiss, p, neigh, n);	// s = poiss*p
			T ps = dotProduct(p, s, n);			// ps = p*s

			// xAcr̍XV
			T alpha = rqo/ps;
			for(int i = 0; i < n; ++i){
				x[i] = x[i]+alpha*p[i];
				r[i] = r[i]-alpha*s[i];
			}

			solverLL(q, ic, r, neigh, n);
			T rqn = dotProduct(r, q, n);	// rqn = r*q

			// XVc2敽ςG[lƂĎɎgp
			if(rqn/n < tol) break;


			T beta = rqn/rqo;
			rqo = rqn;
			for(int i = 0; i < n; ++i){
				p[i] = q[i]+beta*p[i];
			}
		}

		delete [] r;
		delete [] p;
		delete [] q;
		delete [] s;

		max_iter = k;
		return k;
	}

	
	/*!
	 * z(CG)\o
	 * @param[in] poiss ̍s
	 * @param[out] x i[
	 * @param[in] b EӍ
	 * @param[inout] max_iter ő唽
	 * @param[inout] tol e덷
	 * @param[in] neigh ߖTʒu\z
	 * @param[in] size_m zTCY
	 */
	template<class T> 
	static int cgSolver(T **A, T *x, T *b, int **neigh, int n, int &max_iter, T &tol)
	{
		T *r, *p, *s;
		r = new T[n];
		p = new T[n];
		s = new T[n];

		copy(x, (T)(0.0), n);
		copy(r, b, n);
		copy(p, r, n);
		copy(s, p, n);

		T rr0 = dotProduct(p, r, n);

		int k = 0;
		for(k = 0; k < max_iter; k++){
			// s = Ap ̌vZ
			multMV(p, A, s, neigh, n);

			// alpha = r*r/(P*AP)̌vZ
			T alpha = rr0/dotProduct(p, s, n);

			// xAcr̍XV
			for(int i = 0; i < n; ++i){
				x[i] += alpha*s[i];
				r[i] -= alpha*p[i];
			}

			// XVc2敽ςG[lƂĎɎgp
			T rr1 = dotProduct(r, r, n);
			if(rr1/n < tol) break;

			copy(p, r, n);

			T beta = rr1/rr0;
			for(int i = 0; i < n; ++i){
				s[i] = p[i]+beta*s[i];
			}

			rr0 = rr1;
		}

		delete [] r;
		delete [] p;
		delete [] s;

		tol = rr0;
		max_iter = k;
		return k;
	}



} // namespace numerical

#endif // _RX_NUMERICAL_H_