/*!
  @file rx_common.h
	
  @brief vŐʃwb_
 
  @author Makoto Fujisawa
  @date 2013
*/

#ifndef _RX_COMMON_H_
#define _RX_COMMON_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
// CW
#include <cmath>
#include <cstdio>
#include <direct.h>

#include <iostream>
#include <fstream>
#include <sstream>

// STL
#include <vector>
#include <string>


// OpenGL
#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include "rx_utility.h"

using namespace std;


//-----------------------------------------------------------------------------
// `
//-----------------------------------------------------------------------------
typedef unsigned char byte;
typedef unsigned int uint;

#ifndef RXREAL
#define RXREAL float
#endif


//-----------------------------------------------------------------------------
// Xg[o
//-----------------------------------------------------------------------------
#define RXCOUT std::cout


//-----------------------------------------------------------------------------
// lXȊ֐
//-----------------------------------------------------------------------------
/*!
 * pXgqɂĎo
 * @param[in] path t@CpX
 * @return ()gq
 */
inline string GetExtension(const string &path)
{
	string ext;
	size_t pos1 = path.rfind('.');
	if(pos1 != string::npos){
		ext = path.substr(pos1+1, path.size()-pos1);
		string::iterator itr = ext.begin();
		while(itr != ext.end()){
			*itr = tolower(*itr);
			itr++;
		}
		itr = ext.end()-1;
		while(itr != ext.begin()){	// pX̍Ō\0Xy[XƂ̑΍
			if(*itr == 0 || *itr == 32){
				ext.erase(itr--);
			}
			else{
				itr--;
			}
		}
	}
 
	return ext;
}

/*!
 * pXt@C菜pX𒊏o
 * @param[in] path pX
 * @return tH_pX
 */
inline string GetFolderPath(const string &path)
{
	size_t pos1;
 
	pos1 = path.rfind('\\');
	if(pos1 != string::npos){
		return path.substr(0, pos1+1);
		
	}
 
	pos1 = path.rfind('/');
	if(pos1 != string::npos){
		return path.substr(0, pos1+1);
	}
 
	return "";
}

/*!
 * fBNg쐬(KwΉ)
 * @param[in] dir 쐬fBNgpX
 * @return 1,s0 (fBNgłɂꍇ1Ԃ)
 */
static int MkDir(string dir)
{
	if(_mkdir(dir.c_str()) != 0){
		char cur_dir[512];
		_getcwd(cur_dir, 512);	// JgtH_mۂĂ
		if(_chdir(dir.c_str()) == 0){	// chdirŃtH_݃`FbN
			cout << "MkDir : " << dir << " is already exist." << endl;
			_chdir(cur_dir);	// JgtH_ɖ߂
			return 1;
		}
		else{
			size_t pos = dir.find_last_of("\\/");
			if(pos != string::npos){	// Kw̉\L
				int parent = MkDir(dir.substr(0, pos+1));	// efBNgċAIɍ쐬
				if(parent){
					if(_mkdir(dir.c_str()) == 0){
						return 1;
					}
					else{
						return 0;
					}
				}
			}
			else{
				return 0;
			}
		}
	}

	return 1;
}

//! Of[VF̐
template<class T> 
inline void RX_COLOR_RAMP(T t, T *r)
{
	const int ncolors = 7;
	T c[ncolors][3] = {	{1.0, 0.0, 0.0},
						{1.0, 0.5, 0.0},
						{1.0, 1.0, 0.0},
						{0.0, 1.0, 0.0},
						{0.0, 1.0, 1.0},
						{0.0, 0.0, 1.0},
						{1.0, 0.0, 1.0} };
	t = t*(ncolors-1);
	int i = (int)t;
	T u = t-floor(t);
	r[0] = RX_LERP(c[i][0], c[i+1][0], u);
	r[1] = RX_LERP(c[i][1], c[i+1][1], u);
	r[2] = RX_LERP(c[i][2], c[i+1][2], u);
}

/*!
 * ->->->ƕωOf[VF
 * @param[out] col ꂽF
 * @param[in] x l
 * @param[in] xmin ŏl
 * @param[in] xmax ől
 */
inline void Gradation(double col[3], double x, const double xmin = 0.0, const double xmax = 1.0)
{
	double l = xmax-xmin;
	if(fabs(l) < 1e-10) return;
	
	const int ncolors = 7;
	double base[ncolors][3] = { {0.0, 0.0, 0.0},
								{0.0, 0.0, 1.0},
								{0.0, 1.0, 1.0},
								{0.0, 1.0, 0.0},
								{1.0, 1.0, 0.0},
								{1.0, 0.0, 0.0},
								{1.0, 1.0, 1.0} };
	x = RX_CLAMP(((x-xmin)/l), 0.0, 1.0)*(ncolors-1);
	int i = (int)x;
	double dx = x-floor(x);
	col[0] = RX_LERP(base[i][0], base[i+1][0], dx);
	col[1] = RX_LERP(base[i][1], base[i+1][1], dx);
	col[2] = RX_LERP(base[i][2], base[i+1][2], dx);
}


//-----------------------------------------------------------------------------
// Ԋ֐
//-----------------------------------------------------------------------------
inline int GET_INDEX(int i, int j, int k, int nx, int ny){ return k*nx*ny+j*nx+i; }
inline int GET_INDEX_2D(int i, int j, int nx){ return j*nx+i; }

/*!
 * `Ԃɂl̎擾
 * @param[in] x,y W
 * @param[in] f tB[hl(double)i[Ăz
 * @param[in] dx,dy Obh
 * @param[in] nx,ny Obh
 * @return `Ԓl
 */
inline double LERP2D(RXREAL x, RXREAL y, const RXREAL *f, RXREAL dx, RXREAL dy, int nx, int ny)
{

	int i0, i1, j0, j1;
	RXREAL s0, s1, t0, t1;

	// (x,y,z)̃Obhʒu
	i0 = (int)(x/dx); i1 = i0+1;
	j0 = (int)(y/dy); j1 = j0+1;

	s1 = (x/dx-i0);
	t1 = (y/dy-j0);

	// ObhOO
	if(i0 < 0   ){ i0 = 0; i1 = 1; s1 = 0.0; }
	if(j0 < 0   ){ j0 = 0; j1 = 1; t1 = 0.0; }
	if(i1 > nx-1){ i0 = nx-2; i1 = nx-1; s1 = 1.0; }
	if(j1 > ny-1){ j0 = ny-2; j1 = ny-1; t1 = 1.0; }

	s0 = 1.0-s1; 
	t0 = 1.0-t1; 

	// (x0,y0)ɂl`Ԃŋ߂
	return s0*(t0*f[GET_INDEX_2D(i0, j0, nx)]+t1*f[GET_INDEX_2D(i0, j1, nx)])+
		   s1*(t0*f[GET_INDEX_2D(i1, j0, nx)]+t1*f[GET_INDEX_2D(i1, j1, nx)]);
}
/*!
 * `Ԃɂl̎擾(3D)
 * @param[in] x,y,z W
 * @param[in] f tB[hl(double)i[Ăz
 * @param[in] dx,dy,dz Obh
 * @param[in] nx,ny,nz Obh
 * @return `Ԓl
 */
inline double LERP3D(RXREAL x, RXREAL y, RXREAL z, RXREAL *f, 
					 RXREAL dx, RXREAL dy, RXREAL dz, int nx, int ny, int nz)
{
	int i0, i1, j0, j1, k0, k1;
	RXREAL s0, s1, t0, t1, u0, u1;

	// (x,y,z)̃Obhʒu
	i0 = (int)(x/dx); i1 = i0+1;
	j0 = (int)(y/dy); j1 = j0+1;
	k0 = (int)(z/dz); k1 = k0+1;

	s1 = (x/dx-i0);
	t1 = (y/dy-j0);
	u1 = (z/dz-k0);

	// ObhOO
	if(i0 < 0   ){ i0 = 0; i1 = 1; s1 = 0.0; }
	if(j0 < 0   ){ j0 = 0; j1 = 1; t1 = 0.0; }
	if(k0 < 0   ){ k0 = 0; k1 = 1; u1 = 0.0; }
	if(i1 > nx-1){ i0 = nx-2; i1 = nx-1; s1 = 1.0; }
	if(j1 > ny-1){ j0 = ny-2; j1 = ny-1; t1 = 1.0; }
	if(k1 > nz-1){ k0 = nz-2; k1 = nz-1; u1 = 1.0; }

	s0 = 1.0-s1; 
	t0 = 1.0-t1; 
	u0 = 1.0-u1;

	// (x0,y0,z0)ɂl`Ԃŋ߂
	return u0*(s0*(t0*f[GET_INDEX(i0, j0, k0, nx, ny)]+t1*f[GET_INDEX(i0, j1, k0, nx, ny)])+ \
			   s1*(t0*f[GET_INDEX(i1, j0, k0, nx, ny)]+t1*f[GET_INDEX(i1, j1, k0, nx, ny)]))+ \
		   u1*(s0*(t0*f[GET_INDEX(i0, j0, k1, nx, ny)]+t1*f[GET_INDEX(i0, j1, k1, nx, ny)])+ \
			   s1*(t0*f[GET_INDEX(i1, j0, k1, nx, ny)]+t1*f[GET_INDEX(i1, j1, k1, nx, ny)]));

}


//-----------------------------------------------------------------------------
// s
//-----------------------------------------------------------------------------
/*!
 * IC[p]s𐶐
 * @param[out] m ]s
 * @param[in] pitch,yaw,roll IC[j
 */
inline void EulerToMatrix(double *m, double pitch, double yaw, double roll)
{
	yaw   = RX_TO_RADIANS(yaw);
	pitch = RX_TO_RADIANS(pitch);
	roll  = RX_TO_RADIANS(roll);
 
	double cy = cos(yaw); 
	double sy = sin(yaw); 
	double cp = cos(pitch); 
	double sp = sin(pitch); 
	double cr = cos(roll);
	double sr = sin(roll);
 
	double cc = cy*cr; 
	double cs = cy*sr; 
	double sc = sy*cr; 
	double ss = sy*sr;
 
	m[0]  = cc+sp*ss;
	m[1]  = cs-sp*sc;
	m[2]  = -sy*cp;
	m[3]  = 0.0;
	
	m[4]  = -cp*sr;
	m[5]  = cp*cr; 
	m[6]  = -sp;
	m[7]  = 0.0;
	
	m[8]  = sc-sp*cs;
	m[9]  = ss+sp*cc;
	m[10] = cy*cp;
	m[11] = 0.0;
 
	m[12] = 0.0;
	m[13] = 0.0;
	m[14] = 0.0;
	m[15] = 1.0;
}

/*!
 * 3x3s񂩂OpenGL̕ϊs擾
 * @param[in] 
 * @param[out] 
 * @return 
 */
inline void GetGLMatrix(RXREAL m[9], GLfloat glm[16])
{
	glm[0]  = m[0];
	glm[1]  = m[3]; 
	glm[2]  = m[6]; 
	glm[3]  = 0.0f; 
	glm[4]  = m[1]; 
	glm[5]  = m[4]; 
	glm[6]  = m[7]; 
	glm[7]  = 0.0f; 
	glm[8]  = m[2]; 
	glm[9]  = m[5]; 
	glm[10] = m[8]; 
	glm[11] = 0.0f; 
	glm[12] = 0.0f; 
	glm[13] = 0.0f; 
	glm[14] = 0.0f; 
	glm[15] = 1.0f;
}

/*!
 * 3x3s̍s񎮂̌vZ
 *  | m[0] m[1] m[2] |
 *  | m[3] m[4] m[5] |
 *  | m[6] m[7] m[8] |
 * @param[in] m ̍s
 * @return s񎮂̒l
 */
inline RXREAL CalDeterminant3x3(const RXREAL *m)
{
	return m[0]*m[4]*m[8]+m[3]*m[7]*m[2]+m[6]*m[1]*m[5]
	      -m[0]*m[7]*m[5]-m[6]*m[4]*m[2]-m[3]*m[1]*m[8];
}
/*!
 * 3x3s̋tšvZ
 *  | m[0] m[1] m[2] |
 *  | m[3] m[4] m[5] |
 *  | m[6] m[7] m[8] |
 * @param[in] m ̍s
 * @param[out] invm ts
 */
inline void CalInverse3x3(const RXREAL *m, RXREAL *invm)
{
	RXREAL d = CalDeterminant3x3(m);

	if(d == 0) d = 1;

	RXREAL inv_det = 1.0/d;
	invm[0] = inv_det*(m[4]*m[8]-m[5]*m[7]);
	invm[1] = inv_det*(m[2]*m[7]-m[1]*m[8]);
	invm[2] = inv_det*(m[1]*m[5]-m[2]*m[4]);
	
	invm[3] = inv_det*(m[5]*m[6]-m[3]*m[8]);
	invm[4] = inv_det*(m[0]*m[8]-m[2]*m[6]);
	invm[5] = inv_det*(m[2]*m[3]-m[0]*m[5]);
	
	invm[6] = inv_det*(m[3]*m[7]-m[4]*m[6]);
	invm[7] = inv_det*(m[1]*m[6]-m[0]*m[7]);
	invm[8] = inv_det*(m[0]*m[4]-m[1]*m[3]);
}

//-----------------------------------------------------------------------------
// OpenGL`
//-----------------------------------------------------------------------------
/*!
 * xyz`(x:,y:,z:)
 * @param[in] len ̒
 */
static int DrawAxis(float len, float line_width = 5.0)
{
	glLineWidth((GLfloat)line_width);

	// x
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(len, 0.0, 0.0);
	glEnd();

	// y
	glColor3f(0.0, 1.0, 0.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, len, 0.0);
	glEnd();

	// z
	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINES);
	glVertex3d(0.0, 0.0, 0.0);
	glVertex3d(0.0, 0.0, len);
	glEnd();

	return 1;
}

/*! 
 * ̂̃C[t[`
 * @param[in] min ŏWl
 * @param[in] max őWl
 * @param[in] color `F
 */
static void DrawWireCuboid(const Vec3 &minp, const Vec3 &maxp)
{
	glPushMatrix();

	glBegin(GL_LINES);
	
	// xs
	glVertex3f(minp[0], minp[1], minp[2]);	glVertex3f(maxp[0], minp[1], minp[2]);
	glVertex3f(minp[0], minp[1], maxp[2]); glVertex3f(maxp[0], minp[1], maxp[2]);
	glVertex3f(minp[0], maxp[1], minp[2]);	glVertex3f(maxp[0], maxp[1], minp[2]);
	glVertex3f(minp[0], maxp[1], maxp[2]);	glVertex3f(maxp[0], maxp[1], maxp[2]);
	
	// zs
	glVertex3f(minp[0], minp[1], minp[2]);	glVertex3f(minp[0], minp[1], maxp[2]);
	glVertex3f(minp[0], maxp[1], minp[2]);	glVertex3f(minp[0], maxp[1], maxp[2]);
	glVertex3f(maxp[0], minp[1], minp[2]);	glVertex3f(maxp[0], minp[1], maxp[2]);
	glVertex3f(maxp[0], maxp[1], minp[2]);	glVertex3f(maxp[0], maxp[1], maxp[2]);

	// zs
	glVertex3f(minp[0], minp[1], minp[2]);	glVertex3f(minp[0], maxp[1], minp[2]);
	glVertex3f(minp[0], minp[1], maxp[2]);	glVertex3f(minp[0], maxp[1], maxp[2]);
	glVertex3f(maxp[0], minp[1], minp[2]);	glVertex3f(maxp[0], maxp[1], minp[2]);
	glVertex3f(maxp[0], minp[1], maxp[2]);	glVertex3f(maxp[0], maxp[1], maxp[2]);
	
	glEnd();

	glPopMatrix();
}


/*!
 * _S̉~̃C[t[`
 * @param rad ~̔a
 * @param n 
 */
static void DrawWireCircle(const double &rad, const int &n)
{
	double t = 0.0;
	double dt = 2.0*RX_PI/(double)n;
 
	glBegin(GL_LINE_LOOP);
	do{
		glVertex3f(rad*cos(t), rad*sin(t), 0.0);
		t += dt;
	}while(t < 2.0*RX_PI);
	glEnd();
}
 
/*!
 * _S̉~̃C[t[`(XZ)
 * @param rad ~̔a
 * @param n 
 */
static void DrawWireCircleXZ(const double &rad, const int &n)
{
	double t = 0.0;
	double dt = 2.0*RX_PI/(double)n;
 
	glBegin(GL_LINE_LOOP);
	do{
		glVertex3f(rad*cos(t), 0.0, rad*sin(t));
		t += dt;
	}while(t < 2.0*RX_PI);
	glEnd();
}
 
/*!
 * ̃C[t[`
 * @param cen ̒S
 * @param rad ̔a
 * @param col `F
 */
static void DrawWireSphere(const Vec3 &cen, const float &rad, const Vec3 &col)
{
	glDisable(GL_LIGHTING);
	glPushMatrix();
	glTranslatef(cen[0], cen[1], cen[2]);
	glRotatef(90, 1.0, 0.0, 0.0);
	glColor3f(col[0], col[1], col[2]);
 
	// ܓx(x-yʂɕs)
	float z, dz;
	dz = 2.0*rad/8.0f;
	z = -(rad-dz);
	do{
		glPushMatrix();
		glTranslatef(0.0, 0.0, z);
		DrawWireCircle(sqrt(rad*rad-z*z), 32);
		glPopMatrix();
		z += dz;
	}while(z < rad);
 
	// ox(z܂ɉ])
	float t, dt;
	t = 0.0f;
	dt = 180.0/8.0;
	do{
		glPushMatrix();
		glRotatef(t,  0.0, 0.0, 1.0);
		DrawWireCircleXZ(rad, 32);
		glPopMatrix();
 
		t += dt;
	}while(t < 180);
 
	glPopMatrix();
}


#endif // #ifndef _RX_COMMON_H_