/*!
  @file rx_obj.h
	
  @brief OBJ/MTL File Input/Output
 
  @author Makoto Fujisawa
  @date   2011
*/

#ifndef _RX_OBJ_H_
#define _RX_OBJ_H_


//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------
#include "rx_mesh.h"

//-----------------------------------------------------------------------------
// Name Space
//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
// rxOBJNX̐錾 - OBJ`̓ǂݍ
//-----------------------------------------------------------------------------
class rxOBJ
{
	rxMTL m_mapMaterials;	//!< xƃf[^̃}bv
	string m_strCurrentMat;	//!< ݂̃f[^x

	//vector<rxMaterialOBJ> m_vMaterials;
	//int m_iCurrentMat;

public:
	//! RXgN^
	rxOBJ()
	{
		m_strCurrentMat = "";
	}

	//! fXgN^
	~rxOBJ()
	{
	}

	/*!
	 * OBJt@Cǂݍ
	 * @param[in] file_name t@C(tpX)
	 * @param[out] vrts _W
	 * @param[out] vnms _@
	 * @param[out] poly |S
	 * @param[out] mats ގ
	 * @param[in] triangle |S̎Op`tO
	 */
	bool Read(string file_name, vector<Vec3> &vrts, vector<Vec3> &vnms, vector<rxFace> &plys, rxMTL &mats, bool triangle = true);

	/*!
	 * OBJt@C
	 * @param[in] file_name t@C(tpX)
	 * @param[in] vrts _W
	 * @param[in] vnms _@
	 * @param[in] plys |S
	 * @param[in] mats ގ
	 */
	bool Save(string file_name, const vector<Vec3> &vrts, const vector<Vec3> &vnms, const vector<rxFace> &plys, const rxMTL &mats);

	//! ގXg̎擾
	rxMTL GetMaterials(void){ return m_mapMaterials; }

private:
	int loadFace(string &buf, vector<int> &vidxs, vector<int> &nidxs, vector<int> &tidxs);
	int loadMTL(const string &mtl_fn);
	int saveMTL(const string &mtl_fn, const rxMTL &mats);
};

/*!
 * OBJt@Cǂݍ
 * @param[in] filename objt@C̃pX
 * @param[inout] poly 3DIuWFNg
 * @param[in] cen,ext,ang ͂3D`̐Kp[^
 */
static void ReadOBJ(const string filename, rxPolygons &polys)
{
	if(!polys.vertices.empty()){
		polys.vertices.clear();
		polys.normals.clear();
		polys.faces.clear();
		polys.materials.clear();
		polys.open = 0;
	}
	rxOBJ obj;
	if(obj.Read(filename, polys.vertices, polys.normals, polys.faces, polys.materials, true)){
		if(polys.normals.empty()){
			CalVertexNormals(polys);
		}
		polys.open = 1;
	}
}



/*!
 * OBJt@C
 * @param[in] filename objt@C̃pX
 */
static void SaveOBJ(const string filename, rxPolygons &polys)
{
	rxOBJ obj;
	if(obj.Save(filename, polys.vertices, polys.normals, polys.faces, polys.materials)){
		std::cout << filename << " have been saved (OBJ)." << endl;
	}
}

/*!
 * OBJt@Cǂݍ
 * @param[in] file_name t@C(tpX)
 * @param[out] vrts _W
 * @param[out] vnms _@
 * @param[out] poly |S
 * @param[out] mats ގ
 * @param[in] triangle |S̎Op`tO
 */
inline bool rxOBJ::Read(string file_name, 
						vector<Vec3> &vrts, 
						vector<Vec3> &vnms, 
						vector<rxFace> &plys, 
						rxMTL &mats, bool triangle)
{
	ifstream file;

	file.open(file_name.c_str());
	if(!file || !file.is_open() || file.bad() || file.fail()){
		cout << "rxOBJ::Read : Invalid file specified" << endl;
		return false;
	}

	vector<Vec3> vnormals;
	vector<Vec2> vtexcoords;

	string buf;
	string::size_type comment_start = 0;
	while(!file.eof()){
		getline(file, buf);

		// '#'ȍ~̓RgƂĖ
		if( (comment_start = buf.find('#')) != string::size_type(-1) )
			buf = buf.substr(0, comment_start);

		// s̃Xy[XC^u폜
		DeleteHeadSpace(buf);

		// s͖
		if(buf.empty())
			continue;

		if(buf[0] == 'v'){
			if(buf[1] == 'n'){		// _@
				Vec3 n;
				if(StringToVec3s(buf, "vn", n)){
					vnormals.push_back(n);
				}
			}
			else if(buf[1] == 't'){	// eNX`W
				Vec2 tc;
				if(StringToVec2s(buf, "vt", tc)){
					vtexcoords.push_back(tc);
				}
			}
			else{					// _W
				Vec3 v;
				if(StringToVec3s(buf, "v", v)){
					vrts.push_back(v);
					vnms.push_back(Vec3(0.0));
				}
			}
		}
		else if(buf[0] == 'f'){		// 
			int num_face = 0;
			vector<int> vidxs, nidxs, tidxs;
			if(!(num_face = loadFace(buf, vidxs, nidxs, tidxs))) continue;

			if(triangle && num_face >= 4){
				PolyToTri(plys, vidxs, tidxs, vtexcoords, m_strCurrentMat);
			}
			else{
				rxFace face;
				face.vert_idx.resize(num_face);;
				face.material_name = m_strCurrentMat;
				face.texcoords.resize(num_face);
				bool tc = !vtexcoords.empty();
				for(int i = 0; i < num_face; ++i){
					face[i] = vidxs[i];
					if(tc){
						face.texcoords[i] = vtexcoords[tidxs[i]];
					}
					else{
						face.texcoords[i] = Vec2(0.0);
					}
				}
				plys.push_back(face);
			}

			// _@
			if(!vnormals.empty()){
				for(int i = 0; i < num_face; ++i){
					vnms[vidxs[i]] += vnormals[nidxs[i]];
				}
			}
		}
		else if(buf[0] == 'u'){		// ގ
			string mat_name;
			if(!StringToString(buf, "usemtl", mat_name)) continue;

			if(mat_name == "(null)" || m_mapMaterials.empty()) continue;

			map<string, rxMaterialOBJ>::iterator i = m_mapMaterials.find(mat_name);

			if(i != m_mapMaterials.end()){
				m_strCurrentMat = mat_name;
			}

			//cout << "usemtl " << mat_name << endl;
		}
		else if(buf[0] == 'm'){		// ގt@C
			string mat_file_name;
			if(!StringToString(buf, "mtllib", mat_file_name)) continue;

			string dir = ExtractDirPath(file_name);
			if(!dir.empty()){
				mat_file_name = dir+"\\"+mat_file_name;
			}
			//cout << "mtllib " << mat_file_name << endl;

			if(loadMTL(mat_file_name) == 0){
				//cout << "fail to read " << mat_file_name << endl;
				continue;
			}

		}
	}

	if(vnormals.empty()){
		vnms.clear();
	}

	if(!vnms.empty()){
		for(int i = 0; i < (int)vnms.size(); ++i){
			normalize(vnms[i]);
		}
	}

	mats = m_mapMaterials;

	file.close();

	return true;
}


/*!
 * OBJt@C
 * @param[in] file_name t@C(tpX)
 * @param[in] vrts _W
 * @param[in] vnms _@
 * @param[in] plys |S
 * @param[in] mats ގ
 */
inline bool rxOBJ::Save(string file_name, 
						const vector<Vec3> &vrts, 
						const vector<Vec3> &vnms, 
						const vector<rxFace> &plys, 
						const rxMTL &mats)
{
	if((int)plys.size() == 0) return false;

	ofstream file;

	file.open(file_name.c_str());
	if(!file || !file.is_open() || file.bad() || file.fail()){
		cout << "rxOBJ::Save : Invalid file specified" << endl;
		return false;
	}

	//
	// ގ(*.mtl)
	//
	if(!mats.empty()){
		string mtl_file_name = ExtractPathWithoutExt(file_name)+".mtl";

		if(saveMTL(mtl_file_name, mats)){
			file << "mtllib " << mtl_file_name << endl;
		}
	}

	//
	// WIg(*.obj)
	//
	// _(v)
	int nv = (int)vrts.size();
	for(int i = 0; i < nv; ++i){
		file << "v " << Vec3ToString(Vec3(vrts[i])) << endl;
	}

	// @(vn)
	int nn = (int)vnms.size();
	for(int i = 0; i < nn; ++i){
		file << "vn " << Vec3ToString(Vec3(vnms[i])) << endl;
	}

	// eNX`W(vt)

	// (f)
	int np = (int)plys.size();
	string mat_name = "";
	for(int i = 0; i < np; ++i){
		if(plys[i].material_name != "" && plys[i].material_name != mat_name){
			mat_name = plys[i].material_name;
			file << "usemtl " << mat_name << endl;
		}

		file << "f ";
		int n = (int)plys[i].size();

		if(vnms.empty()){	// v
			for(int j = 0; j < n; ++j){
				int vidx = plys[i][j]+1;
				file << vidx;
				file << ((j == n-1) ? "" : " ");
			}
		}
		else{				// v//vn
			for(int j = 0; j < n; ++j){
				int vidx = plys[i][j]+1;
				file << vidx << "//" << vidx << " ";//((j == n-1) ? "" : " ");
			}
		}
		file << endl;
	}

	file.close();

	return true;
}


/*!
 * ʂ\("f"n܂)sʏ𒊏o
 * @param[in] buf s
 * @param[out] vidxs _CfbNX
 * @param[out] nidxs @CfbNX
 * @param[out] tidxs eNX`WCfbNX
 * @return _
 */
inline int rxOBJ::loadFace(string &buf, vector<int> &vidxs, vector<int> &nidxs, vector<int> &tidxs)
{
	int num_face = CountString(buf, 2, " ")+1;

	if(num_face >= 3){
		vidxs.resize(num_face);
		nidxs.resize(num_face);
		tidxs.resize(num_face);

		int vidx = 0;	// _CfbNX
		int tidx = 0;	// eNX`WCfbNX
		int nidx = 0;	// @CfbNX
		int offset = 2;
		for(int i = 0; i < num_face; ++i){
			if(sscanf(&buf[0]+offset, "%d/%d/%d", &vidx, &tidx, &nidx) != 3){
				if(sscanf(&buf[0]+offset, "%d//%d", &vidx, &nidx) != 2){
					if(sscanf(&buf[0]+offset, "%d/%d", &vidx, &tidx) != 2){
						sscanf(&buf[0]+offset, "%d", &vidx);	// _Ŵ
						tidx = 0;
						nidx = 0;
						offset += (int)log10((double)vidx)+1;
					}
					else{	// _WCeNX`W (*//*)
						nidx = 0;
						offset += (int)log10((double)vidx)+(int)log10((double)tidx)+3;
					}
				}
				else{	// _WC@ (*/*)
					tidx = 0;
					offset += (int)log10((double)vidx)+(int)log10((double)nidx)+4;
				}
			}
			else{ // _WCeNX`WC@ (*/*/*)
				offset += (int)log10((double)vidx)+(int)log10((double)tidx)+1+(int)log10((double)nidx)+4;
			}

			offset++;	// Xy[X

			if(vidx >= 1){
				vidxs[i]   = vidx-1;
				nidxs[i] = nidx-1;
				tidxs[i]  = tidx-1;
			}
			else{
				return 0;
			}
		}
	}

	return num_face;
}


/*!
 * ގt@C(*.mtl)̓ǂݍ
 * @param[in] mtl_fn t@C
 * @return ގ̐
 */
inline int rxOBJ::loadMTL(const string &mtl_fn)
{
	ifstream file;

	file.open(mtl_fn.c_str());
	if(!file || !file.is_open() || file.bad() || file.fail()){
		cout << "rxOBJ::loadMTL : Invalid file specified" << endl;
		return false;
	}

	int num_mat = 0;
	rxMaterialOBJ *cur_mat;

	string buf;
	string::size_type comment_start = 0;
	while(!file.eof()){
		getline(file, buf);

		// '#'ȍ~̓RgƂĖ
		if( (comment_start = buf.find('#')) != string::size_type(-1) )
			buf = buf.substr(0, comment_start);

		// s̃Xy[XC^u폜
		DeleteHeadSpace(buf);

		// s͖
		if(buf.empty())
			continue;

		if(buf[0] == 'n'){
			string mat_name;
			if(!StringToString(buf, "newmtl", mat_name)) continue;

			rxMaterialOBJ mat;
			mat.name      = mat_name;
			mat.diffuse   = Vec4(0.0);
			mat.specular  = Vec4(0.0);
			mat.ambient   = Vec4(0.0);
			mat.color     = Vec4(0.0);
			mat.emission  = Vec4(0.0);
			mat.shininess = 0.0;
			mat.illum = 2;
			mat.tex_file = "";
			mat.tex_name = 0;

			m_mapMaterials[mat_name] = mat;
			cur_mat = &m_mapMaterials[mat_name];

			num_mat++;
		}
		else if(buf[0] == 'K'){
			Vec3 col;
			if(buf[1] == 'd'){		// gUː
				if(StringToVec3s(buf, "Kd", col)){
					cur_mat->diffuse = Vec4(col, cur_mat->diffuse[3]);
				}
			}
			else if(buf[1] == 's'){	// ʔː
				if(StringToVec3s(buf, "Ks", col)){
					cur_mat->specular = Vec4(col, cur_mat->specular[3]);
				}
			}
			else if(buf[1] == 'a'){	// ː
				if(StringToVec3s(buf, "Ka", col)){
					cur_mat->ambient = Vec4(col, cur_mat->ambient[3]);
				}
			}
		}
		else if(buf[0] == 'd'){	// At@l(ߗ)
			double alpha;
			if(StringToDouble(buf, "d", alpha)){
				cur_mat->diffuse[3]  = alpha;
				cur_mat->specular[3] = alpha;
				cur_mat->ambient[3]  = alpha;
				cur_mat->color[3]    = alpha;
				cur_mat->emission[3] = alpha;
			}
		}
		else if(buf[0] == 'T' && buf[1] == 'r'){	// F
			Vec3 emit;
			if(StringToVec3s(buf, "Tr", emit)){
				cur_mat->emission = Vec4(emit, cur_mat->emission[3]);
			}
		}
		else if(buf[0] == 'N' && buf[1] == 's'){	// Shininess
			double shine = 0.0;
			if(StringToDouble(buf, "Ns", shine)){
				cur_mat->shininess = shine;
			}
		}
		else if(buf[0] == 'i' && buf[1] == 'l' && buf[2] == 'l'){	// ؖf(1ŋʔ˖, 2ŗL)
			string illum;
			if(StringToString(buf, "illum", illum)){
				cur_mat->illum = atoi(illum.c_str());
			}
		}
		else if(buf[0] == 'm' && buf[1] == 'a' && buf[2] == 'p'){	// eNX`
			string texfn;
			if(StringToString(buf, "map_Kd", texfn)){
				string dir = ExtractDirPath(mtl_fn);
				if(!dir.empty()){
					texfn = dir+"\\"+texfn;
				}

				cur_mat->tex_file = texfn;
			}
		}
	}

	if(!m_mapMaterials.empty()){
		map<string, rxMaterialOBJ>::iterator iter = m_mapMaterials.begin();
		for(; iter != m_mapMaterials.end(); ++iter){
			cout << "material : " << iter->first << endl;
			cout << "  diffuse   = " << iter->second.diffuse << endl;
			cout << "  specular  = " << iter->second.specular << endl;
			cout << "  ambient   = " << iter->second.ambient << endl;
			cout << "  color     = " << iter->second.color << endl;
			cout << "  emission  = " << iter->second.emission << endl;
			cout << "  shininess = " << iter->second.shininess << endl;
			cout << "  illum     = " << iter->second.illum << endl;
			cout << "  tex_file  = " << iter->second.tex_file << endl;
		}

		m_strCurrentMat = m_mapMaterials.begin()->first;
	}

	return num_mat;
}


/*!
 * ގt@C(*.mtl)̏
 * @param[in] mtl_fn t@C
 * @param[in] mats ގ
 */
inline int rxOBJ::saveMTL(const string &mtl_fn, const rxMTL &mats)
{
	ofstream file;

	file.open(mtl_fn.c_str(), ios::out);
	if(!file || !file.is_open() || file.bad() || file.fail()){
		cout << "rxOBJ::saveMTL : Invalid file specified" << endl;
		return 0;
	}

	if(mats.empty()) return 0;

	map<string, rxMaterialOBJ>::iterator iter = m_mapMaterials.begin();
	for(; iter != m_mapMaterials.end(); ++iter){
		file << "newmtl " << iter->first << endl;
		file << "Kd " << Vec3ToString(Vec3(iter->second.diffuse)) << endl;
		file << "Ks " << Vec3ToString(Vec3(iter->second.specular)) << endl;
		file << "Ka " << Vec3ToString(Vec3(iter->second.ambient)) << endl;
		file << "Tr " << Vec3ToString(Vec3(iter->second.emission)) << endl;
		file << "d " << iter->second.diffuse[3] << endl;
		file << "Ns " << iter->second.shininess << endl;
		file << "illum " << iter->second.illum << endl;
		file << "map_Kd " << iter->second.tex_file << endl;
		file << endl;
	}

	file.close();

	return 1;
}

#endif // _RX_OBJ_H_
