/*! 
 @file rx_shader.h

 @brief GLSLVF[_[
 
 @author Makoto Fujisawa
 @date 2009-11
*/
// FILE --rx_shader.h--


#ifndef _RX_SHADERS_H_
#define _RX_SHADERS_H_


//-----------------------------------------------------------------------------
// CN[ht@C
//-----------------------------------------------------------------------------
#include <cstdio>
#include <iostream>
#include <vector>
#include <string>

#include <GL/glew.h>
#include <GL/glut.h>


#define RXSTR(A) #A

using namespace std;


//-----------------------------------------------------------------------------
// GLSLVF[_
//-----------------------------------------------------------------------------
struct rxGLSL
{
	string VertProg;
	string FragProg;
	string Name;
	GLuint Prog;
};


//-----------------------------------------------------------------------------
// Volume Rendering VF[_
//-----------------------------------------------------------------------------
const char volume_vs[] = RXSTR(
varying vec3 texcoord;	// {[eNX`p̃eNX`W
void main(void)
{
	gl_Position = gl_ProjectionMatrix*gl_Vertex;
	texcoord = (gl_ModelViewMatrixInverse*gl_Vertex).xyz;
	gl_ClipVertex = gl_Vertex;
}
);

const char volume_fs[] = RXSTR(
uniform sampler3D volume_tex;
uniform float thickness;	// XCXԋ(=XCX)
uniform float opacity;		// x
uniform vec3 dens_col;		// {[̐F
varying vec3 texcoord;
void main(void)
{
	float d = texture3D(volume_tex, texcoord).x;
	gl_FragColor = vec4(dens_col.r, dens_col.g, dens_col.b, d*opacity*thickness);
	//gl_FragColor = vec4(d, 1.0, 1.0, 0.5);
}
);

//-----------------------------------------------------------------------------
// Volume Rendering VF[_ (FeNX`gp)
//-----------------------------------------------------------------------------
const char volume_image_vs[] = RXSTR(
varying vec3 texcoord;	// {[eNX`p̃eNX`W
void main(void)
{
	gl_Position = gl_ProjectionMatrix*gl_Vertex;
	texcoord = (gl_ModelViewMatrixInverse*gl_Vertex).xyz;
	gl_ClipVertex = gl_Vertex;
}
);

const char volume_image_fs[] = RXSTR(
uniform sampler3D volume_tex;
uniform sampler2D color_tex;	//!< FpeNX`
uniform float thickness;		//!< XCXԋ(=XCX)
uniform float opacity;			//!< x
uniform vec3 dens_col;			//!< {[̐F
varying vec3 texcoord;
void main(void)
{
	float d = texture3D(volume_tex, texcoord).x;
	vec2 color_texcoord;
	color_texcoord.x = 0.5;
	color_texcoord.y = 1.0-clamp((d-0.1)/0.9, 0.0, 1.0);
	vec4 col = texture2D(color_tex, color_texcoord);

	float a = (d > 0.2) ? d : 0.0;

	gl_FragColor = vec4(col.r, col.g, col.b, a*40.0*thickness);
}
);

//-----------------------------------------------------------------------------
// Shadow Map VF[_
//-----------------------------------------------------------------------------
//Shadow Map vs
const char shadow_vs[] = RXSTR(
varying vec4 vPos;
varying vec3 vNrm;
void main(void)
{
	vPos = gl_ModelViewMatrix*gl_Vertex;
	vNrm = normalize(gl_NormalMatrix*gl_Normal);
	//vNrm = gl_NormalMatrix*gl_Normal;

	gl_Position = gl_ProjectionMatrix*vPos;
	//gl_FrontColor = gl_Color*gl_LightSource[0].diffuse*vec4(max(dot(vNrm, gl_LightSource[0].position.xyz), 0.0));
	gl_FrontColor = gl_Color;
	gl_TexCoord[0] = gl_MultiTexCoord0;
	//gl_TexCoord[0] = gl_TextureMatrix[0] * gl_MultiTexCoord0;

}
);

// Normal Mode fs
const char shadow_single_fs[] = RXSTR( 
#version 110 @
#extension GL_EXT_texture_array : enable @

uniform sampler2D tex;			//!< ͗l
uniform sampler2DArray stex;	//!< fvXleNX`(~䕪)
uniform vec4 far_d;				//!< 䉓
varying vec4 vPos;
varying vec3 vNrm;

float ShadowCoef(void)
{
	// fʒugl_FragCoord.xy ɂfvXl gl_FragCoord.zK؂ȃfvX}bv
	// ȏ̕ɂ0Ă(ႦΕ2Ȃ,far_d.z,far_d.w0)
	int index = 3;
	if(gl_FragCoord.z < far_d.x){
		index = 0;
	}
	else if(gl_FragCoord.z < far_d.y){
		index = 1;
	}
	else if(gl_FragCoord.z < far_d.z){
		index = 2;
	}
	
	// _Wn̈ʒu_ƂWn̂̂ɕϊ
	vec4 shadow_coord = gl_TextureMatrix[index]*vPos;

	// ̃fvXl(Օ)Ҕ
	float light_d = shadow_coord.z;
	
	// ǂ̕p邩
	shadow_coord.z = float(index);
	
	// i[ꂽ̍ŏfvXl擾
	float shadow_d = texture2DArray(stex, shadow_coord.xyz).x;
	
	// ̃fvXlƎՕlfvXl̍߂
	float diff = shadow_d-light_d;

	// e0,1Ԃ
	return clamp(diff*250.0+1.0, 0.0, 1.0);
}

void main(void)
{
	vec4 light_col;
	vec3 N = normalize(vNrm);
	//vec3 L = gl_LightSource[0].position.xyz;
	vec3 L = normalize(gl_LightSource[0].position.xyz-vPos.xyz);	// xNg

	// ̌vZ
	//  - OpenGLvZxƔˌW̐(gl_FrontLightProduct)pD
	light_col = gl_FrontLightProduct[0].ambient;

	// gU˂̌vZ
	//float diff = max(dot(L, N), 0.0);
	float diff = max(dot(vNrm, gl_LightSource[0].position.xyz), 0.0);

	// ʔ˂̌vZ
	if(diff > 0.0){
		// ˃xNǧvZ
		vec3 V = normalize(-vPos.xyz);
		//vec3 R = 2.0*dot(N, L)*N-L;
		vec3 R = reflect(-L, N);
		float spec = pow(abs(dot(R, V)), gl_FrontMaterial.shininess);

		light_col += gl_FrontLightProduct[0].diffuse*diff+
					 gl_FrontLightProduct[0].specular*spec;
	}

	const float shadow_ambient = 0.8;
	vec4 color_tex = texture2D(tex, gl_TexCoord[0].st);
	float shadow_coef = ShadowCoef();

	light_col += color_tex;

	gl_FragColor = shadow_ambient*shadow_coef*light_col+(1.0-shadow_ambient)*color_tex;
	//gl_FragColor = shadow_ambient*shadow_coef*gl_Color+(1.0-shadow_ambient)*gl_Color;
	//gl_FragColor = shadow_ambient*shadow_coef*light_col*color_tex+(1.0-shadow_ambient)*color_tex;

}
);

//! ̃VhETvgp
// Smooth shadows fs
const char shadow_multi_fs[] = RXSTR(
#version 120 @
#extension GL_EXT_texture_array : enable @

uniform sampler2D tex;			//!< ͗l
uniform sampler2DArray stex;	//!< fvXleNX`(~䕪)
uniform vec4 far_d;				//!< 䉓

// o[ebNXVF[_󂯎ϐ
varying vec4 vPos;
varying vec3 vNrm;

// ͂̉eWQƗp
const int nsamples = 8;
uniform vec4 offset[nsamples] = vec4[nsamples](vec4(0.000000, 0.000000, 0.0, 0.0),
											   vec4(0.079821, 0.165750, 0.0, 0.0),
											   vec4(-0.331500, 0.159642, 0.0, 0.0),
											   vec4(-0.239463, -0.497250, 0.0, 0.0),
											   vec4(0.662999, -0.319284, 0.0, 0.0),
											   vec4(0.399104, 0.828749, 0.0, 0.0),
											   vec4(-0.994499, 0.478925, 0.0, 0.0),
											   vec4(-0.558746, -1.160249, 0.0, 0.0));

/*!
 * eW̌vZ
 * @param[in] shadow_coord VhE}bvQƗpeNX`W
 * @param[in] light_d Qƈʒuł̌̃fvXl
 * @return eW(êƂ1, ȊO0)
 */
float GetOccCoef(vec4 shadow_coord, float light_d)
{
	// i[ꂽ̍ŏfvXl擾
	float shadow_d = texture2DArray(stex, shadow_coord.xyz).x;
	
	// ̃fvXlƎՕlfvXl̍߂
	float diff = shadow_d-light_d;
	
	// e0,1Ԃ
	return clamp(diff*250.0+1.0, 0.0, 1.0);
}

/*!
 * ê߂̌W(êƂ1, ȊO0)
 * @return eW(êƂ1, ȊO0)
 */
float ShadowCoef(void)
{
	// fʒugl_FragCoord.xy ɂfvXl gl_FragCoord.zK؂ȃfvX}bv
	// ȏ̕ɂ0Ă(ႦΕ2Ȃ,far_d.z,far_d.w0)
	int index = 3;
	if(gl_FragCoord.z < far_d.x){
		index = 0;
	}
	else if(gl_FragCoord.z < far_d.y){
		index = 1;
	}
	else if(gl_FragCoord.z < far_d.z){
		index = 2;
	}
	
	const float scale = 2.0/4096.0;

	// _Wn̈ʒu_ƂWn̂̂ɕϊ
	vec4 shadow_coord = gl_TextureMatrix[index]*vPos;
	
	// ̃fvXl(Օ)Ҕ
	float light_d = shadow_coord.z;
	
	// ǂ̕p邩
	shadow_coord.z = float(index);
	
	// ͂̉eW܂߂Ď擾
	float shadow_coef = 0.0;
	for(int i = 0; i < nsamples; ++i){
		shadow_coef += GetOccCoef(shadow_coord+scale*offset[i], light_d);
	}
	shadow_coef /= nsamples;
	
	return shadow_coef;
}

void main(void)
{
	vec4 light_col;
	vec3 N = normalize(vNrm);
	//vec3 L = gl_LightSource[0].position.xyz;
	vec3 L = normalize(gl_LightSource[0].position.xyz-vPos.xyz);	// xNg

	// ̌vZ
	//  - OpenGLvZxƔˌW̐(gl_FrontLightProduct)pD
	light_col = gl_FrontLightProduct[0].ambient;

	// gU˂̌vZ
	//float diff = max(dot(L, N), 0.0);
	float diff = max(dot(vNrm, gl_LightSource[0].position.xyz), 0.0);

	// ʔ˂̌vZ
	if(diff > 0.0){
		// ˃xNǧvZ
		vec3 V = normalize(-vPos.xyz);
		//vec3 R = 2.0*dot(N, L)*N-L;
		vec3 R = reflect(-L, N);
		float spec = pow(abs(dot(R, V)), gl_FrontMaterial.shininess);

		light_col += gl_FrontLightProduct[0].diffuse*diff+
					 gl_FrontLightProduct[0].specular*spec;
	}

	const float shadow_ambient = 0.8;
	vec4 color_tex = texture2D(tex, gl_TexCoord[0].st);
	float shadow_coef = ShadowCoef();

	gl_FragColor = shadow_ambient*shadow_coef*light_col+(1.0-shadow_ambient)*light_col;
	//gl_FragColor = shadow_ambient*shadow_coef*gl_Color+(1.0-shadow_ambient)*gl_Color;
}
);


//-----------------------------------------------------------------------------
// Shadow View VF[_
//-----------------------------------------------------------------------------
const char shadow_view_vs[] = RXSTR(
void main(void)
{
	gl_TexCoord[0] = vec4(0.5)*gl_Vertex + vec4(0.5);
	gl_Position = gl_Vertex;
}
);

const char shadow_view_fs[] = RXSTR(
#version 110 @
#extension GL_EXT_texture_array : enable @

uniform sampler2DArray tex;
uniform float layer;

void main(void)
{
	vec4 tex_coord = vec4(gl_TexCoord[0].x, gl_TexCoord[0].y, layer, 1.0);
	gl_FragColor = texture2DArray(tex, tex_coord.xyz);
//	gl_FragColor = vec4(1.0, 0.0, 0.0, 1.0);
}
);




//-----------------------------------------------------------------------------
// pointspriteVF[_
//-----------------------------------------------------------------------------
const char ps_vs[] = RXSTR(
uniform float prad;  // |CgTCY
uniform float pscale;   // |CgTCYsNZŕ\Ƃ̃XP[W
void main(void)
{
	// EBhEԂł̃|CgTCYvZ
	vec3 pos_in_camera = vec3(gl_ModelViewMatrix*vec4(gl_Vertex.xyz, 1.0));
	float dist = length(pos_in_camera);
	gl_PointSize = prad*(pscale/dist);

	gl_TexCoord[0] = gl_MultiTexCoord0;
	gl_Position = gl_ModelViewProjectionMatrix*vec4(gl_Vertex.xyz, 1.0);

	gl_FrontColor = gl_Color;
}
);

const char ps_fs[] = RXSTR(
void main(void)
{
	vec3 light_dir = vec3(0.577, 0.577, 0.577);

	// eNX`WvZ
	vec3 nrm;
	nrm.xy = gl_TexCoord[0].xy*vec2(2.0, -2.0)+vec2(-1.0, 1.0);
	float mag = dot(nrm.xy, nrm.xy);
	if (mag > 1.0) discard;   // kill pixels outside circle
	nrm.z = sqrt(1.0-mag);

	// FvZ
	float diffuse = max(0.0, dot(light_dir, nrm));

	gl_FragColor = gl_Color*diffuse;
}
);



//-----------------------------------------------------------------------------
// PhongVF[_
//-----------------------------------------------------------------------------

//! Phong _VF[_
const char phong_vs[] = RXSTR(
#version 110 @
#extension GL_EXT_texture_array : enable @

// tOgVF[_ɒln߂̕ϐ
varying vec4 vPos;
varying vec3 vNrm;
varying vec3 vObjPos;

void main(void)
{
	// _ʒuƖ@
	vPos = gl_ModelViewMatrix*gl_Vertex;
	vNrm = normalize(gl_NormalMatrix*gl_Normal);

	gl_Position = ftransform();
	gl_TexCoord[0] = gl_TextureMatrix[0] * gl_ModelViewMatrix * gl_Vertex;
	
	vObjPos = gl_Vertex.xyz;
}
);

//! Phong sNZVF[_
const char phong_fs[] = RXSTR(
#version 110 @
#extension GL_EXT_texture_array : enable @

uniform vec3 Ke; // ːF
uniform vec3 Ka; // 
uniform vec3 Kd; // gU
uniform vec3 Ks; // ʔ
uniform float shine;
uniform vec3 La;	// Cg
uniform vec3 Ld;	// CggUˌ
uniform vec3 Ls;	// Cgʔˌ
uniform vec3 Lpos;	// Cgʒu
uniform vec3 eyePosition;

// o[ebNXVF[_󂯎ϐ
varying vec4 vPos;
varying vec3 vNrm;
varying vec3 vObjPos;

void main(void)
{
	vec3 P = vObjPos.xyz;
	vec3 N = normalize(vNrm);
	
	// ːF̌vZ
	vec3 emissive = Ke;

	// ̌vZ
	vec3 ambient = Ka*La;

	// gU˂̌vZ
	vec3 L = normalize(Lpos-P);
	float diffuseLight = max(dot(L, N), 0.0);
	vec3 diffuse = Kd*Ld*diffuseLight;

	// ʔ˂̌vZ
	vec3 V = normalize(eyePosition-P);
	vec3 H = normalize(L+V);
	float specularLight = pow(max(dot(H, N), 0.0), shine);
	if(diffuseLight <= 0.0) specularLight = 0.0;
	vec3 specular = Ks*Ls*specularLight;

	gl_FragColor.xyz = emissive+ambient+diffuse+specular;
	gl_FragColor.w = 1.0;
}
);

//-----------------------------------------------------------------------------
// FresnelVF[_
//-----------------------------------------------------------------------------

//! Fresnel _VF[_
const char fresnel_vs[] = RXSTR(
#version 120 @
#extension GL_EXT_texture_array : enable @

uniform float fresnelBias;
uniform float fresnelScale; 
uniform float fresnelPower; 
uniform float etaRatio;
uniform vec3 eyePosition;

// tOgVF[_ɒln߂̕ϐ
varying vec4 vPos;
varying vec3 vNrm;
varying float fNoise;
varying vec3 oR;
varying vec3 oT;
varying float oFresnel;

void main(void)
{
	// _ʒuƖ@
	fNoise = gl_Vertex.w;
	vec4 vert = gl_Vertex;
	vert.w = 1.0f;

	vPos = gl_ModelViewMatrix*vert;
	vNrm = normalize(gl_NormalMatrix*gl_Normal);

	gl_Position = gl_ModelViewProjectionMatrix*vert;//ftransform();
	gl_TexCoord[0] = gl_TextureMatrix[0]*gl_ModelViewMatrix*vert;

	// ˁCˁC܃xNǧvZ
	vec3 I = vPos.xyz-eyePosition;
	I = normalize(I);
	oR = reflect(I, vNrm);
	oT = refract(I, vNrm, etaRatio);

	// ˈ̌vZ
	oFresnel = fresnelBias+fresnelScale*pow(min(0.0, 1.0-dot(I, vNrm)), fresnelPower);
}
);

//! Fresnel sNZVF[_
const char fresnel_fs[] = RXSTR(
#version 120 @
#extension GL_EXT_texture_array : enable @

uniform samplerCube envmap;

uniform float maxNoise;
uniform float minNoise;
uniform vec3 FoamColor;

// o[ebNXVF[_󂯎ϐ
varying vec4 vPos;
varying vec3 vNrm;
varying float fNoise;
varying vec3 oR;
varying vec3 oT;
varying float oFresnel;

vec4 lerp(vec4 a, vec4 b, float s)
{
	return vec4(a + (b - a) * s);       
}

vec3 lerp(vec3 a, vec3 b, float s)
{
	return vec3(a + (b - a) * s);       
}

void main(void)
{
	// ˊF̎擾
	vec4 reflecColor = textureCube(envmap, oR);
	reflecColor.a = 1.0;

	// ܊F̌vZ
	vec4 refracColor;
	refracColor.rgb = textureCube(envmap, oT).rgb;
	refracColor.a = 1.0;

	vec4 cout = lerp(refracColor, reflecColor, oFresnel);

	if(fNoise > minNoise){
		float foam_rate = (fNoise-minNoise)/(maxNoise-minNoise);
		gl_FragColor.rgb = lerp(cout.rgb, FoamColor, foam_rate);
	}
	else{
		gl_FragColor.rgb = cout.rgb;
	}	

	gl_FragColor.a = oFresnel*0.5+0.5;
}
);


//-----------------------------------------------------------------------------
// GLSLRpC
//-----------------------------------------------------------------------------
/*!
 * GLSLvÕRpC
 * @param[in] vsource vertexVF[_vOe
 * @param[in] fsource pixel(fragment)VF[_vOe
 * @return GLSLvOԍ
 */
static GLuint CompileProgram(const char *vsource, const char *fsource)
{
	GLuint vertexShader = glCreateShader(GL_VERTEX_SHADER);
	GLuint fragmentShader = glCreateShader(GL_FRAGMENT_SHADER);

	glShaderSource(vertexShader, 1, &vsource, 0);
	glShaderSource(fragmentShader, 1, &fsource, 0);
	
	glCompileShader(vertexShader);
	glCompileShader(fragmentShader);

	GLuint program = glCreateProgram();

	glAttachShader(program, vertexShader);
	glAttachShader(program, fragmentShader);

	glLinkProgram(program);

	// check if program linked
	GLint success = 0;
	glGetProgramiv(program, GL_LINK_STATUS, &success);

	if (!success) {
		char temp[256];
		glGetProgramInfoLog(program, 256, 0, temp);
		printf("Failed to link program:\n%s\n", temp);
		glDeleteProgram(program);
		program = 0;
	}

	return program;
}

/*!
 * GLSLVF[_RpC
 * @param[in] target ^[Qbg(GL_VERTEX_SHADER,GL_FRAGMENT_SHADER)
 * @param[in] shader VF[_R[h
 * @return GLSLIuWFNg
 */
inline GLuint CompileGLSLShader(GLenum target, const char* shader)
{
	// GLSLIuWFNg쐬
	GLuint object = glCreateShader(target);

	if(!object) return 0;

	glShaderSource(object, 1, &shader, NULL);
	glCompileShader(object);

	// RpCԂ̊mF
	GLint compiled = 0;
	glGetShaderiv(object, GL_COMPILE_STATUS, &compiled);

	if(!compiled){
		char temp[256] = "";
		glGetShaderInfoLog( object, 256, NULL, temp);
		fprintf(stderr, " Compile failed:\n%s\n", temp);

		glDeleteShader(object);
		return 0;
	}

	return object;
}

/*!
 * GLSLVF[_RpC
 * @param[in] target ^[Qbg(GL_VERTEX_SHADER,GL_FRAGMENT_SHADER)
 * @param[in] fn VF[_t@CpX
 * @return GLSLIuWFNg
 */
inline GLuint CompileGLSLShaderFromFile(GLenum target, const char* fn)
{
	FILE *fp;

	// oCiƂăt@Cǂݍ
	fopen_s(&fp, fn, "rb");
	if(fp == NULL) return 0;

	// t@C̖Ɉړ݈ʒu(t@CTCY)擾
	fseek(fp, 0, SEEK_END);
	long size = ftell(fp);

	fseek(fp, 0, SEEK_SET);

	// VF[_̓ei[
	char *text = new char[size+1];
	fread(text, size, 1, fp);
	text[size] = '\0';

	//printf("%s\n", text);


	fclose(fp);

	// VF[_RpC
	printf("Compile %s\n", fn);
	GLuint object = CompileGLSLShader(target, text);

	delete [] text;

	return object;
}

/*!
 * o[ebNXƃtOgVF[_ō\GLSLvO쐬
 * @param[in] vs o[ebNXVF[_IuWFNg
 * @param[in] fs tOgVF[_IuWFNg
 * @return GLSLvOIuWFNg
 */
inline GLuint LinkGLSLProgram(GLuint vs, GLuint fs)
{
	// vOIuWFNg쐬
	GLuint program = glCreateProgram();

	// VF[_IuWFNgo^
	glAttachShader(program, vs);
	glAttachShader(program, fs);

	// vÕN
	glLinkProgram(program);

	// G[o
	GLint charsWritten, infoLogLength;
	glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLogLength);

	char * infoLog = new char[infoLogLength];
	glGetProgramInfoLog(program, infoLogLength, &charsWritten, infoLog);
	printf(infoLog);
	delete [] infoLog;

	// JeXg
	GLint linkSucceed = GL_FALSE;
	glGetProgramiv(program, GL_LINK_STATUS, &linkSucceed);
	if(linkSucceed == GL_FALSE){
		glDeleteProgram(program);
		return 0;
	}

	return program;
}


/*!
 * o[ebNX/WIg/tOgVF[_ō\GLSLvO쐬
 * @param[in] vs o[ebNXVF[_IuWFNg
 * @param[in] gs WIgVF[_IuWFNg
 * @param[in] inputType WIgVF[_ւ̓̓^Cv
 * @param[in] vertexOut o[ebNX̏o
 * @param[in] outputType WIgVF[_̏o̓^Cv
 * @param[in] fs tOgVF[_IuWFNg
 * @return GLSLvOIuWFNg
 */
inline GLuint LinkGLSLProgram(GLuint vs, GLuint gs, GLint inputType, GLint vertexOut, GLint outputType, GLuint fs)
{
	// vOIuWFNg쐬
	GLuint program = glCreateProgram();

	// VF[_IuWFNgo^
	glAttachShader(program, vs);
	glAttachShader(program, gs);

	glProgramParameteriEXT(program, GL_GEOMETRY_INPUT_TYPE_EXT, inputType);
	glProgramParameteriEXT(program, GL_GEOMETRY_VERTICES_OUT_EXT, vertexOut);
	glProgramParameteriEXT(program, GL_GEOMETRY_OUTPUT_TYPE_EXT, outputType);
	glAttachShader(program, fs);

	// vÕN
	glLinkProgram(program);

	// G[o
	GLint charsWritten, infoLogLength;
	glGetProgramiv(program, GL_INFO_LOG_LENGTH, &infoLogLength);

	char * infoLog = new char[infoLogLength];
	glGetProgramInfoLog(program, infoLogLength, &charsWritten, infoLog);
	printf(infoLog);
	delete [] infoLog;

	// JeXg
	GLint linkSucceed = GL_FALSE;
	glGetProgramiv(program, GL_LINK_STATUS, &linkSucceed);
	if(linkSucceed == GL_FALSE){
		glDeleteProgram(program);
		return 0;
	}

	return program;
}


/*!
 * GLSL̃RpCEN(t@C)
 * @param[in] vs _VF[_t@CpX
 * @param[in] fs tOgVF[_t@CpX
 * @param[in] name vO
 * @return GLSLIuWFNg
 */
inline rxGLSL CreateGLSLFromFile(const string &vs, const string &fs, const string &name)
{
	rxGLSL gs;
	gs.VertProg = vs;
	gs.FragProg = fs;
	gs.Name = name;

	GLuint v, f;
	printf("compile the vertex shader : %s\n", name.c_str());
	if(!(v = CompileGLSLShaderFromFile(GL_VERTEX_SHADER, vs.c_str()))){
		// skip the first three chars to deal with path differences
		v = CompileGLSLShaderFromFile(GL_VERTEX_SHADER, &vs.c_str()[3]);
	}

	printf("compile the fragment shader : %s\n", name.c_str());
	if(!(f = CompileGLSLShaderFromFile(GL_FRAGMENT_SHADER, fs.c_str()))){
		f = CompileGLSLShaderFromFile(GL_FRAGMENT_SHADER, &fs.c_str()[3]);
	}

	gs.Prog = LinkGLSLProgram(v, f);
	//gs.Prog = GLSL_CreateShaders(gs.VertProg.c_str(), gs.FragProg.c_str());

	return gs;
}

/*!
 * #versionȂǂ̃vvZbT𕶎ƂďꂽVF[_Ɋ܂ޏꍇCs܂Ȃ̂ŁC
 *  #version 110 @ ̂悤ɍŌ@tCsɕϊ
 * @param[in] s  VF[_
 * @param[in] vs ϊ̃VF[_
 */
inline void CreateGLSLShaderString(const char* s, vector<char> &vs)
{
	int idx = 0;
	char c = s[0];
	while(c != '\0'){
		if(c == '@') c = '\n'; // #versionȂǂ\ɂ邽߂@sɕϊ

		vs.push_back(c);
		idx++;
		c = s[idx];
	}
	vs.push_back('\0');
}

/*!
 * GLSL̃RpCEN()
 * @param[in] vs _VF[_e
 * @param[in] fs tOgVF[_e
 * @param[in] name vO
 * @return GLSLIuWFNg
 */
inline rxGLSL CreateGLSL(const char* vs, const char* fs, const string &name)
{
	rxGLSL gs;
	gs.VertProg = "from char";
	gs.FragProg = "from char";
	gs.Name = name;

	vector<char> vs1, fs1;
	CreateGLSLShaderString(vs, vs1);
	CreateGLSLShaderString(fs, fs1);
	
	//printf("vertex shader : %d\n%s\n", vs1.size(), &vs1[0]);
	//printf("pixel shader  : %d\n%s\n", fs1.size(), &fs1[0]);

	GLuint v, f;
	printf("compile the vertex shader : %s\n", name.c_str());
	v = CompileGLSLShader(GL_VERTEX_SHADER, &vs1[0]);
	printf("compile the fragment shader : %s\n", name.c_str());
	f = CompileGLSLShader(GL_FRAGMENT_SHADER, &fs1[0]);
	gs.Prog = LinkGLSLProgram(v, f);

	return gs;
}



#endif // #ifndef _RX_SHADERS_H_