/*!
  @file rx_vrml.h

  @brief VRML Input/Output
	- ShapeIndexedFaceSet̂ݑΉ

  @author Makoto Fujisawa
  @date   2011
*/

#ifndef _RX_VRML_H_
#define _RX_VRML_H_


//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------
#include "rx_mesh.h"
#include <stack>

//-----------------------------------------------------------------------------
// Name Space
//-----------------------------------------------------------------------------
using namespace std;


//-----------------------------------------------------------------------------
// VRMLNX̐錾 - VRML`̓ǂݍ
//-----------------------------------------------------------------------------
class rxVRML
{
public:
	//! RXgN^
	rxVRML()
	{
	}

	//! fXgN^
	~rxVRML()
	{
	}

	/*!
	 * VRMLt@Cǂݍ
	 * @param[in] file_name t@C(tpX)
	 * @param[out] vrts _W
	 * @param[out] vnms _@
	 * @param[out] poly |S
	 * @param[out] mats ގ
	 * @param[in] triangle |S̎Op`tO
	 */
	bool Read(string file_name, vector<Vec3> &vrts, vector<Vec3> &vnms, vector<rxFace> &plys, rxMTL &mats, bool tri = true);

	/*!
	 * VRMLt@C
	 * @param[in] file_name t@C(tpX)
	 * @param[in] vrts _W
	 * @param[in] vnms _@
	 * @param[in] plys |S
	 * @param[in] mats ގ
	 */
	bool Save(string file_name, const vector<Vec3> &vrts, const vector<Vec3> &vnms, const vector<rxFace> &plys, const rxMTL &mats);

};

/*!
 * VRMLt@Cǂݍ
 * @param[in] filename wrlt@C̃pX
 * @param[inout] poly 3DIuWFNg
 * @param[in] cen,ext,ang ͂3D`̐Kp[^
 */
static void ReadVRML(const string filename, rxPolygons &polys)
{
	if(!polys.vertices.empty()){
		polys.vertices.clear();
		polys.normals.clear();
		polys.faces.clear();
		polys.materials.clear();
		polys.open = 0;
	}
	rxVRML vrml;
	if(vrml.Read(filename, polys.vertices, polys.normals, polys.faces, polys.materials, true)){
		if(polys.normals.empty()){
			CalVertexNormals(polys);
		}
		polys.open = 1;
	}
}

/*!
 * VRMLt@C
 * @param[in] filename wrlt@C̃pX
 */
static void SaveVRML(const string filename, rxPolygons &polys)
{
	rxVRML vrml;
	if(vrml.Save(filename, polys.vertices, polys.normals, polys.faces, polys.materials)){
		std::cout << filename << " have been saved (VRML)." << endl;
	}
}


/*!
 * OBJt@Cǂݍ
 * @param[in] file_name t@C(tpX)
 * @param[out] vrts _W
 * @param[out] vnms _@
 * @param[out] poly |S
 * @param[out] mats ގ
 * @param[in] triangle |S̎Op`tO
 */
inline bool rxVRML::Read(string file_name, 
						 vector<Vec3> &vrts, 
						 vector<Vec3> &vnms, 
						 vector<rxFace> &plys, 
						 rxMTL &mats, 
						 bool triangle)
{
	ifstream file;

	file.open(file_name.c_str());
	if(!file || !file.is_open() || file.bad() || file.fail()){
		cout << "rxVRML::Read : Invalid file specified" << endl;
		return false;
	}

	string buf;

	// ŏ̈sǂݍVRML̃o[W`FbN
	getline(file, buf);
	if(buf.find("#VRML") == string::npos){
		cout << "rxVRML::Read : The file is not VRML file" << endl;
		return false;
	}
	if(buf.find("V2.0") == string::npos){
		cout << "rxVRML::Read : Only V2.0 is supported" << endl;
		return false;
	}

	// f[^ǂݍ
	vector< vector<int> > idxes, color_idxes;
	stack<string> node;

	int shape_verts = 0;
	bool shape = false;
	bool indexed_face_set = false;
	string data;
	string prev_buf;
	string::size_type comment_start = 0;
	while(!file.eof()){
		getline(file, buf);

		// '#'ȍ~̓RgƂĖ
		if( (comment_start = buf.find('#')) != string::size_type(-1) )
			buf = buf.substr(0, comment_start);

		// s͖
		if(buf.empty())	continue;

		// s̃Xy[XC^u폜
		DeleteHeadSpace(buf);

		data = buf;

		// "{""["m[hX^bNɊi[
		size_t pos;
		if((pos = buf.find_first_of("[{")) != string::npos){
			if(pos == 0){
				if(!prev_buf.empty()) node.push(prev_buf);
			}
			else{
				string tag_name = buf.substr(0, pos);
				DeleteSpace(tag_name);
				if(!tag_name.empty()) node.push(tag_name);
			}
			data = buf.substr(pos+1, string::npos);

			if(node.top() == "Shape"){
				shape = true;
			}
			if(shape && node.top().find("IndexedFaceSet") != string::npos){
				indexed_face_set = true;
			}
		}
		else{
			prev_buf = buf;
		}

		if(shape){
			// ގ
			if(node.top() == "Material"){
				Vec3 col(0.0);
				if(data.find("diffuseColor") != string::npos){
					StringToVec3s(data, "diffuseColor", col);
					cout << "diffuse : " << col << endl;
				}
			}

			if(indexed_face_set){
				// _W
				if(node.top() == "point"){
					vector< vector<double> > vrts0;
					int n = ExtractVector(data, ",", " ", 3, vrts0);
					for(int i = 0; i < n; ++i){
						vrts.push_back(Vec3(vrts0[i][0], vrts0[i][1], vrts0[i][2]));
					}
				}

				// _@
				if(node.top() == "vector"){
					vector< vector<double> > vnms0;
					int n = ExtractVector(data, ",", " ", 3, vnms0);
					for(int i = 0; i < n; ++i){
						vnms.push_back(Vec3(vnms0[i][0], vnms0[i][1], vnms0[i][2]));
					}
				}

				// |S_CfbNX
				if(node.top() == "coordIndex"){
					vector< vector<int> > idxes0;
					int n = ExtractVector(data, "-1", ",", 3, idxes0);
					for(int i = 0; i < n; ++i){
						for(int j = 0; j < (int)idxes0[i].size(); ++j){
							idxes0[i][j] += shape_verts;
						}
						idxes.push_back(idxes0[i]);
					}
				}

				// J[CfbNX
				if(node.top() == "colorIndex"){
					int n = ExtractVector(data, "-1,", ",", 3, color_idxes);
				}
			}
		}

		// "}""["m[hX^bN|bv
		if((pos = buf.find_first_of("]}")) != string::npos){
			if(node.top().find("IndexedFaceSet") != string::npos){
				indexed_face_set = false;
			}
			if(node.top() == "Shape"){
				shape = false;
				shape_verts = (int)vrts.size();
			}
			node.pop();
		}

	}

	file.close();

	if(!idxes.empty()){
		// |S(rxFace)̍쐬
		if(triangle){
			for(int i = 0; i < (int)idxes.size(); ++i){
				int n = (int)idxes[i].size();

				if(n == 3){
					rxFace face;
					face.resize(3);
					face.material_name = "";
					for(int j = 0; j < 3; ++j){
						face[j] = idxes[i][j];
					}
					plys.push_back(face);
				}
				else if(n >= 4){
					PolyToTri(plys, idxes[i], "");
				}
			}
		}
		else{
			for(int i = 0; i < (int)idxes.size(); ++i){
				int n = (int)idxes[i].size();

				rxFace face;
				face.resize(n);
				face.material_name = "";
				for(int j = 0; j < n; ++j){
					face[j] = idxes[i][j];
				}
				plys.push_back(face);
			}
		}

		// _@vZ
		if(vnms.empty()){
			CalVertexNormals(vrts, (int)vrts.size(), plys, (int)plys.size(), vnms);
		}

		return true;
	}

	return true;


	return false;
}

/*!
 * OBJt@C
 * @param[in] file_name t@C(tpX)
 * @param[in] vrts _W
 * @param[in] vnms _@
 * @param[in] plys |S
 * @param[in] mats ގ
 */
inline bool rxVRML::Save(string file_name, 
						 const vector<Vec3> &vrts, 
						 const vector<Vec3> &vnms, 
						 const vector<rxFace> &plys, 
						 const rxMTL &mats)
{
	if((int)plys.size() == 0) return false;

	ofstream file;

	file.open(file_name.c_str());
	if(!file || !file.is_open() || file.bad() || file.fail()){
		cout << "rxOBJ::Save : Invalid file specified" << endl;
		return false;
	}

	// VRMLo[W

	file << "#VRML V2.0 utf8\n";
	file << "Transform\n"; 
	file << "{\n";
	file << "  children [\n";
	file << "	   Group\n";
	file << "	   {\n";
	file << "	   children [\n";
	file << "	   Group\n";
	file << "	   {\n";
	file << "	   children [\n";
	file << endl;
	file << "	Shape {\n";
	file << "	  geometry IndexedFaceSet\n"; 
	file << "	  {" << endl;

	int n;
	Vec3 pos;

	// _W̏o
	file << "		 coord Coordinate\n";
	file << "		 {\n"; 
	file << "		   point [\n";
	n = (int)vrts.size();
	for(int i = 0; i < n; ++i){
		file << "			 " << vrts[i][0] << " " << vrts[i][1] << " " << vrts[i][2];
		file << ((i == n-1) ? "\n" : ",\n");
	}
	file << "		   ]\n";
	file << "		  }" << endl;

	// _@̏o
	if(!vnms.empty()){
		file << "		 normal Normal\n";
		file << "		 {\n"; 
		file << "		   vector [\n";
		n = (int)vnms.size();
		for(int i = 0; i < n; ++i){
			file << "			 " << vnms[i][0] << " " << vnms[i][1] << " " << vnms[i][2];
			file << ((i == n-1) ? "\n" : ",\n");
		}
		file << "		   ]\n";
		file << "		  }" << endl;
	}

	// ʑ̏o
	vector<int> idx;
	bool first_flag = true;

	file << "		  coordIndex [\n";
	n = (int)plys.size();
	for(int i = 0; i < n; ++i){
		int m = (int)plys[i].size();
		file << "			";
		for(int j = 0; j < m; ++j){
			file << plys[i][j] << ", ";
		}
		file << ((i == n-1) ? "-1\n" : "-1,\n");
	}
	file << "		  ]\n";
	file << endl;


	file << "solid FLASE\n";
	file << endl;
	file << "   }\n";
	file << "   }\n";
	file << "   ]\n"; 
	file << "   }\n";
	file << "  ]\n"; 
	file << endl;
	file << "	}\n";
	file << "  ]\n";
	file << "  bboxSize 0 0 0\n";
	file << "}" << endl;

	file.close();

	return true;
}




#endif // _RX_VRML_H_
